//=============================================================================
//  Unified IO - C++ interface for industrial IO cards
//  Copyright (C) 2000, UAB "BBD SOFT" ( http://www.bbdsoft.com/ )
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. Any use is at your own risk.
//
// Permission to use or copy this software for any purpose is hereby granted
// without fee, provided the above notices are retained on all copies.
// Permission to modify the code and to distribute modified code is granted,
// provided the above notices are retained, and a notice that the code was
// modified is included with the above copyright notice.
//
//  The author of this program may be contacted at developers@bbdsoft.com
//=============================================================================


#ifndef _BTRACE_
   #include "btrace.hpp"
#endif

#ifndef _PCI72XXCARD_
   #include "PCI72XXCard.hpp"
#endif

#ifndef _STDEXCEPT_
   #include <stdexcept>
#endif

#ifndef _IOPORT_
   #include "ioport.hpp"
#endif

#ifndef _IOTRACK_
   #include "iotrack.hpp"
#endif

#ifndef _PCIBUS_
   #include "pcibus.hpp"
#endif

using namespace std;
using namespace Uniio;

//-----------------------------------------------------------------------------
PCI72xxCard::PCI72xxCard( const unsigned inCardId )
   : IOCard ("PCI-7248/7296")
   , theIrqNo (0)
   , theBaseAddr (0)
   , theLcrBase (0)
   , theInitFlag (0)
   , theMask (NULL)
{
BFUNCTRACE_DEVELOP();

  const PCIDeviceInfo * anInfo = PCIBus::pciBus().getDeviceInfo(0x10b5, 0x9050, inCardId );
  if (anInfo)
  {
     // we found older style Arbor technologies card
     // lets verify it
     if (  anInfo -> subVendorId() != 0x9999
        && anInfo -> subSystemId() != 0x7248
        && anInfo -> subSystemId() != 0x7296
        )
     {
         throw runtime_error ("PCI72xx: Card not found in the system.");
     } // end if
     theInitFlag = anInfo -> subSystemId();
  }
  else
  {
     // lets try to find new style card
     anInfo = PCIBus::pciBus().getDeviceInfo ( 0x144A, 0x7248, inCardId );
     if (!anInfo)
     {
        anInfo = PCIBus::pciBus().getDeviceInfo ( 0x144A, 0x7296, inCardId );
        if (!anInfo)
        {
           throw runtime_error ("PCI72xx: Card not found in the system.");
        }
     }
     theInitFlag = anInfo -> deviceId();
  }

   theIrqNo =  anInfo -> interruptLine();
   theBaseAddr = anInfo -> baseAddress(2) & (~0x03);
   theLcrBase = anInfo -> baseAddress(1);

   if( theIrqNo >15 || theIrqNo <9 ) theIrqNo = 0;

   // if the lcrBase meet the bug of PLX 9050 then modify lcrBase
   // to avoid PLX 9050 bug
   if( theLcrBase & 0x0080 )
   {
      theLcrBase = (theBaseAddr + 0x100) & 0xFF00;
   }
   else
   {
      theLcrBase &= 0xfffc;
   }


   maxChannels = theInitFlag == 0x7248 ? maxPCI7248Channels : maxPCI7296Channels;
   theMask = new unsigned char [maxChannels*maxBanks];
   for (int i = 0; i <maxChannels*maxBanks; i++ )
   {
      theMask[i] = 0;
   }
}


//-----------------------------------------------------------------------------
PCI72xxCard::~PCI72xxCard()
{
BFUNCTRACE_DEVELOP();
delete [] theMask;
}

//-----------------------------------------------------------------------------
IOCard & PCI72xxCard::updateInputs ( const IOTrack & inTrack
                               , long * inImage
                               , long & inTrackState
                               )
{
BFUNCTRACE_DEVELOP ();

for ( int i = 0
    ; i < maxBanks
    ; i++)
{
   // to increase processing speed update only banks that have at least one input
   if ( theMask[inTrack.startingPoint()*maxBanks + i] != 0xff )
   {
      IOPort dataRegister ( theBaseAddr+i+(numberOfRegisters*inTrack.startingPoint())
                          , false               // write access not required
                          );
      unsigned char newValue = dataRegister.readChar ();

      // PCI72xx card used inverted logic - 0 means ON, and 1 means OFF
      newValue = ~newValue;
      for (int j = 0; j<maxNumberOfPointsInBank; j++)
      {
         inImage[i * maxNumberOfPointsInBank+j] = newValue %2;
         newValue /= 2;
      } // endfor
   } // endif
} // endfor

return * this;
} // end PCI72xxCard::updateInputs (..)


//-----------------------------------------------------------------------------
IOCard & PCI72xxCard::updateOutputs ( const IOTrack & inTrack
                                , const long * const inImage
                                , long & inTrackState
                                )
{
BFUNCTRACE_DEVELOP ();

for ( int i = 0
    ; i < maxBanks
    ; i++)
{
   // to increase processing speed update only banks that have at least one output
   if (theMask[inTrack.startingPoint()*maxBanks+i])
   {
      unsigned char newValue = 0;
      for (int j = 0; j<maxNumberOfPointsInBank; j++)
      {
         if ( inImage[i*maxNumberOfPointsInBank+j] )
         {
            newValue |= (1 << j);
         } // endif
      } // endfor

      // this is needed to prevent writing 1 to input points
      newValue &= theMask[inTrack.startingPoint()*maxBanks+i]; 

      // PCI72xx card used inverted logic - 0 means ON, and 1 means OFF
      newValue = ~newValue;

      IOPort dataRegister (theBaseAddr+i+(numberOfRegisters*inTrack.startingPoint()));
      dataRegister.write (newValue);
   } // endif
} // endfor

return * this;
} // end PCI72xxCard::updateOutputs (..)


//-----------------------------------------------------------------------------
IOCard & PCI72xxCard::addTrack ( IOTrack & inTrack )
{
BFUNCTRACE_DEVELOP ();

if ( inTrack.numberOfPoints() != maxNumberOfPoints )
{
   logic_error exc ("PCI72xx track can be only 24 points long.");
   throw exc;
} // endif

if ( inTrack.startingPoint() >= maxChannels )
{
   logic_error exc ("PCI72xx track starting point too big.");
   throw exc;
} // endif

   // pack IO mask for 24 points into one long
   unsigned long aChannelMask = 0;
   for (int j = 0; j < maxNumberOfPoints; j++)
   {
      if ( inTrack.pointType(j) == IOTrack::digitalOutput )
      {
         aChannelMask |= (1 << j);
      }
      else
      {
         if (   inTrack.pointType(j) != IOTrack::digitalInput
             && inTrack.pointType(j) != IOTrack::notUsed
            )
         {
            logic_error exc ("PCI72xx card only supports digital IO.");
            throw exc;
         } // endif
      } // endifelse
   } // endfor


   unsigned char (&byteArray)[4] = (unsigned char(&)[4])aChannelMask;

   if (byteArray [0] != 0 && byteArray [0] != 0xFF)
   {
      logic_error exc ("PCI72xx track first 8 points can not be mixed.");
   }
   if (byteArray [1] != 0 && byteArray [1] != 0xFF)
   {
      logic_error exc ("PCI72xx track second 8 points can not be mixed.");
   }
   if (  byteArray [2] != 0    && byteArray [2] != 0xFF
      && byteArray [2] != 0xF0 && byteArray [2] != 0x0F
      )
   {
      logic_error exc ("PCI72xx track Last 8 points can be mixed only in groups of 4.");
   }

   unsigned char ctrlValue = 0;
  
   // configure Bank A
   unsigned char bankValue = ~ byteArray [0];
   // if INPUT
   if (bankValue)
   {
      // set bit 5 to 1
      ctrlValue += 1 << 4;
   }
   // configure Bank B
   bankValue = ~ byteArray [1];
   // if INPUT
   if (bankValue)
   {
      // set bit 2 to 1
      ctrlValue += 1 << 1;
   }
   // configure Bank C
   bankValue = ~ byteArray [2];

   unsigned char valueCL = bankValue & 0xF0;
   unsigned char valueCU = bankValue & 0x0F;
   // if INPUT
   if (valueCU)
   {
      // set bit 1 to 1
      ctrlValue += 1;
   }
   // if INPUT
   if (valueCL)
   {
      // set bit 4 to 1
      ctrlValue += 1 << 3;
   }
   IOPort aPort ( theBaseAddr+regControl+(numberOfRegisters*inTrack.startingPoint()));
   aPort.write ( ctrlValue );

   for (int i = 0; i< maxBanks; i++)
   {
      theMask[(inTrack.startingPoint()*maxBanks)+i] = byteArray [i];
   } // endfor

return * this;
} // end PCI72xxCard::addTrack (..)

