/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// Command-line XS Board configure/upload/download utility
//

// $Log: xsload.cpp $
// Revision 1.8  2000/09/10 17:51:48  DEVB
// Replaced endl with newline since xserror objects don't like endl anymore.
//
// Revision 1.7  1999/09/24 18:24:08  DEVB
// added options for selecting programmable oscillator clock source
// and for selecting the directory where the RAM interface bitstreams
// are stored.
//
// Revision 1.6  1999/07/06 09:18:52  DEVB
// added option to upload data from XS Board RAM to a hex file.
//
// Revision 1.5  1999/06/29 16:49:04  DEVB
// added ability to set programmable osc on XS95 Board.
//
// Revision 1.4  1999/06/23 02:11:18  DEVB
// added option to set clock divisor of the DS1075 programmable oscillator
// on the XS40 Board.
//
// Revision 1.3  1999/03/30 19:46:28  DEVB
// Added code to display release version number
//
// Revision 1.2  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 21:49:31  DEVB
// Initial revision
//

#include <stdlib.h>
#include <assert.h>
#include <strstrea.h>
#include <fstream.h>
#include <string>
using std::string;

#include "XSError.h"
#include "XS95Brd.h"
#include "XSVBoard.h"
#include "XSABoard.h"
#include "XS40Brd.h"
#include "Utils.h"

// version info
string version("$Name: V4_0_0 $");

// global variables
XSError errMsg(cerr);		// setup error channel
string lpt;					// parallel port identifier
unsigned int portNum = -1;	// default parallel port number
string brdModel;			// XS Board model number
string ramFileFormat;		// upload RAM file format
string flashFileFormat;		// upload Flash file format
bool upload = false;		// enable reading of RAM or Flash
unsigned int loAddr, hiAddr;	// address boundaries for reading RAM
int nBitstreams = 0;		// number of FPGA/CPLD bitstream files
int posBitstreams = -1;		// position of bitstreams in argument array
int nRAMfiles = 0;			// number of RAM data files
int posRAMfiles = -1;		// position of RAM data files in argument array
int nFlashfiles = 0;		// number of Flash data files
int posFlashfiles = -1;		// position of Flash data files in argument array
bool doBatch = false;	// batch test or requires user confirmation?


// begin methods and functions

// print a string that shows the options and how to use this program.
void Usage(string& progName)
{
	cerr << progName.data() 
		<< "\n\t[-[h|help]]: get help"
		<< "\n\t[-[p|port] [1|2|3]]: select parallel port"
		<< "\n\t[-[b|board] <type>]: select the board type to test"
		<< "\n\t[-[fpga|cpld] <files>]: download files to FPGA or CPLD"
		<< "\n\t[-[ram] <files>]: download files to RAM"
		<< "\n\t[-[flash|seeprom] <files>]: download files to flash or serial EEPROM"
		<< "\n\t[-[u|upload] <lo> <hi>]: upload RAM or Flash data between low and high addresses"
		<< "\n\t[-[f|format] [hex|mcs|exo-[16|24|32]|xess-[16|24|32]]]: file format for upload"
		<< "\n";
	int pos = version.find("Name:");
	if(pos < 0)
	{
		cerr << "Version number unavailable!!" << endl;
	}
	else
	{
		string versionNumber = version.substr(pos+5,version.length()-pos-5);
		for(int i=versionNumber.length()-1; i>=0; i--)
		{
			if(versionNumber[i]=='_')
				versionNumber[i] = '.';
			else if(versionNumber[i]=='$')
				versionNumber[i] = ' ';
		}
		cerr << "Version " << versionNumber.data() << endl;
	}
}


// process the command line options and set the appropriate flags.
bool ProcessOpts(int* argc, char** argv)
{
	int *numFiles = NULL;	// pointer to number of FPGA/RAM/Flash files
	int *posFiles = NULL;	// pointer to position of FPGA/RAM/Flash files

	bool ok = true;	// no errors yet

	for(int i=1; i< *argc; i++)
	{
		int cnt, j;
		string arg = argv[i];
		
		if(arg[0] == '-')  // found an option
		{
			arg = ConvertToUpperCase(arg);	// convert option to upper case
			
			// ask for help but we don't give much
			if(arg=="-H" || arg=="-HELP")
			{
				Usage(arg);
				cnt = 1;
			}
			
			// specify the parallel port being used
			else if(arg=="-PORT" || arg=="-P")
			{
				istrstream argstr(argv[i+1]);
				argstr >> portNum;
				lpt = (string)"LPT" + (string)(argv[i+1]);
				cnt = 2;
			}
			
			// enable batch operation
			else if(arg=="-BATCH")
			{
				doBatch = true;
				cnt = 1;
			}
			
			// specify the board model number
			else if(arg=="-BOARD" || arg=="-B")
			{
				brdModel = ConvertToUpperCase((string)(argv[i+1]));
				cnt = 2;
			}
			
			// specify that the programmable logic device is to be programmed
			else if(arg=="-FPGA" || arg=="-CPLD")
			{
				if(nBitstreams > 0)
				{
					ok = false;
					errMsg.SetSeverity(XSErrorMinor);
					errMsg << "cannot have more than one list of Flash/SEEPROM files" << "\n";
					errMsg.EndMsg();
					cnt = 1;
				}
				else
				{
					numFiles = &nBitstreams;
					posFiles = &posBitstreams;
					cnt = 1;
				}
			}
			
			// specify that the RAM is to be programmed
			else if(arg=="-RAM")
			{
				if(nRAMfiles > 0)
				{
					ok = false;
					errMsg.SetSeverity(XSErrorMinor);
					errMsg << "cannot have more than one list of RAM files" << "\n";
					errMsg.EndMsg();
					cnt = 1;
				}
				else
				{
					numFiles = &nRAMfiles;
					posFiles = &posRAMfiles;
					cnt = 1;
				}
			}
			
			// specify that the Flash or serial EEPROM is to be programmed
			else if(arg=="-FLASH" || arg=="-SEEPROM")
			{
				if(nFlashfiles > 0)
				{
					ok = false;
					errMsg.SetSeverity(XSErrorMinor);
					errMsg << "cannot have more than one list of Flash/SEEPROM files" << "\n";
					errMsg.EndMsg();
					cnt = 1;
				}
				else
				{
					numFiles = &nFlashfiles;
					posFiles = &posFlashfiles;
					cnt = 1;
				}
			}

			// specify what range of XS Board RAM addresses should be uploaded into a file
			else if(arg=="-UPLOAD" || arg=="-U")
			{
				upload = true;
				istrstream argstr1(argv[i+1]);
				argstr1 >> hex >> loAddr;
				istrstream argstr2(argv[i+2]);
				argstr2 >> hex >> hiAddr;
				cnt = 3;
			}
			
			// specify the upload file format
			else if(arg=="-FORMAT" || arg=="-F")
			{
				ramFileFormat = ConvertToUpperCase((string)(argv[i+1]));
				flashFileFormat = ConvertToUpperCase((string)(argv[i+1]));
				cnt = 2;
			}
			
			// can't figure out what the user wants
			else
			{
				ok = false;
				errMsg.SetSeverity(XSErrorMinor);
				errMsg << "unknown option: " << arg.data() << "\n";
				errMsg.EndMsg();
				cnt = 1;
			}
		}
		else	// found something which is not an option, so it is must be a file name
		{
			if(posFiles==NULL || numFiles==NULL)
			{
				ok = false;
				errMsg.SetSeverity(XSErrorMinor);
				errMsg << "this file is misplaced: " << arg.data() << "\n";
				errMsg.EndMsg();
				cnt = 1;
			}
			else
			{
				if(*posFiles<0)
					*posFiles = i;	// record start position of list of files
				(*numFiles)++;	// increment the number of files in the list
				cnt = 0;
				i++;			// keep the file name on the argument list
			}
		}
			
		// remove the processed options from the argument list
		for(j=i+cnt; j< *argc; j++)
			argv[j-cnt] = argv[j];
		*argc -= cnt;
		i--;
	}

	return ok;
}


main( int argc, char **argv )
{
	// get the board type used in the most recent invocation of an XSTOOLS utility
	brdModel = GetXSTOOLSParameter((string)"BoardType");

	// get the RAM hex file type used in the most recent invocation of an XSTOOLS utility
	ramFileFormat = GetXSTOOLSParameter((string)"RAMFormat");

	// get the Flash hex file type used in the most recent invocation of an XSTOOLS utility
	flashFileFormat = GetXSTOOLSParameter((string)"FlashFormat");

	// get the parallel port used in the most recent invocation of an XSTOOLS utility
	lpt = GetXSTOOLSParameter((string)"LPT");
	if(lpt != "")
	{
		int n = sscanf(lpt.data()+3,"%d",&portNum);
		assert(n != 0);
	}

	// process command-line options
	if(ProcessOpts(&argc,argv) == false)
	{
		Usage(string(argv[0]));
		return 1;	// exit if error in program options
	}

	if(doBatch)
	{
		errMsg.EnableBatch(true);	// turn off error messages
		EnableBatch(true);		// turn off user prompts
	}

	// it's an error if the board model being tested is not specified
	if(brdModel == "")
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"No XS Board model was specified!\n");
		Usage(string(argv[0]));
		return 1;
	}

	// store the model type of the board that was specified
	SetXSTOOLSParameter((string)"BoardType",brdModel);

	// use LPT1 if no other parallel port was specified
	if(portNum == -1)
	{
		portNum = 1;
		lpt = "LPT1";
	}

	// store the parallel port that was specified
	SetXSTOOLSParameter((string)"LPT",lpt);

	// determine the type of XS Board and set the pointer to the board object
	XSBoard* brdPtr;
	if(strncmp(brdModel.data(),"XS95",strlen("XS95"))==0)
		brdPtr = new XS95Board();
	else if(strncmp(brdModel.data(),"XS40",strlen("XS40"))==0)
		brdPtr = new XS40Board();
	else if(strncmp(brdModel.data(),"XSP",strlen("XSP"))==0)
		brdPtr = new XS40Board();
	else if(strncmp(brdModel.data(),"XSA",strlen("XSA"))==0)
		brdPtr = new XSABoard();
	else if(strncmp(brdModel.data(),"XSV",strlen("XSV"))==0)
		brdPtr = new XSVBoard();
	else
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XS Board was specified!\n");
		return 1;
	}

	// initialize the board object
	if(brdPtr->Setup(&errMsg,brdModel.data(),portNum) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Invalid parallel port was selected!\n");
		delete brdPtr;
		return 1;
	}
	
	if(upload==true)
	{ // uploading takes precedence over anything else
		if(nRAMfiles<=0 && nFlashfiles<=0)
		{
			errMsg.SimpleMsg(XSErrorMajor,(string)"No file name given for uploaded data!!\n");
		}
		if(nRAMfiles>0)
		{
			if(ramFileFormat == "")
				ramFileFormat = "HEX";
			
			// store the RAM HEX file format that was specified
			SetXSTOOLSParameter((string)"RAMFormat",ramFileFormat);

			if(brdPtr->UploadRAM(string(argv[posRAMfiles]),ramFileFormat.data(),loAddr,hiAddr) == false)
			{
				delete brdPtr;
				return 1;
			}
		}
		if(nFlashfiles>0)
		{
			if(flashFileFormat == "")
				flashFileFormat = "HEX";
			
			// store the Flash HEX file format that was specified
			SetXSTOOLSParameter((string)"FlashFormat",flashFileFormat);
			
			if(brdPtr->UploadFlash(string(argv[posFlashfiles]),flashFileFormat.data(),loAddr,hiAddr) == false)
			{
				delete brdPtr;
				return 1;
			}
		}
	}
	else
	{ // download RAM and Flash files, then download any FPGA/CPLD files
		// if any files in the RAM list are selected, then download them into RAM
		if(nRAMfiles > 0)
		{
			for(int i=0; i<nRAMfiles; i++)
			{
				if(brdPtr->DownloadRAM(string(argv[posRAMfiles+i]), i==0, i==nRAMfiles-1) == false)
				{
					string instructions = "An error occurred while downloading to the RAM\n\nContinue?";
					if(PromptUser(instructions,PROMPT_OKCANCEL) == RESPONSE_CANCEL)
					{
						delete brdPtr;
						return 1;
					}
				}
				else if(errMsg.IsError())
				{
					string instructions = "An error occurred while downloading to the RAM\n\nContinue?";
					if(PromptUser(instructions,PROMPT_OKCANCEL) == RESPONSE_CANCEL)
					{
						delete brdPtr;
						return 1;
					}
				}
			}
		}
		if(nFlashfiles > 0)
		{
			for(int i=0; i<nFlashfiles; i++)
			{
				if(brdPtr->DownloadFlash(string(argv[posFlashfiles+i]), i==0, i==nFlashfiles-1) == false)
				{
					string instructions = "An error occurred while downloading to the Flash\n\nContinue?";
					if(PromptUser(instructions,PROMPT_OKCANCEL) == RESPONSE_CANCEL)
					{
						delete brdPtr;
						return 1;
					}
				}
				else if(errMsg.IsError())
				{
					string instructions = "An error occurred while downloading to the Flash\n\nContinue?";
					if(PromptUser(instructions,PROMPT_OKCANCEL) == RESPONSE_CANCEL)
					{
						delete brdPtr;
						return 1;
					}
				}
			}
		}
		if(nBitstreams > 0)
		{
			if(brdPtr->Configure(string(argv[posBitstreams])) == false)
			{
				delete brdPtr;
				return 1;
			}
			else if(errMsg.IsError())
			{
				delete brdPtr;
				return 1;
			}
		}
	}

	delete brdPtr;

	return 0;	// 0 indicates a successful download/upload
}
