/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The XSPORT program lets the user force binary values onto the
// data pins of the parallel port
//

// $Log: xsport.cpp $
// Revision 1.5  2000/09/10 17:49:59  DEVB
// Replaced endl with newlines since xserror objects don't like endl anymore.
//
// Revision 1.4  2000/02/10 14:43:43  DEVB
// added -CONTROL option to allow setting of parallel port control bits.
//
// Revision 1.3  1999/03/30 19:46:28  DEVB
// Added code to display release version number
//
// Revision 1.2  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:07:04  DEVB
// Initial revision
//

#include <stdlib.h>
#include <assert.h>
#include <strstrea.h>
#include <fstream.h>
#include <string>
using std::string;

#include "XSError.h"
#include "PPort.h"
#include "Utils.h"

// version info
string version("$Name: V4_0_0 $");

// global variables
string lpt;			// parallel port identifier
int portNum = -1;	// parallel port number

XSError errMsg(cerr); // setup error channel
bool writeControlBits = false; // true if bitstring is destined for control bits


// print a string that shows the options and how to use this program.
void Usage(string& progName)
{
	cerr << progName.data() << " [-[h|help]] [-[p|port] [1|2|3]] <binary number>" << endl;
	int pos = version.find("Name:");
	if(pos < 0)
	{
		cerr << "Version number unavailable!!" << endl;
	}
	else
	{
		string versionNumber = version.substr(pos+5,version.length()-pos-5);
		for(int i=versionNumber.length()-1; i>=0; i--)
		{
			if(versionNumber[i]=='_')
				versionNumber[i] = '.';
			else if(versionNumber[i]=='$')
				versionNumber[i] = ' ';
		}
		cerr << "Version " << versionNumber.data() << endl;
	}
}


// process the command line options and set the appropriate flags.
bool ProcessOpts(int* argc, char** argv)
{
	XSError err(cerr); // setup error channel

	bool ok = true;	// no error yet

	for(int i=1; i< *argc; i++)
	{
		int cnt, j;
		string arg = argv[i];
		
		if(arg[0] == '-')  // found an option
		{
			arg = ConvertToUpperCase(arg);	// convert option to upper case
			
			// ask for help but we don't give much
			if(arg=="-H" || arg=="-HELP")
			{
				Usage(arg);
				cnt = 1;
			}
			
			// specify the printer port being used
			else if(arg=="-PORT" || arg=="-P")
			{
				istrstream argstr(argv[i+1]);
				argstr >> portNum;
				lpt = (string)"LPT" + (string)(argv[i+1]);
				cnt = 2;
			}
			
			// can't figure out what the user wants
			else
			{
				ok = false;
				err.SetSeverity(XSErrorMinor);
				err << "unknown option: " << arg.data() << "\n";
				err.EndMsg();
				cnt = 1;
			}
		}
		else	// found something which is not an option, so it must be a bit string
		{ // skip over this argument and leave it in the argument array
			i++;
			cnt = 0;
		}
			
		// remove the processed options from the argument list
		for(j=i+cnt; j< *argc; j++)
			argv[j-cnt] = argv[j];
		*argc -= cnt;
		i--;
	}

	return ok;
}


int main( int argc, char **argv )
{
	XSError err(cerr); // setup error channel

	lpt = GetXSTOOLSParameter((string)"LPT");
	if(lpt != "")
	{
		int n = sscanf(lpt.data()+3,"%d",&portNum);
		assert(n != 0);
	}
	
	if(ProcessOpts( &argc, argv ) == false)
	{
		Usage(string(argv[0]));
		return 1;	// exit if error in program options
	}

	// use LPT1 if no other parallel port was specified
	if(portNum == -1)
	{
		portNum = 1;
		lpt = "LPT1";
	}

	// store the parallel port that will be used
	SetXSTOOLSParameter((string)"LPT",lpt);

	if(argc<2)
		return 1;	// exit if no binary string of bits

	// create the parallel port object
	PPort port;
	if(port.Setup(&errMsg,portNum,0x3) == false)
	{
		err.SimpleMsg(XSErrorMajor,(string)"Invalid parallel port was selected!\n");
		return 1;
	}
	
	int portValue = port.In(0,7);	// get current value on parallel port

	string bits = argv[1];		// get the string of binary bits
	int i = bits.length()-1;
	for( unsigned int bitMask=1; (i>=0) && (bitMask!=256); bitMask<<=1,i-- )
	{	// iterate through the string while assigning binary values to
		// the parallel port data bits
		switch( bits[i] )
		{
			// binary one is represented by '1', 'h', or 'H'
		case '1':
		case 'H':
		case 'h':
			portValue |= bitMask;
			break;
			// binary zero is represented by '0', 'l', or 'L'
		case '0':
		case 'L':
		case 'l':
			portValue &= ~bitMask;
			break;
			// erroneous binary value if we get here!!
		default:
			errMsg.SetSeverity(XSErrorMinor);
			errMsg << "illegal binary value in " << bits.data() << "\n";
			errMsg.EndMsg();
			break;
		}
	}
	
	// exit without changing the parallel port if there was an error
	if(errMsg.IsError())
		return 1;

	port.Out(portValue,0,7);	// output the new value on the parallel port

	return 0;
}
