/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

// $Log: xssetclk.cpp $

//
// Command-line XS Board oscillator programming utility.
//

#include <ctype.h>
#include <assert.h>
#include <time.h>
#include <strstrea.h>
#include <string>
using std::string;

#include "XSError.h"
#include "XS95Brd.h"
#include "XSVBoard.h"
#include "XSABoard.h"
#include "XS40Brd.h"
#include "Utils.h"

// version info
string version("$Name: V4_0_0 $");

// global variables
string lpt;					// parallel port identifier
int portNum = -1;			// parallel port number
string brdModel;			// XS Board model number
int divisor = 2;			// default divisor for master oscillator frequency
bool externalClk = false;	// enable external master oscillator

// begin methods and functions

// print a string that shows the options and how to use this program.
void Usage(string& progName)
{
	cerr << progName.data() 
		<< "\n\t[-[h|help]]: get help"
		<< "\n\t[-[p|port] [1|2|3]]: select parallel port"
		<< "\n\t[-[b|board] <type>]: select the board type to test"
		<< "\n\t[-[d|divisor] <#>]: clock divisor"
		<< "\n\t[-[e|external]]: use an external clock source"
		<< "\n";
	int pos = version.find("Name:");
	if(pos < 0)
	{
		cerr << "Version number unavailable!!" << "\n";
	}
	else
	{
		string versionNumber = version.substr(pos+5,version.length()-pos-5);
		for(int i=versionNumber.length()-1; i>=0; i--)
		{
			if(versionNumber[i]=='_')
				versionNumber[i] = '.';
			else if(versionNumber[i]=='$')
				versionNumber[i] = ' ';
		}
		cerr << "Version " << versionNumber.data() << "\n";
	}
}


// process the command line options and set the appropriate flags.
bool ProcessOpts(int* argc, char** argv)
{
	XSError err(cerr); // setup error channel

	bool ok = true;	// no error yet

	for(int i=1; i< *argc; i++)
	{
		int cnt, j;
		string arg = argv[i];
		
		if(arg[0] == '-')  // found an option
		{
			arg = ConvertToUpperCase(arg);	// convert option to upper case
			
			// ask for help but we don't give much
			if(arg=="-H" || arg=="-HELP")
			{
				Usage(arg);
				cnt = 1;
			}
			
			// specify the printer port being used
			else if(arg=="-PORT" || arg=="-P")
			{
				istrstream argstr(argv[i+1]);
				argstr >> portNum;
				lpt = (string)"LPT" + (string)(argv[i+1]);
				cnt = 2;
			}
			
			// specify the board model number
			else if(arg=="-BOARD" || arg=="-B")
			{
				brdModel = ConvertToUpperCase((string)(argv[i+1]));
				cnt = 2;
			}
			
			// specify the oscillator divisor
			else if(arg=="-DIVISOR" || arg=="-D")
			{
				istrstream argstr(argv[i+1]);
				argstr >> divisor;
				cnt = 2;
			}
			
			// select an external master oscillator
			else if(arg=="-E" || arg=="-EXTERNAL")
			{
				externalClk = true;
				cnt = 1;
			}
			
			// can't figure out what the user wants
			else
			{
				ok = false;
				err.SetSeverity(XSErrorMinor);
				err << "unknown option: " << arg.data() << "\n";
				err.EndMsg();
				cnt = 1;
			}
		}
		else	// found something which is not an option, so it is illegal
		{
			ok = false;
			err.SetSeverity(XSErrorMinor);
			err << "unknown option: " << arg.data() << "\n";
			err.EndMsg();
			cnt = 1;
		}
			
		// remove the processed options from the argument list
		for(j=i+cnt; j< *argc; j++)
			argv[j-cnt] = argv[j];
		*argc -= cnt;
		i--;
	}

	return ok;
}


int main(int argc, char** argv)
{
	XSError err(cerr); // setup error channel

	// get the board type used in the most recent invocation of an XSTOOLS utility
	brdModel = GetXSTOOLSParameter((string)"BoardType");

	// get the parallel port used in the most recent invocation of an XSTOOLS utility
	lpt = GetXSTOOLSParameter((string)"LPT");
	if(lpt != "")
	{
		int n = sscanf(lpt.data()+3,"%d",&portNum);
		assert(n != 0);
	}

	// process command-line options
	if(ProcessOpts(&argc,argv) == false)
	{
		Usage(string(argv[0]));
		return 1;	// exit if error in program options
	}

	// it's an error if the board model being tested is not specified
	if(brdModel == "")
	{
		err.SimpleMsg(XSErrorMajor,(string)"No XS Board model was specified!\n");
		Usage(string(argv[0]));
		return 1;
	}

	// store the model type of the board that was specified
	SetXSTOOLSParameter((string)"BoardType",brdModel);

	// use LPT1 if no other parallel port was specified
	if(portNum == -1)
	{
		portNum = 1;
		lpt = "LPT1";
	}

	// store the parallel port that was specified
	SetXSTOOLSParameter((string)"LPT",lpt);

	// determine the type of XS Board and set the pointer to the board object
	XSBoard* brdPtr;
	if(strncmp(brdModel.data(),"XS95",strlen("XS95"))==0)
		brdPtr = new XS95Board();
	else if(strncmp(brdModel.data(),"XS40",strlen("XS40"))==0)
		brdPtr = new XS40Board();
	else if(strncmp(brdModel.data(),"XSP",strlen("XSP"))==0)
		brdPtr = new XS40Board();
	else if(strncmp(brdModel.data(),"XSA",strlen("XSA"))==0)
		brdPtr = new XSABoard();
	else if(strncmp(brdModel.data(),"XSV",strlen("XSV"))==0)
		brdPtr = new XSVBoard();
	else
	{
		err.SimpleMsg(XSErrorMajor,(string)"Unknown type of XS Board was specified!\n");
		return 1;
	}

	// initialize the board object
	if(brdPtr->Setup(&err,brdModel.data(),portNum) == false)
	{
		err.SimpleMsg(XSErrorMajor,(string)"Invalid parallel port was selected!\n");
		delete brdPtr;
		return 1;
	}

	brdPtr->SetFreq(divisor,externalClk);

	delete brdPtr;

	return 0;	// 0 indicates a successful download/upload
}