/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The AT17Port object specializes the PPort object so that an XS40 Board serial EEPROM can
// be downloaded through the parallel port.
//

// $Log: at17prt.cpp $
// Revision 1.1  2001/12/30 15:34:59  DEVB
// Initial revision
//
// Revision 1.1  2001/12/30 15:26:33  DEVB
// Handles downloads to AT17C serial EEPROM through parallel port
//



#include <fstream.h>
#include <assert.h>
#include "Utils.h"
#include "HexRecrd.h"
#include "AT17Prt.h"



// begin methods and functions



// Create a Flash upload/download port
AT17Port::AT17Port(void)
{
	progressGauge = NULL;
}

// Create a serial EEPROM download port
AT17Port::AT17Port(
				   XSError* e,	// pointer to error reporting object
				   unsigned int portNum,	// parallel port number
				   unsigned int invMask,	// inversion mask for the parallel port
				   unsigned int pos_clk,	// bit position in parallel port of EEPROM clock pin
				   unsigned int pos_eece,	// bit position in parallel port of EEPROM chip-enable
				   unsigned int pos_eeoe,	// bit position in parallel port of EEPROM output-enable
				   unsigned int pos_din		// bit position in parallel port of EEPROM serial data input
				   )
{
	progressGauge = NULL;
	Setup(e,portNum,invMask,pos_clk,pos_eece,pos_eeoe,pos_din);
}

// Setup a serial EEPROM download port
int AT17Port::Setup(
					XSError* e,	// pointer to error reporting object
				   unsigned int portNum,	// parallel port number
				   unsigned int invMask,	// inversion mask for the parallel port
				   unsigned int pos_clk,	// bit position in parallel port of EEPROM clock pin
				   unsigned int pos_eece,	// bit position in parallel port of EEPROM chip-enable
				   unsigned int pos_eeoe,	// bit position in parallel port of EEPROM output-enable
				   unsigned int pos_din		// bit position in parallel port of EEPROM serial data input
				   )
{
	posCLK   = pos_clk;
	posEECE = pos_eece;
	posEEOE = pos_eeoe;
	posDIN = pos_din;
	return PPort::Setup(e,portNum,invMask);
}

// download the serial EEPROM with the contents of a BIT file
bool AT17Port::ProgramEEPROM(string& bitfileName)
{
	XSError& err = GetErr();
	
	if(bitfileName.length()==0)
		return false;  // stop if no bit file was given
	
	ifstream is(bitfileName.data(),ios::nocreate|ios::binary);  // otherwise open bit file
	if(is.fail() || is.eof()!=0)
	{ // error - couldn't open bitstream file
		err.SetSeverity(XSErrorMajor);
		err << "could not open " << bitfileName.data() << "\n";
		err.EndMsg();
		return false;
	}
	
	// determine the size of the file
	is.seekg(0,ios::end);	// move pointer to end of file
	streampos streamEndPos = is.tellg();	// pointer position = position of end of file
	is.seekg(0,ios::beg);	// return pointer to beginning of file

	progressGauge = new Progress(&err,(string)"Serial EEPROM Download",(string)"Downloading "+StripPrefix(bitfileName),0,streamEndPos);

	bool status = ProgramEEPROM(is);
	
	delete progressGauge;
	
	is.close();  // close-up the bitstream file
	
	return status;
}

// compute an integer from a vector of hex numbers in a stream
static long unsigned int GetInteger(istream& is, int len)
{
	assert(len<=4); // can't handle lengths greater than 32 bits
	long unsigned int lui = 0;
	for(int i=0; i<len; i++)
	{ // process hex digits starting from MSDigit
		unsigned char digit;
		is.read(&digit,1);
		assert(is.eof()==0);
		lui = lui*256 + (long unsigned int)digit;
	}
	return lui;
}

// get the field type indicator from a stream
static int GetType(istream& is)
{
	unsigned char type;
	is.read(&type,1);
	assert(is.eof()==0);
	return type;
}

// pass over fields until a certain field is found
static bool ScanForField(istream& is, unsigned char searchType)
{
	while(true)
	{
		unsigned char type = GetType(is);
		if(type==searchType)
			return true;
		unsigned int length = GetInteger(is,2);
		is.ignore(length);
		if(is.eof()!=0)
			break;
	}
	return false;
}

// program a serial EEPROM with a bitstream
bool AT17Port::ProgramEEPROM(istream& is)
{
	XSError& err = GetErr();

	if(is.eof()!=0)
		return false; // exit if no BIT stream exists

	// Setup the pins of the Atmel EEPROM for programming.
	// The SER_EN_ should already be held low by a shunt on Jumper J6.
	// This holds the EEPROM CE_ low.
	Out(0,posEECE,posEECE);
	Out(0,posEEOE,posEEOE);
	
	// jump over the first field of the BIT file
	long unsigned int fieldLength = GetInteger(is,2);
	assert(fieldLength!=0);
	is.ignore(fieldLength);
	
	// process the second field
	fieldLength = GetInteger(is,2);
	assert(fieldLength==1);
	
	// now look for the field with the bitstream data
	const int BitstreamFieldType = 0x65;	// field type for the bitstream data
	bool status = ScanForField(is,BitstreamFieldType);
	assert(status==true);
	
	// now we are at the start of the configuration bits
	fieldLength = GetInteger(is,4); // get the length of the bit stream
	assert(fieldLength>0);
	
	// now load the EEPROM memory
	unsigned char page[64];
	for(long unsigned int i=0; i<fieldLength; i++)
	{ 
		is.read(&page[i&0x3f],1);	// read the bytes and into a 64-byte page
		if(is.eof())
		{
			break;
		}
		assert(is.eof()==0);	// shouldn't reach end-of-file
		if((i%1000)==0)		// after every 1000 bytes,
			progressGauge->Report(is.tellg());	// give some feedback during the programming process
		if((i&0x3f)==0x3f)
		{ // got a full 64-byte page, so program the EEPROM
			ProgramEEPROMPage(page,i&~0x3F,64);
		}
	}
	if((i&0x3F)!=0)
	{ // BIT file ended before a full 64-byte page was assembled
		ProgramEEPROMPage(page,i&~0x3F,(i&0x3F)+1);
	}
	progressGauge->Report(is.tellg());	// should set gauge to 100%

	// now set the output-enable polarity for the EEPROM to active-high
	Out(1,posEECE,posEECE);	// raise chip-enable high
	Out(1,posEEOE,posEEOE);	// raise output-enable high
	// write 0xFF to address 0x3FFF to set the output-enable polarity
	unsigned char allOne = 0xFF;
	ProgramEEPROMPage(&allOne,0x3FFF,1);
		
	// programming done, so disable the EEPROM
	Out(1,posEECE,posEECE);

	return true;
}

// Send a byte into an Atmel EEPROM
void AT17Port::SendEEPROMByte(unsigned char byte)
{
	for(int i=0; i<8; i++)
	{
		// configuration data pin of parallel port connects to data pin of Atmel EEPROM
		Out(byte&0x80 ? 1:0,posDIN,posDIN);  // byte is sent MSB first
		byte <<= 1;
		Out(1,posCLK,posCLK);
		Out(0,posCLK,posCLK);
	}
	Out(0,posDIN,posDIN);	// this is for the acknowledge bit from the EEPROM
	Out(1,posCLK,posCLK);
	Out(0,posCLK,posCLK);
}

// Program a page of an Atmel EEPROM
void AT17Port::ProgramEEPROMPage(unsigned char* buf,	// buffer with data bytes
							 unsigned int pageAddr,	// address where data will be stored
							 unsigned int len		// number of bytes in buffer
							 )
{
	assert(len<=64);	// an Atmel page is 64 bytes or less
	assert(pageAddr<0x20000L);	// Atmel byte addresses are less than 128K
	
	// initialize clock and data pins
	Out(0,posCLK,posCLK);
	Out(1,posDIN,posDIN);
	// give start signal (data falls while clock is high)
	Out(1,posCLK,posCLK);
	Out(0,posDIN,posDIN);
	Out(0,posCLK,posCLK);
	// send device-address byte with A2=0 and R/W=0 (write enabled)
	SendEEPROMByte(0xA6);
	// send two-byte address of EEPROM page
	SendEEPROMByte(pageAddr>>8);	// send upper byte
	SendEEPROMByte(pageAddr&0xFF);	// send lower bytes
	// send data bytes in page
	for(unsigned int i=0; i<len; i++)
		SendEEPROMByte(buf[i]);
	// give stop signal (data rises while clock is high)
	Out(0,posDIN,posDIN);
	Out(1,posCLK,posCLK);
	Out(1,posDIN,posDIN);
	// wait 10 ms
	InsertDelay(10,MILLISECONDS);
}
