/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The AT49FPort object specializes the PPort object so that an XS Board AT49F Flash can
// be downloaded and uploaded through the parallel port.
//

// $Log: at49fprt.cpp $
// Revision 1.1  2001/12/30 15:36:29  DEVB
// Initial revision
//
// Revision 1.1  2001/12/30 15:29:36  DEVB
// <Empty change comment>
//



#include <fstream.h>
#include <assert.h>
#include "Utils.h"
#include "AT49FPrt.h"



// begin methods and functions



// Create a Flash upload/download port
AT49FPort::AT49FPort(void)
{
	;
}

// Create a Flash upload/download port
AT49FPort::AT49FPort(XSError* e,	// pointer to error reporting object
				   unsigned int portNum,	// parallel port number
				   unsigned int invMask,	// inversion mask for the parallel port
				   unsigned int pos_reset,	// bit position in parallel port of Flash RESET pin
				   unsigned int pos_clk,	// bit position in parallel port of Flash CLK pin
				   unsigned int pos_dolsb,	// bit position in parallel port of LSB of Flash data-out pin
				   unsigned int pos_domsb,	// bit position in parallel port of MSB of Flash data-out pin
				   unsigned int pos_dilsb,	// bit position in parallel port of LSB of Flash data-in pin
				   unsigned int pos_dimsb,	// bit position in parallel port of MSB of Flash data-in pin
				   unsigned int pos_stlsb,	// bit position in parallel port of LSB of Flash status pin
				   unsigned int pos_stmsb)	// bit position in parallel port of MSB of Flash status pin
	: FlashPort(e,portNum,invMask,pos_reset,pos_clk,pos_dolsb,pos_domsb,pos_dilsb,pos_dimsb,pos_stlsb,pos_stmsb)
{
	;
}


struct flashBlock
{
	unsigned int address;
	unsigned int length;
	bool erased;
} static flashBlocks[] = 
{
	{0x00000000, 0x00004000, false},
	{0x00004000, 0x00002000, false},
	{0x00006000, 0x00002000, false},
	{0x00008000, 0x00018000, false},
	{0x00020000, 0x00020000, false},
};

// Program a byte in the Flash
bool AT49FPort::ProgramFlash(unsigned int address, unsigned int data)
{
//	fprintf(stdout,"%02x => (%06x)\n",data,address);
	if(WriteFlashByte(0x5555,0xAA) == false)	return false;
	if(WriteFlashByte(0x2AAA,0x55) == false)	return false;
	if(WriteFlashByte(0x5555,0xA0) == false)	return false;
	if(WriteFlashByte(address,data) == false)	return false;

	unsigned int d1, d2;
	if(ReadFlashByte(address,&d1) == false)
		return false;
	d1 &= 0x40;
	if(ReadFlashByte(address,&d2) == false)
		return false;
	d2 &= 0x40;
	while(d1 != d2)
	{
		d1 = d2;
		if(ReadFlashByte(address,&d2) == false)
			return false;
		d2 &= 0x40;
	}

	if(ReadFlashByte(address,&d1) == false)
		return false;
	if(d1 != data)
	{
		XSError& err = GetErr();
		err.SetSeverity(XSErrorNone);
		err << "data mismatch at " << address << ": " << data << " != " << d1 << "\n";
		err.EndMsg();
		return false;
	}
	return true;
}

// erase the entire contents of the Flash RAM
bool AT49FPort::EraseFlash(void)
{
	if(WriteFlashByte(0x5555,0xAA) == false)	return false;
	if(WriteFlashByte(0x2AAA,0x55) == false)	return false;
	if(WriteFlashByte(0x5555,0x80) == false)	return false;
	if(WriteFlashByte(0x5555,0xAA) == false)	return false;
	if(WriteFlashByte(0x2AAA,0x55) == false)	return false;
	if(WriteFlashByte(0x5555,0x10) == false)	return false;

	Progress* progressGauge = new Progress(NULL,(string)"Flash Erase",(string)"Erasing flash device",0,29);
	for(int i=0; i<30; i++)
	{
		InsertDelay(10000/30,MILLISECONDS);
		progressGauge->Report(i);
	}
	delete progressGauge;

#if 0
	// the toggling of D6 to indicate busy does not work!!
	unsigned int d1 = ReadFlashByte(0) & 0x40;
	unsigned int d2 = ReadFlashByte(0) & 0x40;
	while(d1 != d2)
	{
		d1 = d2;
		d2 = ReadFlashByte(0) & 0x40;
	}
#endif

	// initialize the block erasure flags
	for(i=sizeof(flashBlocks)/sizeof(struct flashBlock)-1; i>=0; i--)
		flashBlocks[i].erased = true;
	return true;
}

// erase the contents of a Flash block
bool AT49FPort::EraseFlashBlock(unsigned int blockIndex)
{
	assert(blockIndex < sizeof(flashBlocks)/sizeof(struct flashBlock));
	unsigned int address = flashBlocks[blockIndex].address;
	if(WriteFlashByte(0x5555,0xAA) == false)	return false;
	if(WriteFlashByte(0x2AAA,0x55) == false)	return false;
	if(WriteFlashByte(0x5555,0x80) == false)	return false;
	if(WriteFlashByte(0x5555,0xAA) == false)	return false;
	if(WriteFlashByte(0x2AAA,0x55) == false)	return false;
	if(WriteFlashByte(address,0x30) == false)	return false;

	unsigned int d1, d2;
	if(ReadFlashByte(address,&d1) == false)
		return false;
	d1 &= 0x40;
	if(ReadFlashByte(address,&d2) == false)
		return false;
	d2 &= 0x40;
	while(d1 != d2)
	{
		d1 = d2;
		if(ReadFlashByte(address,&d2) == false)
			return false;
		d2 &= 0x40;
	}

	flashBlocks[blockIndex].erased = true;
	return true;
}

// reset the Flash so data can be read from it
bool AT49FPort::ResetFlash(void)
{
	return true;
}
