/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// Bitstream objects just store strings of boolean values and
// allow certain operations to be performed on them
//

// $Log: bitstrm.cpp $
// Revision 1.3  2001/12/30 16:04:04  DEVB
// Added method to convert Bitstream to a binary string
//
// Revision 1.2  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 21:54:25  DEVB
// Initial revision
//

#include "bitstrm.h"

#define NUM_OF_WORDS(numBits)	((numBits-(numBits?1:0))/bitsPerLong+1)

const unsigned int bitsPerLong = 8 * sizeof(unsigned long)/sizeof(char);


// begin methods and functions

// allocates a bitstream containing at least n bits.
Bitstream::Bitstream(unsigned int n)
{
	numBits = n;		// remember number of bits in the bitstream
	unsigned int numWords = NUM_OF_WORDS(numBits); // number of words to store it
	bits = new unsigned long[numWords];	// allocate storage
	assert(bits!=NULL);					// make sure storage was allocated
	for( unsigned int i=0; i<numWords; i++ )
		bits[i] = 0;	// clear all bits in stream to 0
}


// frees the storage used by a bitstream.
Bitstream::~Bitstream(void)
{
	delete bits;
	bits = NULL;
}


// returns the number of bits in the bitstream 
unsigned int Bitstream::GetLength(void) const
{
	return numBits;
}


// Compares two bitstreams.  Returns true if they match, false otherwise.
bool Bitstream::operator==(Bitstream& b2) const
{
	if(b2.numBits!=numBits)	// can't be equal unless same length
		return false;
	for(unsigned int i=NUM_OF_WORDS(numBits); i>0; i--)
		if(bits[i-1]!=b2.bits[i-1])		// compare word-by-word
			return false;	// failed - bits don't match
		return true;	// ok, they must be equal
}


// compares the end of 1st bitstream against a subfield of 2nd bitstream
bool Bitstream::Subcompare(unsigned int pos,	// position in 1st bitstream
						   Bitstream& b2,		// 2nd bitstream
						   unsigned int b2Pos) const	// position in 2nd bitstream
{
	assert(pos<numBits);       // field 1 goes from pos -> numBits
	assert(b2Pos<b2.numBits);   // field 2 goes from b2pos -> b2.numBits
	if(numBits-pos > b2.numBits-b2Pos)
		return false;     // no match - field 1 longer than field 2
	unsigned int i,j;
	for(i=pos, j=b2Pos; i<numBits; i++,j++)
		if((*this)[i]!=b2[j])
			return false;	// no match - one or more bits don't match
		return true;	// got through the whole thing, so streams match
}


// copies contents of one bitstream into another (also adjusts size).
Bitstream& Bitstream::operator=(Bitstream& b2)
{
	unsigned int numWords2 = NUM_OF_WORDS(b2.numBits); // # words in b2
	if(NUM_OF_WORDS(numBits) != numWords2)	// compare # words in b1 to b2
	{ // change size of b1 bitstream to match b2			
		delete bits;
		numBits = b2.numBits;
		bits = new unsigned long[numWords2];
		assert(bits!=NULL);
	}
	for(unsigned int i=0; i<numWords2; i++)
		bits[i] = b2.bits[i]; // copy b2 to b1 word-by-word
	return *this;
}


// makes a new copy of a bitstream
Bitstream* Bitstream::Copy(void)
{
	Bitstream* bcopy = new Bitstream(numBits);	// make a new bitstream of the same size
	*bcopy = *this;		// copy contents of original to new bitstream
	return bcopy;
}


// returns the value of a bit from a bitstream.
unsigned int Bitstream::operator[](unsigned int bitIndex) const
{
	unsigned int wordIndex = bitIndex / bitsPerLong;	// word containing the bit
	unsigned int shift     = bitIndex % bitsPerLong;	// bit position within word
	
	// now get the word and shift it so the requested bit is in the LSB.
	// Then mask off the rest and return just a 1 or a 0.
	return ((bits[wordIndex]>>shift) & 1L);
}


// clear a bitstream
void Bitstream::Clear(void)
{
	// count from numWords down to 1
	for(unsigned int i=NUM_OF_WORDS(numBits); i>0; )
		bits[--i] = 0; // decrement before access since index is numWords-1 to 0
}


// sets the value of a bit in a bitstream
void Bitstream::SetBit(
					   unsigned int bitIndex,	// position to set
					   unsigned int val			// value to set bit
					   )
{
	assert(bitIndex<numBits);
	unsigned int wordIndex = bitIndex / bitsPerLong;  // index of word containing bit
	unsigned long bitMask  = 1L << (bitIndex % bitsPerLong);
	
	bits[wordIndex] &= ~bitMask;  // force selected bit to zero
	if(val)
		bits[wordIndex] |= bitMask; // then set it if the input value is 1
}


// Sets the value of multiple bits in a bitstream.  End the string of bit values
// with a number that is neither 0 or 1.
void Bitstream::SetBits(
						unsigned int bitIndex,	// position to start changing bits
						int firstBit, ...		// list of bit values
						)
{
	int bit;
	// progress through the states until a -1 is seen (this ends the string)
	va_list ap;
	for(va_start(ap,firstBit),bit=firstBit;  bit==0 || bit==1; bit=va_arg(ap,int))
		SetBit(bitIndex++,(unsigned int)bit);
	va_end(ap);
}


// XOR two bitstreams and return the resulting bitstream.
Bitstream& Bitstream::operator^(Bitstream& b2) const
{
	assert(numBits==b2.numBits); // operands must have the same size
	Bitstream& result = *(new Bitstream(numBits));  // result is same size as operands
	unsigned int numWords = NUM_OF_WORDS(numBits);
	
	// store a 1 in the result anywhere the two bitstreams differ
	for(unsigned int i=0; i<numWords; i++)
		result.bits[i] = bits[i] ^ b2.bits[i];	// XOR word-by-word
	return result;
}

// concatenate two bitstreams to create a third bitstream
Bitstream& Bitstream::operator+(Bitstream& b2) const
{
	Bitstream& sum = *(new Bitstream(numBits + b2.numBits));
	int i, j;
	for(i=sum.numBits-1, j=numBits-1; j>=0; i--,j-- )
		sum.SetBit(i,(*this)[j]);
	assert(j== -1);
	for( j=b2.numBits-1; j>=0; i--,j-- )
		sum.SetBit(i,b2[j]);
	assert(i== -1);
	assert(j== -1);
	return sum;
}

// converts a bistream to a string of 1s and 0s
string Bitstream::ToString(void) const
{
	// start printing with the MSB which is stored at the
	// end of the bitstream.  Then work our way to the beginning.
	string s = "";
	for(unsigned int i=numBits; i>0; )
		s += ((*this)[--i]==1 ? "1":"0");
	return s;
}

// prints a bitstream to an output stream
ostream& Bitstream::PrintBits(ostream& os) const
{
	os << ToString().data() << endl;
	return os;
}


// helper function which prints a bitstream to an output stream
ostream& operator<<(ostream& os, Bitstream& b)
{
	return b.PrintBits(os);
}
