/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The F28Port object specializes the PPort object so that an XSV Board 28F016 Flash can
// be downloaded and uploaded through the parallel port.
//

// $Log: f28port.cpp $
// Revision 1.1  2001/12/30 15:39:41  DEVB
// Initial revision
//



#include <fstream.h>
#include <assert.h>
#include <string.h>
#include "Utils.h"
#include "Progress.h"
#include "F28Port.h"



// begin methods and functions



// Create a Flash upload/download port
F28Port::F28Port(void)
{
	;
}

// Create a Flash upload/download port
F28Port::F28Port(XSError* e,	// pointer to error reporting object
				   unsigned int portNum,	// parallel port number
				   unsigned int invMask,	// inversion mask for the parallel port
				   unsigned int pos_reset,	// bit position in parallel port of Flash RESET pin
				   unsigned int pos_clk,	// bit position in parallel port of Flash CLK pin
				   unsigned int pos_dolsb,	// bit position in parallel port of LSB of Flash data-out pin
				   unsigned int pos_domsb,	// bit position in parallel port of MSB of Flash data-out pin
				   unsigned int pos_dilsb,	// bit position in parallel port of LSB of Flash data-in pin
				   unsigned int pos_dimsb,	// bit position in parallel port of MSB of Flash data-in pin
				   unsigned int pos_stlsb,	// bit position in parallel port of LSB of Flash status pin
				   unsigned int pos_stmsb)	// bit position in parallel port of MSB of Flash status pin
	: FlashPort(e,portNum,invMask,pos_reset,pos_clk,pos_dolsb,pos_domsb,pos_dilsb,pos_dimsb,pos_stlsb,pos_stmsb)
{
	;
}


static const unsigned int rdyBitMask = 1<<7;

struct flashBlock
{
	unsigned int address;
	unsigned int length;
	bool erased;
} static flashBlocks[] = 
{
	{0x00000000, 0x00010000, false},
	{0x00010000, 0x00010000, false},
	{0x00020000, 0x00010000, false},
	{0x00030000, 0x00010000, false},
	{0x00040000, 0x00010000, false},
	{0x00050000, 0x00010000, false},
	{0x00060000, 0x00010000, false},
	{0x00070000, 0x00010000, false},
	{0x00080000, 0x00010000, false},
	{0x00090000, 0x00010000, false},
	{0x000A0000, 0x00010000, false},
	{0x000B0000, 0x00010000, false},
	{0x000C0000, 0x00010000, false},
	{0x000D0000, 0x00010000, false},
	{0x000E0000, 0x00010000, false},
	{0x000F0000, 0x00010000, false},
	{0x00100000, 0x00010000, false},
	{0x00110000, 0x00010000, false},
	{0x00120000, 0x00010000, false},
	{0x00130000, 0x00010000, false},
	{0x00140000, 0x00010000, false},
	{0x00150000, 0x00010000, false},
	{0x00160000, 0x00010000, false},
	{0x00170000, 0x00010000, false},
	{0x00180000, 0x00010000, false},
	{0x00190000, 0x00010000, false},
	{0x001A0000, 0x00010000, false},
	{0x001B0000, 0x00010000, false},
	{0x001C0000, 0x00010000, false},
	{0x001D0000, 0x00010000, false},
	{0x001E0000, 0x00010000, false},
	{0x001F0000, 0x00010000, false},
};

// Program a byte in the Flash
bool F28Port::ProgramFlash(unsigned int address, unsigned int data)
{
	if(WriteFlashByte(address,0x40) == false)	return false;
	if(WriteFlashByte(address,data) == false)	return false;
	unsigned int status;
	if(ReadFlashStatus(&status) == false)	return false;
	while((status & rdyBitMask) == 0)
	{
		if(ReadFlashStatus(&status) == false)
			return false;
	}

	XSError& err = GetErr();
	const unsigned int programErrorMask = (1<<4)|(1<<3)|(1<<1);
	if((status & programErrorMask) != 0)
	{
		err.SetSeverity(XSErrorNone);
		err << "error programming byte at " << address << ": " << data << "\n";
		err.EndMsg();
		ClearFlashStatus();
		return false;
	}
#if 0
	unsigned int d = ReadFlashByte(address);
	if(d != data)
	{
		err.SetSeverity(XSErrorNone);
		err << "data mismatch at " << address << ": " << data << " != " << d << "\n";
		err.EndMsg();
	}
#endif
	return true;
}

// erase the entire contents of the Flash RAM
bool F28Port::EraseFlash(void)
{
	Progress* progressGauge = new Progress(NULL,(string)"Flash Erase",(string)"Erasing flash device",0,sizeof(flashBlocks)/sizeof(struct flashBlock)-1);
	for(int i=0; i<sizeof(flashBlocks)/sizeof(struct flashBlock); i++)
	{
		if(EraseFlashBlock(i) == false)
		{
			delete progressGauge;
			return false;
		}
		progressGauge->Report(i);
	}
	delete progressGauge;
	return true;
}

// erase the contents of a Flash block
bool F28Port::EraseFlashBlock(unsigned int blockIndex)
{
	assert(blockIndex < sizeof(flashBlocks)/sizeof(struct flashBlock));

	static const unsigned int eraseErrorMask = (1<<5)|(1<<4)|(1<<3)|(1<<1);

	unsigned int address = flashBlocks[blockIndex].address;
	if(WriteFlashByte(address,0x20) == false)	return false;
	if(WriteFlashByte(address,0xD0) == false)	return false;
	unsigned int status;
	if(ReadFlashStatus(&status) == false)	return false;
	while((status & rdyBitMask) == 0)
	{
		if(ReadFlashStatus(&status) == false)
			return false;
	}
	if((status & eraseErrorMask) != 0)
	{
		XSError& err = GetErr();
		err.SetSeverity(XSErrorNone);
		err << "error erasing block at " << address << "\n";
		err.EndMsg();
		ClearFlashStatus();
		return false;
	}
	flashBlocks[blockIndex].erased = true;
	return true;
}

// reset the Flash so data can be read from it
bool F28Port::ResetFlash(void)
{
	return WriteFlashByte(0,0xFF);
}

// read the ID from the Flash
bool F28Port::ReadFlashID(unsigned int* id)
{
	if(WriteFlashByte(0,0x90) == false)		return false;
	unsigned int id1,id2;
	if(ReadFlashByte(0x0,&id1) == false)	return false;
	if(ReadFlashByte(0x1,&id2) == false)	return false;
	*id = id1 | (id2<<8);
	return true;
}

// get the status of the Flash
bool F28Port::ReadFlashStatus(unsigned int* status)
{
	if(WriteFlashByte(0,0x70) == false)		return false;
	return ReadFlashByte(0,status);
}

// clear any error bits in the Flash status register
bool F28Port::ClearFlashStatus(void)
{
	return WriteFlashByte(0,0x50);
}

// lock a Flash block so it can't be altered
bool F28Port::LockFlashBlock(unsigned int blockIndex)
{
	assert(blockIndex < sizeof(flashBlocks)/sizeof(struct flashBlock));

	const unsigned int lockBlockErrorMask = (1<<5)|(1<<4)|(1<<3)|(1<<1);

	unsigned int address = flashBlocks[blockIndex].address;
	if(WriteFlashByte(address,0x60) == false)	return false;
	if(WriteFlashByte(address,0x01) == false)	return false;
	unsigned int status;
	if(ReadFlashStatus(&status) == false)	return false;
	while((status & rdyBitMask) == 0)
	{
		if(ReadFlashStatus(&status) == false)
			return false;
	}
	if((status & lockBlockErrorMask) != 0)
	{
		XSError& err = GetErr();
		err.SetSeverity(XSErrorNone);
		err << "error locking block at " << address << "\n";
		err.EndMsg();
		ClearFlashStatus();
		return false;
	}
	return true;
}

// lock all the block lock bits so they cannot be altered
bool F28Port::MasterLockFlash(void)
{
 	const unsigned int lockBlockErrorMask = (1<<5)|(1<<4)|(1<<3)|(1<<1);

	if(WriteFlashByte(0,0x60) == false)		return false;
	if(WriteFlashByte(0,0xF1) == false)		return false;
	unsigned int status;
	if(ReadFlashStatus(&status) == false)	return false;
	while((status & rdyBitMask) == 0)
	{
		if(ReadFlashStatus(&status) == false)
			return false;
	}
	if((status & lockBlockErrorMask) != 0)
	{
		XSError& err = GetErr();
		err.SetSeverity(XSErrorNone);
		err << "error unlocking block lock bits" << "\n";
		err.EndMsg();
		ClearFlashStatus();
		return false;
	}
	return true;
}

// unlock all the Flash blocks
bool F28Port::UnlockFlashBlocks(void)
{
	const unsigned int unlockBlockErrorMask = (1<<5)|(1<<4)|(1<<3)|(1<<1);
	if(WriteFlashByte(0,0x60) == false)		return false;
	if(WriteFlashByte(0,0xD0) == false)		return false;
	unsigned int status;
	if(ReadFlashStatus(&status) == false)	return false;
	while((status & rdyBitMask) == 0)
	{
		if(ReadFlashStatus(&status) == false)
			return false;
	}
	if((status & unlockBlockErrorMask) != 0)
	{
		XSError& err = GetErr();
		err.SetSeverity(XSErrorNone);
		err << "error unlocking blocks" << "\n";
		err.EndMsg();
		ClearFlashStatus();
		return false;
	}
	return true;
}