/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The FlashPort object specializes the PPort object so that an XS Board Flash can
// be downloaded and uploaded through the parallel port.
//

// $Log: flashprt.h $
// Revision 1.1  2001/12/30 15:41:25  DEVB
// Initial revision
//


#ifndef FlashPort_H
#define FlashPort_H

#include "Progress.h"
#include "HexRecrd.h"
#include "pport.h"


class FlashPort : public PPort
{
	public:

	/* begin FlashPort methods */

	/* method: FlashPort */
	// Create a Flash upload/download port
	 FlashPort(void);

	/* method: FlashPort */
	// Create a Flash upload/download port
	 FlashPort(XSError* e, // pointer to error reporting object
		unsigned int portNum, // parallel port number
		unsigned int invMask, // inversion mask for the parallel port
		unsigned int pos_reset, // bit position in parallel port of Flash RESET pin
		unsigned int pos_clk, // bit position in parallel port of Flash CLK pin
		unsigned int pos_dolsb, // bit position in parallel port of LSB of Flash data-out pin
		unsigned int pos_domsb, // bit position in parallel port of MSB of Flash data-out pin
		unsigned int pos_dilsb, // bit position in parallel port of LSB of Flash data-in pin
		unsigned int pos_dimsb, // bit position in parallel port of MSB of Flash data-in pin
		unsigned int pos_stlsb, // bit position in parallel port of LSB of Flash status pin
		unsigned int pos_stmsb) // bit position in parallel port of MSB of Flash status pin
		;

	/* method: Setup */
	// Setup a Flash upload/download port
	int Setup(XSError* e, // pointer to error reporting object
		unsigned int portNum, // parallel port number
		unsigned int invMask, // inversion mask for the parallel port
		unsigned int pos_reset, // bit position in parallel port of Flash RESET pin
		unsigned int pos_clk, // bit position in parallel port of Flash CLK pin
		unsigned int pos_dolsb, // bit position in parallel port of LSB of Flash data-out pin
		unsigned int pos_domsb, // bit position in parallel port of MSB of Flash data-out pin
		unsigned int pos_dilsb, // bit position in parallel port of LSB of Flash data-in pin
		unsigned int pos_dimsb, // bit position in parallel port of MSB of Flash data-in pin
		unsigned int pos_stlsb, // bit position in parallel port of LSB of Flash status pin
		unsigned int pos_stmsb) // bit position in parallel port of MSB of Flash status pin
		;

	/* method: DownloadFlash */
	// download the Flash with the contents of a HEX file
	bool DownloadFlash(
			string& hexfileName,
			bool doErase	// if true, then erase Flash before downloading data from file
			);

	/* method: DownloadFlash */
	// download the XS Board Flash with the contents arriving through a stream
	bool DownloadFlash(
			istream& is,
			bool doErase	// if true, then erase Flash before downloading data from file
			);

	/* method: DownloadHexRecordToFlash */
	// load Flash data at the addresses given in a hex record
	bool DownloadHexRecordToFlash(HexRecord& hx);

	/* method: WriteFlashByte */
	// write a byte to the Flash through the parallel port
	bool WriteFlashByte(unsigned int address, unsigned int data);

	/* method: UploadFlash */
	bool UploadFlash(
		string& hexfileName, // dump uploaded data to this file
		const char* format,		// hex file format
		unsigned long loAddr, // start fetching data from this address
		unsigned long hiAddr // stop at this address
		);

	/* method: UploadFlash */
	// upload the XS Board Flash to a stream
	bool UploadFlash(
		ostream& os, // dump uploaded data to this stream
		const char* format,		// hex file format
		unsigned long loAddr, // start fetching data from this address
		unsigned long hiAddr // stop at this address
		);

	/* method: UploadHexRecordFromFlash */
	// get Flash data at the addresses given and store it in a hex record
	bool UploadHexRecordFromFlash(
		HexRecord& hx, // hex record to store data in
		unsigned long loAddr, // begin upload from this address
		unsigned long hiAddr // end upload at this address
		);

	/* method: ReadFlashByte */
	// read a byte from the Flash through the parallel port
	bool ReadFlashByte(unsigned int address, unsigned int* b);

	/* method: ProgramFlash */
	// Program a byte in the Flash
	virtual bool ProgramFlash(unsigned int address, unsigned int data) = 0;

	/* method: EraseFlash */
	// erase the entire Flash RAM
	virtual bool EraseFlash(void) = 0;

	/* method: EraseFlashBlock */
	// erase the contents of a Flash block
	virtual bool EraseFlashBlock(unsigned int blockIndex) = 0;

	/* method: ResetFlash */
	// reset the Flash so data can be read from it
	virtual bool ResetFlash(void) = 0;

	/* end FlashPort methods */


	protected:

	Progress *progressGauge;	// indicates progress of operations

	
	private:

	unsigned int posRESET; // bit position in parallel port of RAM RESET pin
	unsigned int posCLK;   // bit position in parallel port of RAM CLK pin
	unsigned int posDOLSB; // bit position in parallel port of LSB of RAM data-out pin
	unsigned int posDOMSB; // bit position in parallel port of MSB of RAM data-out pin
	unsigned int posDILSB; // bit position in parallel port of LSB of RAM data-in pin
	unsigned int posDIMSB; // bit position in parallel port of MSB of RAM data-in pin
	unsigned int posSTLSB; // bit position in parallel port of LSB of RAM status pin
	unsigned int posSTMSB; // bit position in parallel port of MSB of RAM status pin
};

/* begin FlashPort functions */

/* end FlashPort functions */

#endif
