/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// HexRecord objects store hexadecimal records consisting of
// a length, beginning address, hexadecimal data values, checksum,
// and tag.  These records are usually created by reading in
// an Intel hex file.
//

// $Log: hexrecrd.h $
// Revision 1.3  2001/12/30 16:07:24  DEVB
// Added support for Motorola and XESS hex file formats and added support
// for Intel hex files that pass the 64K limit.
//
// Revision 1.2  1999/03/30 18:27:41  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:52:12  DEVB
// Initial revision
//

#ifndef HexRecord_H
#define HexRecord_H

#include <iostream.h>

// these are the supported hex file formats
typedef enum
{
	MotorolaFormat	= 0x00,
	IntelFormat		= 0x10,
	XESSFormat		= 0x20,
} HexFileFormat;

// these are the various types of records found in a hex file
typedef enum
{
	MotoStartRecord					= 0x00,	// first record in file
	MotoDataWith16BitAddressRecord	= 0x01,	// actual address and data
	MotoDataWith24BitAddressRecord	= 0x02,	// actual address and data
	MotoDataWith32BitAddressRecord	= 0x03,	// actual address and data
	MotoEndWith16BitAddressRecord	= 0x09,	// actual address and data
	MotoEndWith24BitAddressRecord	= 0x08,	// actual address and data
	MotoEndWith32BitAddressRecord	= 0x07,	// actual address and data
	IntelDataRecord					= 0x10,	// actual address offset and data
	IntelEOFRecord					= 0x11,	// end of file record
	IntelExtSegAddressRecord		= 0x12,	// specifies bits 4-19 of 20-bit address
	IntelExtLinAddressRecord		= 0x14,	// specifies upper 16 bits of 32-bit address
	IntelStartSegAddressRecord		= 0x13,	// specifies 20-bit execution start address
	IntelStartLinAddressRecord		= 0x15,	// specifies 32-bit execution start address
	XESSDataWith16BitAddressRecord	= 0x20,	// actual address and data
	XESSDataWith24BitAddressRecord	= 0x21,	// actual address and data
	XESSDataWith32BitAddressRecord	= 0x22,	// actual address and data
	InvalidRecord					= 0xFF,	// erroneous record
} HexRecordType;

// various types of errors that can occur when handling hex records
typedef enum
{
	NoHexRecordError,
	StartHexRecordError,
	LengthHexRecordError,
	AddressHexRecordError,
	TypeHexRecordError,
	DataHexRecordError,
	CheckSumHexRecordError
} HexRecordError;


class HexRecord
{
	public:

	/* begin HexRecord methods */

	/* method: HexRecord */
	// construct and initialize an empty hex record
	 HexRecord(void);

	/* method: ~HexRecord */
	// delete a hex record
	 ~HexRecord(void);

	/* method: Setup */
	// setup the entries in a hex record
	void Setup(const char* format);

	/* method: SetFileFormat */
	// set the type of file format for the hex record
	void SetFileFormat(HexFileFormat f);

	/* method: GetFileFormat */
	// get the type of file format for the hex record
	HexFileFormat GetFileFormat(void) const;

	/* method: SetRecordType */
	// set the type tag of the hex record
	void SetRecordType(HexRecordType t);

	/* method: GetRecordType */
	// get the tag of the hex record
	HexRecordType GetRecordType(void) const;

	/* method: IsData */
	// does the hex record contain data?
	bool IsData(void) const;

	/* method: IsValid */
	// is the hex record valid?
	bool IsValid(void) const;

	/* method: SetAddress */
	// set the address at which the hex record data should be loaded
	void SetAddress(unsigned int addr);

	/* method: GetAddress */
	// get address at which hex record data will be loaded
	unsigned int GetAddress(void) const;

	/* method: SetBaseAddress */
	// set the base address for the hex record
	void SetBaseAddress(unsigned int baseAddr);

	/* method: GetBaseAddress */
	// get the base address for the hex record
	unsigned int GetBaseAddress(void) const;

	/* method: SetOffset */
	// set the offset at which the hex record data should be loaded
	void SetOffsetAddress(unsigned int off);

	/* method: GetOffset */
	// get offset at which hex record data will be loaded
	unsigned int GetOffsetAddress(void) const;

	/* method: GetAddressLength */
	// get number of bytes in address
	unsigned int GetAddressLength(void);

	/* method: SetAddressMask */
	// set the address mask
	void SetAddressMask(unsigned int mask);

	/* method: GetAddressMask */
	// get the address mask
	unsigned int GetAddressMask(void) const;

	/* method: SetLength */
	// Set the size of the data area in the hex record.  Delete existing data.
	void SetLength(unsigned int l);

	/* method: GetLength */
	// get the number of data bytes in the hex record
	unsigned int GetLength(void) const;

	/* method: operator[] */
	// index into the hex record to get a single byte
	unsigned char& operator[](unsigned int index) const;

	/* method: CalcCheckSum */
	// calculate the 8-bit checksum of the data in the hex record
	void CalcCheckSum(void);

	/* method: GetCheckSum */
	// get the calculated checksum of the hex record
	unsigned char GetCheckSum(void) const;

	/* method: SetError */
	// set the error flag in the hex record
	void SetError(HexRecordError e);

	/* method: GetError */
	// get the error flag from the hex record
	HexRecordError GetError(void) const;

	/* method: IsError */
	// is there an error in the hex record
	bool IsError(void) const;

	/* method: GetErrMsg */
	// return an error message for the hex record
	const char* GetErrMsg(void) const;


	private:

	HexRecordError err;			// hex record error
	HexFileFormat fileFormat;	// hex file format (XESS, Intel, Motorola)
	HexRecordType recordType;	// type of data stored in hex record
	unsigned int length;		// number of data bytes in the hex record
	unsigned int offset;		// destination address for the hex data
	unsigned int base;			// base address added to offset to create final address
	unsigned int mask;			// masks off upper bits for address lengths less than 32-bits
	bool segmented;				// true if record is for Intel 20-bit segmented data
	unsigned char* data;		// data values stored in the hex record
	unsigned char checkSum;		// checksum for the entire hex record

	/* end HexRecord methods */

};

/* begin HexRecord functions */

/* function: ErrMsg */
// return an error message for the given type of error
const char* ErrMsg(HexRecordError e);

// helper function that reads in a hex record from an input stream
extern istream& operator>> (istream& is, HexRecord& hx);

// helper function that dumps a hex record to an output stream
extern ostream& operator<< (ostream& is, HexRecord& hx);

/* end HexRecord functions */

#endif
