/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// JTAGPort objects add JTAG capabilities to a parallel port object.
// A JTAG port can control the TCK, TMS, and TDI pins and read the
// TDO pin.  It stores and updates the state of the TAP state machine
// based on the TMS level at every TCK pulse.
//

// $Log: jtagport.cpp $
// Revision 1.3  2001/12/30 16:14:02  DEVB
// The parallel port pins that attach to the TCK, TMS, TDI and TDO pins
// are now user-assignable rather than hard-coded.
//
// Revision 1.2  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 21:55:45  DEVB
// Initial revision
//



#include <string>
using std::string;



#include "JTAGPort.h"



// The following array stores the transition table for the TAP
// controller.  It tells us what the next state will be by placing the
// current state code in the first index and the value of
// the TMS input in the second index.
static TAPState nextTAPState[16][2] =
{
	// TMS=0                 TMS=1            CURRENT TAP STATE
	{ RunTestIdle,        TestLogicReset },     // TestLogicReset
	{ RunTestIdle,        SelectDRScan   },     // RunTestIdle
	{ CaptureDR,          SelectIRScan   },     // SelectDRScan
	{ CaptureIR,          TestLogicReset },     // SelectIRScan
	{ ShiftDR,            Exit1DR        },     // CaptureDR
	{ ShiftIR,            Exit1IR        },     // CaptureIR
	{ ShiftDR,            Exit1DR        },     // ShiftDR
	{ ShiftIR,            Exit1IR        },     // ShiftIR
	{ PauseDR,            UpdateDR       },     // Exit1DR
	{ PauseIR,            UpdateIR       },     // Exit1IR
	{ PauseDR,            Exit2DR        },     // PauseDR
	{ PauseIR,            Exit2IR        },     // PauseIR
	{ ShiftDR,            UpdateDR       },     // Exit2DR
	{ ShiftIR,            UpdateIR       },     // Exit2IR
	{ RunTestIdle,        SelectDRScan   },     // UpdateDR
	{ RunTestIdle,        SelectDRScan   },     // UpdateIR
};



// begin methods and functions


// Create a JTAG controller port
JTAGPort::JTAGPort(void)
{
	;
}


// Create a JTAG controller port
JTAGPort::JTAGPort(
				   XSError* e,	// pointer to error reporting object
				   unsigned int portNum,	// parallel port number
				   unsigned int invMask,	// inversion mask for the parallel port
				   unsigned int pos_tck,	// bit position in parallel port of TCK pin
				   unsigned int pos_tms,	// bit position in parallel port of TMS pin
				   unsigned int pos_tdi,	// bit position in parallel port of TDI pin
				   unsigned int pos_tdo		// bit position in parallel port of TDO pin
				   )
				   : PPort(e,portNum,invMask)
{
	Setup(e,portNum,invMask,pos_tck,pos_tms,pos_tdi,pos_tdo);
}



// Setup the JTAG controller port
bool JTAGPort::Setup(
				   XSError* e,	// pointer to error reporting object
				   unsigned int portNum,	// parallel port number
				   unsigned int invMask,	// inversion mask for the parallel port
				   unsigned int pos_tck,	// bit position in parallel port of TCK pin
				   unsigned int pos_tms,	// bit position in parallel port of TMS pin
				   unsigned int pos_tdi,	// bit position in parallel port of TDI pin
				   unsigned int pos_tdo		// bit position in parallel port of TDO pin
				 )
{
	posTCK = pos_tck;
	posTMS = pos_tms;
	posTDI = pos_tdi;
	posTDO = pos_tdo;
	SetTraceOnOff(false,cerr);	// don't trace TAP signals
	return PPort::Setup(e,portNum,invMask);	// return non-zero if error occurs
}



// Set the value of the JTAG clock bit
void JTAGPort::SetTCK(unsigned int b)
{
	Out(b,posTCK,posTCK);
}

// Get the value output on the JTAG clock bit
unsigned int JTAGPort::GetTCK(void)
{
	return In(posTCK,posTCK);
}

// Toggle the TCK output twice (return it to its original level)
void JTAGPort::PulseTCK(void)
{
	assert(GetTCK()==0);  // quiescent state of TCK should be zero
	if(traceFlag)
		*osTrace << GetTAPStateLabel(currentTAPState).data() << "\t"
		<< GetTMS() << " "
		<< GetTDI()	<< " "
		<< GetTDO() << endl;
	SetTCK(~GetTCK());  // toggle TCK output
	SetTCK(~GetTCK());	// toggle it again
	currentTAPState = nextTAPState[currentTAPState][GetTMS()]; // update controller state
}

// returns the text label associated with the given state
string JTAGPort::GetTAPStateLabel(TAPState s)
{
	switch(s)
	{
	case TestLogicReset:  return string("Test-Logic-Reset");
	case RunTestIdle:     return string("Run-Test/Idle");
	case SelectDRScan:    return string("Select_DR-Scan");
	case SelectIRScan:    return string("Select_IR-Scan");
	case CaptureDR:       return string("Capture-DR");
	case CaptureIR:       return string("Capture-IR");
	case ShiftDR:         return string("Shift-DR");
	case ShiftIR:         return string("Shift-IR");
	case Exit1DR:         return string("Exit1-DR");
	case Exit1IR:         return string("Exit1-IR");
	case PauseDR:         return string("Pause-DR");
	case PauseIR:         return string("Pause-IR");
	case Exit2DR:         return string("Exit2-DR");
	case Exit2IR:         return string("Exit2-IR");
	case UpdateDR:        return string("Update-DR");
	case UpdateIR:        return string("Update-IR");
	default:
		assert(true==false);  // should never get here!!
	}
	return string("Unknown TAP state");
}



// Set the value of the JTAG TMS bit
void JTAGPort::SetTMS(unsigned int b)
{
	Out(b,posTMS,posTMS);
}

// Get the value output on the JTAG TMS bit
unsigned int JTAGPort::GetTMS(void)
{
	return In(posTMS,posTMS);
}



// Set the value of the JTAG TDI bit
void JTAGPort::SetTDI(unsigned int b)
{
	Out(b,posTDI,posTDI);
}

// Get the value output on the JTAG TDI bit
unsigned int JTAGPort::GetTDI(void)
{
	return In(posTDI,posTDI);
}



// Get the value of the JTAG TDO output that enters the port
unsigned int JTAGPort::GetTDO(void)
{
	return In(posTDO,posTDO);
}



// Initialize the TAP state controller and the attached JTAG device
void JTAGPort::InitTAP(void)
{
	currentTAPState = TestLogicReset; // this is the state the TAP should start in
	SetTCK(0);  // set clock low; TAP changes state on rising edge
	SetTMS(1);	// set TMS high to reset TAP state machine
	PulseTCK();	// pulse TCK 5 times to reset TAP state machine
	PulseTCK();
	PulseTCK();
	PulseTCK();
	PulseTCK();
	currentTAPState = TestLogicReset; // this is state TAP should be in now
}



// enables/disables trace of JTAG signals
void JTAGPort::SetTraceOnOff(bool f,		// trace if true, no trace if false
							 ostream& os	// trace info goes to this stream
							 )
{
	traceFlag = f;
	osTrace = &os;  // trace messages go to this output stream
}



// Sets the TMS line to the appropriate value such that the next TCK pulse
// will move the TAP controller to the requested nextState.
// If the new state of the TAP controller doesn't match the requested
// state, the subroutine will cause the program to abort.
void JTAGPort::GotoNextTAPState(TAPState nextState)
{
	// determine the correct TMS value to move to the desired state
	SetTMS( nextTAPState[currentTAPState][1]==nextState ? 1:0 );
	PulseTCK();  // pulse TCK to move to the next state
	assert( currentTAPState==nextState );  // ensure new state matches desired state
}



// Moves through a sequence of TAP controller states.
// Terminate state changes when a -1 is reached in the sequence.
void JTAGPort::GoThruTAPSequence(TAPState nextState, ...)
{
	// progress through the states until a -1 is seen
	va_list ap;
	for(va_start(ap,nextState); nextState != -1; nextState=va_arg(ap,TAPState))
		GotoNextTAPState(nextState);
	va_end(ap);
}



// Return the bit from TDO, send a bit to TDI, and pulse TCLK.
// This subroutine assumes that the TCK is 0 and the TAP controller state
// is Shift-DR or Shift-IR upon entry.  That's why we pick up the value
// coming in through TDO right away.
unsigned int JTAGPort::SendRcvBit(unsigned int sendBit)
{
	// Make sure we are in the Shift-IR or Shift-DR state.  Gathering
	// and shifting out data isn't legal otherwise.
	assert( currentTAPState==ShiftIR || currentTAPState==ShiftDR );
	
	// get the value output on the TDO output of the external device
	// when TCK is low
	unsigned int TDOValue = GetTDO();
	
	// now output the specified value to the TDI input of the external device
	SetTDI(sendBit & 1);
	
	// pulse TCK to move the bit into the BSIR or BSDR
	PulseTCK(); //clock goes high then low
	
	// return the value we previously captured off TDO
	return TDOValue;
}



#define MAX(a,b)	((a)<(b)?(b):(a))
// Transmit a bitstream through TDI while receiving a bitstream through TDO.
//
// This subroutine assumes the TAP controller state is
// Shift-IR or Shift-DR upon entry.  Upon termination, this subroutine
// will leave the TAP controller in the Exit1-IR or Exit1-DR state.
//
// Either sendBits or rcvBits can be a zero-length bitstream if you don't
// want to transmit or receive bits during a particular call to this subroutine.
// For example, you may want to load the BSDR with a bit pattern but
// you might not care what data gets shifted out of the BSDR during
// the loading.
//
// SendBits and RcvBits can point to the same bitstream without
// causing problems.
//
// Note that the LSB of a bitstream has an index of 0.
void JTAGPort::SendRcvBitstream(Bitstream& sendBits, Bitstream& rcvBits)
{
	assert(currentTAPState==ShiftIR || currentTAPState==ShiftDR);
	
	// lower the TMS line to make sure the TAP controller stays
	// in the Shift-IR or Shift-DR state for the first length-1 cycles
	SetTMS(0);
	unsigned int length = MAX(sendBits.GetLength(),rcvBits.GetLength());
	assert(length>0);
	for( unsigned int i=0; i<length; i++ )
	{
		unsigned int rcvBit;
		
		// On the last bit, raise the TMS line so the TAP
		// controller will move out of the Shift state into
		// the Exit1 state.
		if( i==length-1 )
			SetTMS(1);
		
		// send the next bit if the bitstream is not empty
		if(sendBits.GetLength() > i)
			rcvBit = SendRcvBit(sendBits[i]);
		/* else just shift in a zero */
		else
			rcvBit = SendRcvBit(0);
		
		// store the received bit if the bitstream is not empty
		if(rcvBits.GetLength() > i)
			rcvBits.SetBit(i,rcvBit);
	}
	assert(currentTAPState==Exit1DR || currentTAPState==Exit1IR);
}



// Load the BSIR with an instruction, execute the instruction,
// and then capture and reload the BSDR.
//
// This subroutine assumes the TAP controller is in the Test-Logic-Reset
// or Run-Test/Idle state upon entry.  The TAP controller is returned to the
// Run-Test/Idle state upon exit.
void JTAGPort::LoadBSIRthenBSDR( Bitstream& instruction,
								Bitstream& send, Bitstream& recv )
{
	// Make sure we are in the Test-Logic-Reset or Run-Test/Idle state.
	assert( currentTAPState==RunTestIdle || currentTAPState==TestLogicReset );
	
	// move the TAP controller to the Shift-IR state
	GoThruTAPSequence(RunTestIdle, SelectDRScan,
		SelectIRScan, CaptureIR, ShiftIR, -1);
	
	// load the instruction opcode into the BSIR
	Bitstream nullBitstream(0);	//   
	SendRcvBitstream(instruction, nullBitstream);
	
	// now activate the instruction by entering the Update-IR state.
	GoThruTAPSequence(UpdateIR, -1);
	
	// if the BSDR length is non-zero, then select the BSDR by moving
	// into the Capture-DR state so that the new pin data is
	// parallel-loaded into the BSDR.  Then move into the
	// Shift-DR state so that the captured data can be shifted out
	// while a new pattern is shifted in. Finally, output the
	// newly loaded data onto the parallel output of the BSDR by moving
	// into the Update-DR state.
	if(send.GetLength() > 0 || recv.GetLength() > 0)
	{
		// move into the Capture-DR state so that the new pin data is
		// parallel loaded into the BSDR.  Then move into the Shift-DR state.
		GoThruTAPSequence(SelectDRScan, CaptureDR, ShiftDR, -1);
		
		// shift out the captured data while new data is shifted in
		SendRcvBitstream(send, recv);
		
		// output the freshly loaded data onto the pins of the external device
		GoThruTAPSequence(UpdateDR, -1);
	}
	
	// return to the idle state
	GoThruTAPSequence( RunTestIdle, -1 );
}
