/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// JTAGPort objects add JTAG capabilities to a parallel port object.
// A JTAG port can control the TCK, TMS, and TDI pins and read the
// TDO pin.  It stores and updates the state of the TAP state machine
// based on the TMS level at every TCK pulse.
//

// $Log: jtagport.h $
// Revision 1.3  2001/12/30 16:14:02  DEVB
// The parallel port pins that attach to the TCK, TMS, TDI and TDO pins
// are now user-assignable rather than hard-coded.
//
// Revision 1.2  1999/03/30 18:27:41  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:54:16  DEVB
// Initial revision
//

#ifndef JTAGPort_H
#define JTAGPort_H

#include <assert.h>
#include <stdarg.h>

#include "bitstrm.h"
#include "PPort.h"


// identifiers for all possible TAP states
typedef enum
{
	TestLogicReset=0, RunTestIdle=1,
		SelectDRScan=2,   SelectIRScan=3,
		CaptureDR=4,      CaptureIR=5,
		ShiftDR=6,        ShiftIR=7,
		Exit1DR=8,        Exit1IR=9,
		PauseDR=10,       PauseIR=11,
		Exit2DR=12,       Exit2IR=13,
		UpdateDR=14,      UpdateIR=15
}TAPState;


class JTAGPort : public PPort
{
	public:

	/* begin JTAGPort methods */

	/* method: JTAGPort */
	// Create a JTAG controller port
	 JTAGPort();

	/* method: JTAGPort */
	// Create a JTAG controller port
	 JTAGPort(XSError* e, // pointer to error reporting object
		unsigned int portNum, // parallel port number
		unsigned int invMask, // inversion mask for the parallel port
		unsigned int pos_tck, // bit position in parallel port of TCK pin
		unsigned int pos_tms, // bit position in parallel port of TMS pin
		unsigned int pos_tdi, // bit position in parallel port of TDI pin
		unsigned int pos_tdo); // bit position in parallel port of TDO pin

	/* method: Setup */
	// Setup a JTAG controller port
	bool Setup(XSError* e, // pointer to error reporting object
		unsigned int portNum, // parallel port number
		unsigned int invMask, // inversion mask for the parallel port
		unsigned int pos_tck, // bit position in parallel port of TCK pin
		unsigned int pos_tms, // bit position in parallel port of TMS pin
		unsigned int pos_tdi, // bit position in parallel port of TDI pin
		unsigned int pos_tdo); // bit position in parallel port of TDO pin

	/* method: SetTCK */
	// Set the value of the JTAG clock bit
	void SetTCK(unsigned int b);

	/* method: GetTCK */
	// Get the value output on the JTAG clock bit
	unsigned int GetTCK(void);

	/* method: PulseTCK */
	// Toggle the TCK output twice (return it to its original level)
	void PulseTCK(void);

	/* method: SetTMS */
	// Set the value of the JTAG TMS bit
	void SetTMS(unsigned int b);

	/* method: GetTMS */
	// Get the value output on the JTAG TMS bit
	unsigned int GetTMS(void);

	/* method: SetTDI */
	// Set the value of the JTAG TDI bit
	void SetTDI(unsigned int b);

	/* method: GetTDI */
	// Get the value output on the JTAG TDI bit
	unsigned int GetTDI(void);

	/* method: GetTDO */
	// Get the value of the JTAG TDO output that enters the port
	unsigned int GetTDO(void);

	/* method: InitTAP */
	// Initialize the TAP state controller and the attached JTAG device
	void InitTAP(void);

	/* method: SetTraceOnOff */
	// enables/disables trace of JTAG signals
	void SetTraceOnOff(bool f, // trace if true, no trace if false
		ostream& os // trace info goes to this stream
		);

	/* method: GetTAPStateLabel */
	// returns the text label associated with the given state
	string GetTAPStateLabel(TAPState s);

	/* method: GotoNextTAPState */
	// Sets the TMS line to the appropriate value such that the next TCK pulse
	// will move the TAP controller to the requested nextState.
	// If the new state of the TAP controller doesn't match the requested
	// state, the subroutine will cause the program to abort.
	void GotoNextTAPState(TAPState nextState);

	/* method: GoThruTAPSequence */
	// Moves through a sequence of TAP controller states.
	// Terminate state changes when a -1 is reached in the sequence.
	void GoThruTAPSequence(TAPState nextState, ...);

	/* method: SendRcvBit */
	// Return the bit from TDO, send a bit to TDI, and pulse TCLK.
	// This subroutine assumes that the TCK is 0 and the TAP controller state
	// is Shift-DR or Shift-IR upon entry.  That's why we pick up the value
	// coming in through TDO right away.
	unsigned int SendRcvBit(unsigned int sendBit);

	/* method: SendRcvBitstream */
	// Transmit a bitstream through TDI while receiving a bitstream through TDO.
	//
	// This subroutine assumes the TAP controller state is
	// Shift-IR or Shift-DR upon entry.  Upon termination, this subroutine
	// will leave the TAP controller in the Exit1-IR or Exit1-DR state.
	//
	// Either sendBits or rcvBits can be a zero-length bitstream if you don't
	// want to transmit or receive bits during a particular call to this subroutine.
	// For example, you may want to load the BSDR with a bit pattern but
	// you might not care what data gets shifted out of the BSDR during
	// the loading.
	//
	// SendBits and RcvBits can point to the same bitstream without
	// causing problems.
	//
	// Note that the LSB of a bitstream has an index of 0.
	void SendRcvBitstream(Bitstream& sendBits, Bitstream& rcvBits);

	/* method: LoadBSIRthenBSDR */
	// Load the BSIR with an instruction, execute the instruction,
	// and then capture and reload the BSDR.
	//
	// This subroutine assumes the TAP controller is in the Test-Logic-Reset
	// or Run-Test/Idle state upon entry.  The TAP controller is returned to the
	// Run-Test/Idle state upon exit.
	void LoadBSIRthenBSDR( Bitstream& instruction,
		Bitstream& send, Bitstream& recv );

	/* end JTAGPort methods */


	private:

	TAPState currentTAPState;		// state of Test Access Port
	bool traceFlag;		// trace JTAG states on/off
	ostream* osTrace;	// output stream for trace info

	// These store the bit positions of the JTAG control lines
	unsigned int posTCK; // position of JTAG clock pin
	unsigned int posTMS; // position of JTAG mode pin
	unsigned int posTDI; // position of JTAG data input pin
	unsigned int posTDO; // position of JTAG data output pin

};

/* begin JTAGPort functions */

/* end JTAGPort functions */

#endif
