/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// OscPort objects specialize a PPort to handle programming of
// a DS1075 programmable oscillator.
//

// $Log: oscport.cpp $
// Revision 1.1  2001/12/30 15:46:20  DEVB
// Initial revision
//



#include <dos.h>
#include <stdlib.h>
#include <stdio.h>
#include <fstream.h>
#include <strstrea.h>
#include <time.h>
#include <string.h>

#include <afxwin.h>

#include "OscPort.h"
#include "utils.h"



// Some values for the DS1075 programmable oscillator

const unsigned int OscCmdLength		= 8;	// length of oscillator commands 
const unsigned int OscDataLength	= 9;	// length of oscillator data
const unsigned int OscCmdWriteDIV	= 1;	// command for writing clock divisor 
const unsigned int OscCmdWriteMUX	= 2;	// command for writing clock multiplexer 

const unsigned int OneDuration		= 5L;	// duration of a logic 1 in the osc. data stream
const unsigned int ZeroDuration		= 112L;	// duration of a logic 0
const unsigned int BitDuration		= 150L;	// total bit duration



// begin methods and functions



// Create a prog. oscillator controller port
OscPort::OscPort(void)
{
	;
}

// Create a prog. oscillator controller port
OscPort::OscPort(XSError* e,	// pointer to error reporting object
				   unsigned int portNum,	// parallel port number
				   unsigned int invMask,	// inversion mask for the parallel port
				   unsigned int pos_osc)	// bit position in parallel port of osc. pin
{
	Setup(e,portNum,invMask,pos_osc);
}



// Setup a prog. oscillator controller port
int OscPort::Setup(XSError* e,	// pointer to error reporting object
				   unsigned int portNum,	// parallel port number
				   unsigned int invMask,	// inversion mask for the parallel port
				   unsigned int pos_osc)	// bit position in parallel port of osc. pin
{
	posOsc = pos_osc;
	return PPort::Setup(e,portNum,invMask);
}



// reset the programmable oscillator
void OscPort::ResetOsc()
{
	Out(0,posOsc,posOsc);				// lower data bit 0
	InsertDelay(1200L,MICROSECONDS);	// wait for DS1075 to see reset level
	Out(1,posOsc,posOsc);				// now remove reset level
	InsertDelay(1200L,MICROSECONDS);	// wait for presence pulse from DS1075
}



// send a bit of data to the programmable oscillator.
void OscPort::SendOscBit(unsigned char b)
{
	if(b)	// send a logic 1
	{
		Out(0,posOsc,posOsc);
		InsertDelay(OneDuration,MICROSECONDS);
		Out(1,posOsc,posOsc);
		InsertDelay(BitDuration-OneDuration,MICROSECONDS);
	}
	else	// send a logic 0
	{
		Out(0,posOsc,posOsc);
		InsertDelay(ZeroDuration,MICROSECONDS);
		Out(1,posOsc,posOsc);
		InsertDelay(BitDuration-ZeroDuration,MICROSECONDS);
	}
}



// send a command to the programmable oscillator
void OscPort::IssueOscCmd(unsigned int cmd)
{
	for(int i=0; i<OscCmdLength; i++, cmd>>=1)
		SendOscBit(cmd&1);
}



// send a data word to the programmable oscillator
void OscPort::SendOscData(unsigned int data)
{
	for(int i=0; i<OscDataLength; i++, data>>=1)
		SendOscBit(data&1);
}



// set the DS1075 oscillator frequency
bool OscPort::SetOscFrequency(int div, bool extOscPresent)
{
	XSError& err = GetErr();
	if(div<1)
	{
		err.SimpleMsg(XSErrorMajor,(string)"Clock divisor must be greater than zero!!\n");
		return false;
	}
	if(div>2052)
	{
		err.SimpleMsg(XSErrorMajor,(string)"Clock divisor must be less than 2053!!\n");
		return false;
	}

	int div1 = div==1 ? 1:0;	// set divide-by-1 bit if clock is not divided down
	int prescaleOff = 1;		// turn off prescalar
	int prescaleByTwo = 0;
	if(!extOscPresent)		// setup prescalar circuitry if DS1075 internal oscillator is used
	{
		if(div>1026)
		{
			prescaleOff = 0;	// turn on the prescalar
			prescaleByTwo = 0;	// enable divide-by-four by turning off the divide-by-two prescalar
			div /= 4;			// reduce divisor
		}
		else if(div>513)
		{
			prescaleOff = 0;	// turn on the prescalar
			prescaleByTwo = 1;	// turn on the divide-by-two prescalar
			div /= 2;
		}
	}
	div = div>513 ? 513:div;	// divisor saturates at 513 if external oscillator is used
	// The divisor must be adjusted to get the frequency right as follows:
	//		original	adjusted	resulting
	//		div			div			frequency
	//		----------------------------------
	//		1			 -1			100 MHz (no division)
	//		2			  0			 50 MHz (divide by 2)
	//		3			  1			 33 MHz (divide by 3)
	//		4			  2			 25 MHz (divide by 4)
	//		...			...			...
	//		511			509			
	//		512			510			
	//		513			511
	div -= 2;

	int powerUp = 1;		// keep the oscillator powered up
	int disable0 = 1;		// disable the OUT0 output of the DS1075 (it's not connected on the XS Boards)
	int mux = (disable0<<5) | (powerUp<<4) | (prescaleByTwo<<3) |(prescaleOff<<2) | (div1<<1) | (extOscPresent ? 1:0);

	Out(1,posOsc,posOsc);	// make sure the osc. data pin starts out high

	// program divisor register of the DS1075
	ResetOsc();
	IssueOscCmd(OscCmdWriteDIV);
	SendOscData(div);
	InsertDelay(240,MICROSECONDS);

	// program multiplexor register of the DS1075
	ResetOsc();
	IssueOscCmd(OscCmdWriteMUX);
	SendOscData(mux);
	InsertDelay(240,MICROSECONDS);

	return true;
}
