/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// PPort objects handle the parallel port data, status, and
// control pins.
//

// $Log: pport.cpp $
// Revision 1.4  2001/12/30 16:09:54  DEVB
// PPort can now use either the DRIVERLINX or UNIIO parallel port drivers,
// fetches the parallel port hardware address from RAM, checks the operation
// of the parallel port to detect access problems.
//
// Revision 1.3  2000/09/10 13:27:22  DEVB
// Replaced endl with newline to avoid problems with loging of
// error messages by the xserror object for display in Windows.
//
// Revision 1.2  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 21:57:52  DEVB
// Initial revision
//


#include <stdlib.h>
#include <conio.h>
#include <assert.h>
#include <strstrea.h>
#include <wtypes.h>

#include "DLPORTIO.H"
#include "PPort.h"
#include "Utils.h"

using namespace Uniio;

// begin methods and functions



//
// Here are some routines for handling primitive read/write of ports
//

// get data byte from a given 16-bit I/O address
static unsigned char Inp(unsigned short address)
{
	return DlPortReadPortUchar(address);
}

// send data byte to a given 16-bit I/O address
static void Outp(unsigned short address, unsigned char byte)
{
	DlPortWritePortUchar(address,byte);
}



//
// Here are some routines for handling bit fields
//

#define LOWFIELDMASK(lo,hi)	((1<<((hi)-(lo)+1))-1)
#define FIELDMASK(lo,hi)	(LOWFIELDMASK(lo,hi)<<(lo))

static const unsigned int minBitPos = 0;
static const unsigned int maxBitPos = 23;

// Get the value stored in a bit field
static unsigned int GetField(unsigned int data,		// port data
							 unsigned int loPos,	// low bit position of field
							 unsigned int hiPos		// high bit position of field
							 )
{
	assert(loPos<=maxBitPos);
	assert(hiPos<=maxBitPos);
	assert(hiPos>=loPos);
	return (FIELDMASK(loPos,hiPos) & data) >> loPos;
}

// Set the value in a bit field and return the entire data value after the
// field is updated.
static unsigned int SetField(unsigned int data,		// port data
							 unsigned int loPos,	// low bit position of field
							 unsigned int hiPos,	// high bit position of field
							 unsigned int newData	// new data for field
							 )
{
	assert(loPos<=maxBitPos);
	assert(hiPos<=maxBitPos);
	assert(hiPos>=loPos);
	return (data & ~FIELDMASK(loPos,hiPos)) |
		((newData & FIELDMASK(0,hiPos-loPos))<<loPos);
}




// begin PPort object methods



// Constructor for a parallel port object.
PPort::PPort(void)
{
	dataPort = statusPort = controlPort = NULL;
}

// Constructor for a parallel port object.  Inversion masks are used to
// correct the effects of the various inverters in the PC and XS Board.
PPort::PPort(
			 XSError* e,				// error reporting channel 
			 unsigned int n,			// parallel port number
			 unsigned int invMask		// inversion mask for data, status, and control output bits
			 )
{
	dataPort = statusPort = controlPort = NULL;
	Setup(e,n,invMask);
}



// Destructor for parallel port object
PPort::~PPort(void)
{
	if(dataPort != NULL) delete dataPort;
	if(statusPort != NULL) delete statusPort;
	if(controlPort != NULL) delete controlPort;
}



// Setup the parallel port object
bool PPort::Setup(
			 XSError* e,				// error reporting channel 
			 unsigned int n,			// parallel port number
			 unsigned int invMask		// inversion mask for data, status, and control output bits
			)
{
	assert(sizeof(unsigned int)/sizeof(char) >= 3);	// value needs to be at least 24-bits wide

	// find out whether to use the alternate parallel port driver
	useAlternateIODriver = (GetXSTOOLSParameter((string)"LPTDRIVER")=="DRIVERLINX");

	// see if the parallel port should be checked for correct operation
	char s[20];
	sprintf(s,"LPT%1dCHECK",n);	// get the parallel port check flag from the parameter file
	string lptCheck;
	if((lptCheck=GetXSTOOLSParameter((string)s)) == "NO")
		chkCounter = 0;				// don't bother to check the port for correct operation
	else
		chkCounter = 100;			// number of times to initially check port for correct operation

	SetErr(e);						// set error reporting channel
	SetInvMask(invMask);			// set read, write inversion masks
	return SetLPTNum(n);			// return false if could not assign to the given parallel port address
}



// Sets the error reporting channel
void PPort::SetErr(XSError* e)
{
	err = e;
}

// Provides access to the error reporting channel
XSError& PPort::GetErr(void)
{
	return *err;
}



// Assignment operator for parallel port objects
PPort& PPort::operator=(PPort& src)
{
	err				= src.err;
	num				= src.num;
	address			= src.address;
	invMask			= src.invMask;
	return *this;
}



// Global parameters for the parallel port
static const unsigned int minPortNum = 1;		// minimum parallel port #
static const unsigned int maxPortNum = 4;		// maximum parallel port #

#include <crtdbg.h>
// Sets up a port object for a given parallel port n
bool PPort::SetLPTNum(unsigned int n)
{
	//	assert( n>=minPortNum && n<=maxPortNum );
	num = n;
	if(n <= maxPortNum)
	{ // find the I/O address for the given parallel port #
		address = 0;
		char s[20];
		sprintf(s,"LPT%1dADDRESS",num);
		string lptAddress;
		if((lptAddress=GetXSTOOLSParameter((string)s)) != "")
			sscanf(lptAddress.data(),"%x",&address);
		if(address==0)
		{
			char lptAddresses[8];
			PhysicalMemory lptAddrBlock(0x00000408,8,false);
			lptAddrBlock.readMemoryBlock((void*)lptAddresses,8, 0);
			address = (lptAddresses[(num-1)*2+1]<<8) + lptAddresses[(num-1)*2];
		}
		
		// return false if no parallel port is at this address
		if(address==0)
			return false;
	}
	else
	{ // directly assign I/O address when parallel port number is not in range
		address = n;
	}
	
	if(!useAlternateIODriver)
	{
		// create ports to read/write the data, control, and status fields of the pport
		OSInterface& o = OSInterface::osinterface(); // this just gets the driver running on WinNT, 2000, XP
		dataPort = new IOPort(address,true);		// writeable data port
		statusPort = new IOPort(address+1,false);	// readable status port
		controlPort = new IOPort(address+2,true);	// writeable control 
		assert((dataPort!=NULL) && (statusPort!=NULL) && (controlPort!=NULL));
	}
	
	return true;	// return true if a parallel port was found
}



// Gets the parallel port number for a parallel port object
unsigned int PPort::GetLPTNum(void) const
{
	return num;
}



// Output a value on the designated pins of the concatenated 24-bit parallel port field
// Bits  0 -  7:	data pins
// Bits 15 -  8:	status pins
// Bits 23 - 16:	control pins
void PPort::Out(unsigned int v,
				unsigned int loPos,	// low bit position of field
				unsigned int hiPos	// high bit position of field
				)
{
	unsigned int d, new_d;
	assert(loPos<=23);
	assert(hiPos<=23);
	assert(loPos<=hiPos);
	if(useAlternateIODriver)
	{
		if(hiPos<8)
			d = Inp(address);
		else if(loPos>15)
			d = Inp(address+2)<<16;
		else if(loPos>=8 && hiPos<=15)
			d = Inp(address+1)<<8;
		else
			d = Inp(address) | (Inp(address+1)<<8) | (Inp(address+2)<<16);
		new_d = (d & ~FIELDMASK(loPos,hiPos)) | (((v<<loPos)^invMask) & FIELDMASK(loPos,hiPos));
		if(hiPos<8)
			Outp(address,new_d);
		else if(loPos>15)
			Outp(address+2,new_d>>16);
		else if(loPos>=8 && hiPos<=15)
			Outp(address+1,new_d>>8);
		else
		{
			Outp(address,new_d);
			Outp(address+1,new_d>>8);
			Outp(address+2,new_d>>16);
		}
		if(chkCounter > 0)
		{ // check the value output on the port to make sure it matches the value that was sent
			// this check is only performed for the first few uses of the parallel port
			chkCounter--;
			if(hiPos<8)
			{
				d = Inp(address);
				new_d &= 0xff;	// zero the unused part of new output data
			}
			else if(loPos>15)
			{
				d = Inp(address+2)<<16;
				new_d &= 0xff0000;	// zero the unused part of new output data
			}
			else if(loPos>=8 && hiPos<=15)
			{
				d = Inp(address+1)<<8;
				new_d &= 0xff00;	// zero the unused part of new output data
			}
			else
			{
				d = Inp(address) | (Inp(address+1)<<8) | (Inp(address+2)<<16);
				new_d &= 0xffffff;	// zero the unused part of new output data
			}
			if(d != new_d)
			{
				err->SimpleMsg(XSErrorFatal,(string)"ERRLOC: Parallel port not responding!!\n\nCHECK YOUR PARALLEL PORT HARDWARE ADDRESS!!\n");
			}
		}
	}
	else
	{
#define WRITECHAR write
		if(hiPos<8)
			d = dataPort->readChar();
		else if(loPos>15)
			d = controlPort->readChar()<<16;
		else if(loPos>=8 && hiPos<=15)
			d = statusPort->readChar()<<8;
		else
			d = dataPort->readChar() | (statusPort->readChar()<<8) | (controlPort->readChar()<<16);
		new_d = (d & ~FIELDMASK(loPos,hiPos)) | (((v<<loPos)^invMask) & FIELDMASK(loPos,hiPos));
		if(hiPos<8)
			dataPort->WRITECHAR((unsigned char)new_d);
		else if(loPos>15)
			controlPort->WRITECHAR((unsigned char)(new_d>>16));
		else if(loPos>=8 && hiPos<=15)
			statusPort->WRITECHAR((unsigned char)(new_d>>8));
		else
		{
			dataPort->WRITECHAR((unsigned char)new_d);
			statusPort->WRITECHAR((unsigned char)(new_d>>8));
			controlPort->WRITECHAR((unsigned char)(new_d>>16));
		}
		
		if(chkCounter > 0)
		{ // check the value output on the port to make sure it matches the value that was sent
			// this check is only performed for the first few uses of the parallel port
			chkCounter--;
			if(hiPos<8)
			{
				d = (dataPort->readChar()) & 0xff;
				new_d &= 0xff;	// zero the unused part of new output data
			}
			else if(loPos>15)
			{
				d = (controlPort->readChar() & 0xff) << 16;
				new_d &= 0xff0000;	// zero the unused part of new output data
			}
			else if(loPos>=8 && hiPos<=15)
			{
				d = (statusPort->readChar() & 0xff) << 8;
				new_d &= 0xff00;	// zero the unused part of new output data
			}
			else
			{
				d = (dataPort->readChar() & 0xff) | ((statusPort->readChar() & 0xff)<<8) | ((controlPort->readChar() & 0xff)<<16);
				new_d &= 0xffffff;	// zero the unused part of new output data
			}
			if(d != new_d)
			{
				cerr << "loPos = " << hex << loPos << "; hiPos = " << hex << hiPos << "; write = " << hex << new_d << "; read = " << hex << d << "\n";
				err->SimpleMsg(XSErrorFatal,(string)"ERRLOC: Parallel port not responding!!\n\nCHECK YOUR PARALLEL PORT HARDWARE ADDRESS!!\n");
			}
		}
	}
}



// Return the current values on the designated pins of the concatenated 24-bit parallel port field
// Bits  0 -  7:	data pins
// Bits 15 -  8:	status pins
// Bits 23 - 16:	control pins
unsigned int PPort::In(unsigned int loPos,	// low bit position of field
						unsigned int hiPos	// high bit position of field
						) const
{
	unsigned int d;
	assert(loPos<=23);
	assert(hiPos<=23);
	assert(loPos<=hiPos);
	if(useAlternateIODriver)
	{
		if(hiPos<8)
			d = Inp(address);
		else if(loPos>15)
			d = Inp(address+2)<<16;
		else if(loPos>=8 && hiPos<=15)
			d = Inp(address+1)<<8;
		else
			d = Inp(address) | (Inp(address+1)<<8) | (Inp(address+2)<<16);
	}
	else
	{
		if(hiPos<8)
			d = dataPort->readChar();
		else if(loPos>15)
			d = controlPort->readChar()<<16;
		else if(loPos>=8 && hiPos<=15)
			d = statusPort->readChar()<<8;
		else
			d = dataPort->readChar() | (statusPort->readChar()<<8) | (controlPort->readChar()<<16);
	}
	return ((d^invMask) & FIELDMASK(loPos,hiPos)) >> loPos;
}



// Set the inversion mask for the concatenated 24-bit parallel port field
// Bits  0 -  7:	inversion mask for data pins
// Bits 15 -  8:	inversion mask for status pins
// Bits 23 - 16:	inversion mask for control pins
void PPort::SetInvMask(unsigned int mask)
{
	invMask = mask;
}