/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// Pport objects handle the parallel port data, status, and
// control pins.
//

// $Log: pport.h $
// Revision 1.3  2001/12/30 16:09:54  DEVB
// PPort can now use either the DRIVERLINX or UNIIO parallel port drivers,
// fetches the parallel port hardware address from RAM, checks the operation
// of the parallel port to detect access problems.
//
// Revision 1.2  1999/03/30 18:27:41  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:53:42  DEVB
// Initial revision
//


#ifndef PPort_H
#define PPort_H

#include "ioport.hpp"
#include "physmem.hpp"
#include "osiface.hpp"
#include "btrace.hpp"
#include "XSError.h"

using namespace Uniio;

class PPort
{
	public:

	/* begin PPort methods */

	/* method: PPort */
	// Constructor for a parallel port object.
	PPort(void);

	/* method: PPort */
	// Constructor for a parallel port object.  Inversion masks are used to
	// correct the effects of the various inverters in the PC and XS Board.
	PPort(
		XSError* e, // error reporting channel 
		unsigned int n, // parallel port number
		unsigned int invMask // inversion mask for data, status, and control output bits
		);

	/* method: ~PPort */
	// Destructor for parallel port object
	~PPort(void);

	/* method: Setup */
	// Setup the parallel port object
	bool Setup(
		XSError* e, // error reporting channel 
		unsigned int n, // parallel port number
		unsigned int invMask // inversion mask for data, status, and control output bits
		);

	/* method: SetErr */
	// sets the error reporting channel
	void SetErr(XSError* e);

	/* method: GetErr */
	// Provides access to the error reporting channel
	XSError& GetErr(void);

	/* method: operator= */
	// Assignment operator for parallel port objects
	PPort& operator=(PPort& src);

	/* method: SetNum */
	// minimum parallel port static const unsigned int maxPortNum = 4;		// maximum parallel port
	// Sets up a port object for a given parallel port n
	bool SetLPTNum(unsigned int n);

	/* method: GetNum */
	// Gets the parallel port number for a parallel port object
	unsigned int GetLPTNum(void) const;

	/* method: Out */
	// Output a value on the designated pins of the concatenated 24-bit parallel port field
	// Bits  0 -  7:	data pins
	// Bits 15 -  8:	status pins
	// Bits 23 - 16:	control pins
	void Out(unsigned int v,
		unsigned int loPos, // low bit position of field
		unsigned int hiPos // high bit position of field
		);

	/* method: In */
	// Return the current values on the designated pins of the concatenated 24-bit parallel port field
	// Bits  0 -  7:	data pins
	// Bits 15 -  8:	status pins
	// Bits 23 - 16:	control pins
	unsigned int In(unsigned int loPos, // low bit position of field
		unsigned int hiPos // high bit position of field
		) const;

	/* method: SetInvMask */
	// Set the inversion mask for the concatenated 24-bit parallel port field
	// Bits  0 -  7:	inversion mask for data pins
	// Bits 15 -  8:	inversion mask for status pins
	// Bits 23 - 16:	inversion mask for control pins
	void SetInvMask(unsigned int mask);

	/* end PPort methods */


	private:

	IOPort *dataPort, *controlPort, *statusPort;	// IO port objects
	XSError* err;					// error reporting object
	unsigned int num;				// parallel port num 1,2,3,4
	unsigned int address;			// I/O address
	unsigned int invMask;			// concatenated inversion mask for data, status, and control
	unsigned int chkCounter;		// counts the number of initial checks to make on parallel port
	bool useAlternateIODriver;		// true if DRIVERLINX is used instead of UNIIO parallel port driver

};

/* begin PPort functions */

/* function: Inp */
//
// Here are some routines for handling primitive read/write of ports
//
// get data byte from a given 16-bit I/O address
static unsigned char Inp(unsigned short address);

/* function: Outp */
// send data byte to a given 16-bit I/O address
static void Outp(unsigned short address, unsigned char byte);

/* function: GetField */
// Get the value stored in a bit field
static unsigned int GetField(unsigned int data, // port data
	unsigned int loPos, // low bit position of field
	unsigned int hiPos // high bit position of field
	);

/* function: SetField */
// Set the value in a bit field and return the entire data value after the
// field is updated.
static unsigned int SetField(unsigned int data, // port data
	unsigned int loPos, // low bit position of field
	unsigned int hiPos, // high bit position of field
	unsigned int newData // new data for field
	);

/* end pport functions */

#endif