/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The RAMPort object specializes the PPort object so that an XS Board RAM can
// be downloaded and uploaded through the parallel port.
//

// $Log: ramport.cpp $
// Revision 1.1  2001/12/30 15:45:21  DEVB
// Initial revision
//



#include <fstream.h>
#include <assert.h>
#include "Utils.h"
#include "HexRecrd.h"
#include "RAMPort.h"



// begin methods and functions



// Create a RAM upload/download port
RAMPort::RAMPort(void)
{
	;
}

// Create a RAM upload/download port
RAMPort::RAMPort(XSError* e,	// pointer to error reporting object
				   unsigned int portNum,	// parallel port number
				   unsigned int invMask,	// inversion mask for the parallel port
				   unsigned int pos_reset,	// bit position in parallel port of RAM RESET pin
				   unsigned int pos_clk,	// bit position in parallel port of RAM CLK pin
				   unsigned int pos_dolsb,	// bit position in parallel port of LSB of RAM data-out pin
				   unsigned int pos_domsb,	// bit position in parallel port of MSB of RAM data-out pin
				   unsigned int pos_dilsb,	// bit position in parallel port of LSB of RAM data-in pin
				   unsigned int pos_dimsb,	// bit position in parallel port of MSB of RAM data-in pin
				   unsigned int pos_stlsb,	// bit position in parallel port of LSB of RAM status pin
				   unsigned int pos_stmsb,	// bit position in parallel port of MSB of RAM status pin
				   unsigned int dWidth,	// width of data word
				   unsigned int aWidth		// width of address
				   )
{
	Setup(e,portNum,invMask,pos_reset,pos_clk,
		pos_dolsb,pos_domsb,
		pos_dilsb,pos_dimsb,
		pos_stlsb,pos_stmsb,
		dWidth,aWidth);
}



// Setup a RAM upload/download port
int RAMPort::Setup(XSError* e,	// pointer to error reporting object
				   unsigned int portNum,	// parallel port number
				   unsigned int invMask,	// inversion mask for the parallel port
				   unsigned int pos_reset,	// bit position in parallel port of RAM RESET pin
				   unsigned int pos_clk,	// bit position in parallel port of RAM CLK pin
				   unsigned int pos_dolsb,	// bit position in parallel port of LSB of RAM data-out pin
				   unsigned int pos_domsb,	// bit position in parallel port of MSB of RAM data-out pin
				   unsigned int pos_dilsb,	// bit position in parallel port of LSB of RAM data-in pin
				   unsigned int pos_dimsb,	// bit position in parallel port of MSB of RAM data-in pin
				   unsigned int pos_stlsb,	// bit position in parallel port of LSB of RAM status pin
				   unsigned int pos_stmsb,	// bit position in parallel port of MSB of RAM status pin
				   unsigned int dWidth,	// width of data word
				   unsigned int aWidth		// width of address
				   )
{
	progressGauge = NULL;
	posRESET = pos_reset;
	posCLK   = pos_clk;
	posDOLSB = pos_dolsb;
	posDOMSB = pos_domsb;
	posDILSB = pos_dilsb;
	posDIMSB = pos_dimsb;
	posSTLSB = pos_stlsb;
	posSTMSB = pos_stmsb;
	dataWidth = dWidth;
	addrWidth = aWidth;
	return PPort::Setup(e,portNum,invMask);
}



// download the RAM with the contents of a HEX file
bool RAMPort::DownloadRAM(string& hexfileName)
{
	XSError& err = GetErr();
	
	if(hexfileName.length()==0)
		return false;  // stop if no hex file was given
	
	ifstream is(hexfileName.data(),ios::nocreate);  // otherwise open hex file
	if(is.fail() || is.eof()!=0)
	{ // error - couldn't open hex file
		err.SetSeverity(XSErrorMajor);
		err << "could not open " << hexfileName.data() << "\n";
		err.EndMsg();
		return false;
	}
	
	// determine the size of the file
	is.seekg(0,ios::end);	// move pointer to end of file
	streampos streamEndPos = is.tellg();	// pointer position = position of end of file
	is.seekg(0,ios::beg);	// return pointer to beginning of file

	progressGauge = new Progress(&err,(string)"RAM Download",(string)"Downloading "+StripPrefix(hexfileName),0,streamEndPos);

	bool status = DownloadRAM(is);
	
	delete progressGauge;
	
	is.close();  // close-up the hex file
	
	return status;
}



// download the XS Board RAM with the contents arriving through a stream
bool RAMPort::DownloadRAM(istream& is)
{
	XSError& err = GetErr();

	assert(progressGauge != NULL);	//	make sure progress indicator is initialized
	progressGauge->Report(0);	// start progress indicator at zero

	// read hex records from file and place data in the RAM of the board
	HexRecord hx;
	while(is.eof()==0)
	{
		is >> hx;
		if(is.eof()!=0)
			break; // terminate loading loop when stream goes dry
		if(err.IsError())
			break;
		if(is.fail())
		{
			err.SetSeverity(XSErrorMajor);
			err << "error reading stream\n";
			err.EndMsg();
			break;
		}
		if(hx.IsError())
		{ // some error in the hex record itself
			err.SetSeverity(XSErrorMajor);
			err << "hex record: " << hx.GetErrMsg() << "\n";
			err.EndMsg();
			break;
		}
		else
		{ // send out an indication for each hex record that's loaded
			progressGauge->Report(is.tellg());
			if(DownloadHexRecordToRAM(hx) == false) // download hex record
				return false;	// an error occurred while downloading the hex record
		}
	}
	progressGauge->Report(is.tellg());	// should set gauge to 100%
	
	return err.IsError() ? false:true;
}



bool RAMPort::UploadRAM(
					   string& hexfileName,		// dump uploaded data to this file
					   const char* format,		// hex file format
					   unsigned long loAddr,	// start fetching data from this address
					   unsigned long hiAddr		// stop at this address
					   )
{
	XSError& err = GetErr();
	
	if(hexfileName.length()==0)
		return false;  // stop if no hex file was given
	
	ofstream os(hexfileName.data());  // otherwise open hex file
	if(os.fail() || os.eof()!=0)
	{ // error - couldn't open hex file
		err.SetSeverity(XSErrorMajor);
		err << "could not open " << hexfileName.data() << "\n";
		err.EndMsg();
		return false;
	}
	
	progressGauge = new Progress(&err,(string)"RAM Upload",(string)"Uploading "+StripPrefix(hexfileName),loAddr,hiAddr);
	
	bool status = UploadRAM(os,format,loAddr,hiAddr);
	
	delete progressGauge;

	os.close();  // close-up the hex file

	return status;
}



// upload the XS Board RAM to a stream
bool RAMPort::UploadRAM(
					   ostream& os,				// dump uploaded data to this stream
					   const char* format,		// hex file format
					   unsigned long loAddr,	// start fetching data from this address
					   unsigned long hiAddr		// stop at this address
					   )
{
	XSError& err = GetErr();

	assert(progressGauge != NULL);	//	make sure progress indicator is initialized
	progressGauge->Report(0);	// start progress indicator at zero
	
	// read hex records from RAM and place data in the hex file
	HexRecord hx;
	hx.Setup(format);
	for(unsigned long addr=loAddr; (addr|0xF)<=hiAddr; addr=(addr+16)&~0xF)
	{
		if(UploadHexRecordFromRAM(hx,addr,addr|0xF) == false) // get 16 bytes from RAM
			return false;	// an error occurred while uploading the hex record
		os << hx;		// send hex record to output stream
		progressGauge->Report(addr);	// give feedback on progress
	}
	if(addr<=hiAddr)
	{ // handle the last few bytes of an upload
		if(UploadHexRecordFromRAM(hx,addr,hiAddr) == false) // get the last few bytes from RAM
			return false;	// an error occurred while uploading the hex record
		os << hx;		// send hex record to output stream
	}
	progressGauge->Report(hiAddr);	// should set gauge to 100%
	
	return true;
}



// load RAM data at the addresses given in a hex record
bool RAMPort::DownloadHexRecordToRAM(HexRecord& hx)
{
	XSError& err = GetErr();

	if(!hx.IsData()) // don't download unless this is RAM data
		return true;	// no error, just didn't do anything

	// the address has to be right-shifted by one bit for each doubling of the data width.
	unsigned int addrScale;
	if(dataWidth<=8)
		addrScale = 0;	// addresses for byte-wide data need no scaling because each byte fits in a hex record byte
	else if(dataWidth<=16)
		addrScale = 1;	// addresses for 16-bit data are halved because each word takes two bytes in the hex record
	else if(dataWidth<=32)
		addrScale = 2;	// addresses for 32-bit data are quartered because each word takes four bytes in the hex record
	else
		assert(1==0);	// we can't handle data widths > 32 bits
	unsigned int stride = 1 << addrScale;

	unsigned int addrMask = ~((1 << addrScale) - 1);

	if(hx.GetLength() % stride)
	{
		err.SimpleMsg(XSErrorMajor,(string)"Cannot download an odd number of bytes to word-wide RAM!\n");
		return false;
	}
	unsigned int loAddr = hx.GetAddress();
	if(loAddr & ~addrMask)
	{
		err.SimpleMsg(XSErrorMajor,(string)"Cannot download to word-wide RAM using an odd byte-starting address!\n");
		return false;
	}
	
	Out(1,posRESET,posRESET);	// reset the downloading state machine
	Out(0,posCLK,posCLK);		// force clock low
	Out(0,posRESET,posRESET);	// release the reset
	unsigned int statusChk = 0;	// set status check to reset state id
	unsigned int status = In(posSTLSB,posSTMSB);
	assert(status==statusChk);	// make sure we are in reset state
	if(status != statusChk)
		return false;	// error - couldn't reset RAM interface state machine
	
	unsigned int address = loAddr >> addrScale;	// get starting address for hex record
	for(int j=addrWidth-4; j>=0; j-=4)
	{
		Out((address>>j)&0xf,posDOLSB,posDOMSB);
		Out(1,posCLK,posCLK);	// latch 4 address bits into downloading circuit
		statusChk++;			// increment status check but don't check it
		Out(0,posCLK,posCLK);
	}
	status = In(posSTLSB,posSTMSB);
	assert(status==statusChk);
	if(status != statusChk)
		return false;	// error - RAM interface state machine not in the right state
	
	// process the bytes from a data-type hex record
	unsigned int statusChkSave = statusChk;
	for(unsigned int i=0; i<hx.GetLength(); i+=stride)
	{
		unsigned int data = 0;	// assemble next data word from hex record
		for(j=0; j<stride; j++)
			data = (data<<8) | (hx[i+j] & 0xFF);
		for(j=dataWidth-4; j>0; j-=4)
		{
			Out((data>>j)&0xf,posDOLSB,posDOMSB);
			Out(1,posCLK,posCLK);	// latch 4 data bits into downloading circuit
			// get the last 4 bits out while the clock is high because the
			// write to RAM will occur when the clock goes low in a few statements 
			if(j==4)
				Out(data&0xf,posDOLSB,posDOMSB);
			statusChk++;				// increment to next state id
			Out(0,posCLK,posCLK);
			status = In(posSTLSB,posSTMSB);
			assert(status==statusChk);
			if(status != statusChk)
				return false;	// error - RAM interface state machine not in the right state
		}
		Out(1,posCLK,posCLK);		// terminate RAM write pulse
		statusChk = statusChkSave;	// loop status check once data word is downloaded
		Out(0,posCLK,posCLK);
		status = In(posSTLSB,posSTMSB);
		assert(status==statusChk);
		if(status != statusChk)
			return false;	// error - RAM interface state machine not in the right state
	}
	
	Out(1,posRESET,posRESET);	// reset the downloading state machine
	Out(0,posRESET,posRESET);	// release the reset
	statusChk = 0;				// set status check to reset state id
	status = In(posSTLSB,posSTMSB);
	assert(status==statusChk);	// make sure we are in reset state
	if(status != statusChk)
		return false;	// error - RAM interface state machine not in the right state
	else
		return true;
}



// get RAM data at the addresses given and store it in a hex record
bool RAMPort::UploadHexRecordFromRAM(
						HexRecord& hx,			// hex record to store data in
						unsigned long loAddr,	// begin upload from this address
						unsigned long hiAddr	// end upload at this address
						)
{
	XSError& err = GetErr();

	// the address has to be right-shifted by one bit for each doubling of the data width.
	unsigned int addrScale;
	if(dataWidth<=8)
		addrScale = 0;	// addresses for byte-wide data need no scaling because each byte fits in a hex record byte
	else if(dataWidth<=16)
		addrScale = 1;	// addresses for 16-bit data are halved because each word takes two bytes in the hex record
	else if(dataWidth<=32)
		addrScale = 2;	// addresses for 32-bit data are quartered because each word takes four bytes in the hex record
	else
		assert(1==0);	// we can't handle data widths > 32 bits
	unsigned int stride = 1 << addrScale;

	unsigned int addrMask = ~((1 << addrScale) - 1);

	if((hiAddr-loAddr+1) % stride)
	{
		err.SimpleMsg(XSErrorMajor,(string)"Cannot upload an odd number of bytes from word-wide RAM!\n");
		return false;
	}
	if(loAddr & ~addrMask)
	{
		err.SimpleMsg(XSErrorMajor,(string)"Cannot upload from word-wide RAM using an odd byte-starting address!\n");
		return false;
	}

	// clear the lower bits of the address bounds
	loAddr = loAddr & addrMask;
	hiAddr = hiAddr & addrMask;

	hx.SetAddress(loAddr);	// set beginning and ending addresses for the hex record
	hx.SetLength(hiAddr-loAddr+stride);
	
	Out(1,posRESET,posRESET);	// reset the downloading state machine
	Out(0,posCLK,posCLK);		// force clock low
	Out(0,posRESET,posRESET);	// release the reset
	unsigned int statusChk = 0;	// set status check to reset state id
	unsigned int status = In(posSTLSB,posSTMSB);
	assert(status==statusChk);	// make sure we are in reset state
	if(status != statusChk)
		return false;	// error - RAM interface state machine not in the right state

	// set initial address into the downloading state machine
	for(int j=addrWidth-4; j>=0; j-=4)
	{
		Out((loAddr>>(addrScale+j))&0xf,posDOLSB,posDOMSB);
		Out(1,posCLK,posCLK);	// latch 4 address bits into downloading circuit
		statusChk++;			// increment status check but don't check it
		Out(0,posCLK,posCLK);
	}
	status = In(posSTLSB,posSTMSB);
	assert(status==statusChk);
	if(status != statusChk)
		return false;	// error - RAM interface state machine not in the right state

	Out(1,posRESET,posRESET);	// reset the downloading state machine
	Out(0,posRESET,posRESET);	// release the reset
	statusChk = 0;				// set status check to reset state id
	status = In(posSTLSB,posSTMSB);
	assert(status==statusChk);	// make sure we are in reset state
	if(status != statusChk)
		return false;	// error - RAM interface state machine not in the right state

	unsigned int data;	// holds data read from RAM

	// process the bytes from a data-type hex record.
	// we have to read up to hiAddr+1 in order to get the data from hiAddr.
	for(unsigned long i=loAddr+stride; i<=hiAddr+stride; i+=stride)
	{
		int k;
		data = 0;
		for(j=addrWidth-4, k=dataWidth; j>=0; j-=4, k-=4)
		{
			Out((i>>(addrScale+j))&0xf,posDOLSB,posDOMSB);
			Out(1,posCLK,posCLK);	// latch 4 address bits into downloading circuit
			statusChk++;			// increment status check
			Out(0,posCLK,posCLK);
			if(k>0)
			{ // read 4 bits of data
				data = (data<<4) | (In(posDILSB,posDIMSB) & 0xf);
			}
			else
			{ // check status if not reading data
				status = In(posSTLSB,posSTMSB);
				assert(status==statusChk);	// make sure we are in the correct state
				if(status != statusChk)
					return false;	// error - RAM interface state machine not in the right state
			}
		}
		for(j=dataWidth-8, k=0; j>=0; j-=8, k++)
			hx[(i-stride)+k-loAddr] = (data>>j) & 0xFF;
		Out(1,posRESET,posRESET);	// reset the downloading state machine
		Out(0,posRESET,posRESET);	// release the reset
		statusChk = 0;				// set status check to reset state id
		status = In(posSTLSB,posSTMSB);
		assert(status==statusChk);	// make sure we are in reset state
		if(status != statusChk)
			return false;	// error - RAM interface state machine not in the right state
	}
	
	hx.CalcCheckSum();	// put the checksum into the hex record
	return true;
}
