/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

// $Log: utils.cpp $
// Revision 1.3  2001/12/30 16:21:42  DEVB
// Moved time delay function into this file.
// Added utility for prompting the user for input in both GUI and command-line.
// Added utility for storing/retrieving info in registry and text file.
// Added utility for retrieving XS Board info from a text file.
//
// Revision 1.2  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 21:58:49  DEVB
// Initial revision
//

//
// Utility routines used by other programs.
//

#include <ctype.h>
#include <assert.h>
#include <time.h>
#include <io.h>
#include <fcntl.h>
#include <string>
using std::string;

#include "Utils.h"

#include <afxole.h>
#include <shlobj.h>
#include <winreg.h>
#include <afxwin.h>
#include <afxcmn.h>

#include "XSError.h"


// begin methods and functions



// convert a string to upper case
string& ConvertToUpperCase(string& s)
{
	for(int i=0; i<s.length(); i++)
		s[i] = toupper(s[i]);
	return s;
}



// strip suffix from a file name
string& StripSuffix(string& fileName)
{
	assert(fileName.empty()==false);
	for(int i=fileName.length()-1; i>=0; i-- )
	{
		if(fileName[i]=='.') // found the last period that delimits the suffix
		{
			fileName.erase(i,fileName.length()-i); // cut off the suffix
			break;
		}
		else if(fileName[i]=='\\') // found the file/directory delimiter
			break;        //  so there is no suffix
	}
	return fileName;  // return the suffix-stripped file name
}



// get suffix from a file name
string GetSuffix(string& fileName)
{
	assert(fileName.empty()==false);
	for(int i=fileName.length()-1; i>=0; i--)
	{
		if(fileName[i]=='.') // found the last period that delimits the suffix
			return ConvertToUpperCase(fileName.substr(i+1,fileName.length()-(i+1)));
		else if(fileName[i]=='\\') // found the file/directory delimiter
			break;
	}
	return string(""); // no suffix. return empty string
}



// strip prefixed directory path from a file name
string& StripPrefix(string& fileName)
{
	assert(fileName.empty()==false);
	for(int i=fileName.length()-1; i>=0; i--)
		if(fileName[i]=='\\') // found the last file/directory delimiter
		{
			// return only the last part of the file name
			fileName = fileName.substr(i+1,fileName.length()-(i+1));
			return fileName;
		}
	return fileName;  // no prefix, so return the original file name
}



// get prefixed directory path from a file name
string GetPrefix(string& fileName)
{
	assert(fileName.empty()==false);
	for(int i=fileName.length()-1; i>=0; i--)
		if(fileName[i]=='\\') // found the last file/directory delimiter
			break;
	return fileName.substr(0,i);	// return the directory path without the file name
}



#pragma optimize( "", off ) // turn off optimization so timing loops don't get removed

// subroutine for delaying by a given number of microseconds or milliseconds
void InsertDelay(unsigned long d, unsigned int time_units)
{
	static long loopsPerMillisecond = -1;
	
	if(loopsPerMillisecond < 0)
	{ // determine the timing of the computer if we haven't already done so
		long unsigned iterations = 10000000;
		clock_t start, finish;
		do
		{
		start = clock();	// get starting time
		assert(start>=0);
		for(long i = iterations; i>0L; i--)
			;	// do a whole bunch of empty loops
		finish = clock();	// get ending time
		assert(finish>0);
		iterations *= 2;	// double iterations for next try if needed
		} while ((finish-start)<10);	// keep going until we have enough precision
		// compute the number of empty loops per millisecond
		loopsPerMillisecond = ((iterations/(2*1000))*CLOCKS_PER_SEC)/(finish-start);
		assert(loopsPerMillisecond>1000);	// we need enough precision
#if DEBUG
		cerr << "finish = " << finish << "   start = " << start << "\n";
		cerr << "loops per ms = " << loopsPerMillisecond << "\n";
		cerr.flush();
#endif
	}
	
	if(time_units==MICROSECONDS && d>10000)
	{
		time_units = MILLISECONDS;
		d /= 1000;
	}
	if(time_units == MILLISECONDS)
	{	// doing millisecond timing here
		long s = clock();
		assert(s>0);	// clock() returns -1 if there is an error
		long t = s + (d * CLOCKS_PER_SEC)/ 1000;
		if(t==s) t++;	// must wait at least 1 clock tick
		while(clock()<t)
			;
	}
	else
	{	// doing microsecond timing here
		for(long loops = (loopsPerMillisecond*d)/1000; loops>0; loops--)
			;
	}
}

#pragma optimize( "", on )


static bool batch = false;	// stops prompts to user when true

// subroutine for enabling/disabling prompts to user
void EnableBatch(bool b)
{
	batch = b;	// disable user prompts (i.e. enable batch processing) when true
}

// subroutine for prompting the user for a response
int PromptUser(string& msg, int action)
{
	if(batch)
		return RESPONSE_CONTINUE;

#ifdef _WINDOWS
	int response;
	switch(action)
	{
	case PROMPT_OKCANCEL:
		response = AfxMessageBox(msg.data(),MB_ICONINFORMATION | MB_OKCANCEL);
		return response==IDCANCEL ? RESPONSE_CANCEL : RESPONSE_CONTINUE;
	case PROMPT_OK:
	default:
		AfxMessageBox(msg.data(),MB_ICONINFORMATION);
		return RESPONSE_CONTINUE;
	}
#else
	cerr << msg.data();
	char c;
	switch(action)
	{
	case PROMPT_OKCANCEL:
		cerr << "\nContinue (Y/N)?...";
		c = toupper(getchar());
		fflush(stdin);
		cerr << "\n";
		if(c == 'Y')
			return RESPONSE_CONTINUE;
		else
			return RESPONSE_CANCEL;
	case PROMPT_OK:
	default:
		cerr << "\nPress any key to continue...";
		getchar();
		fflush(stdin);
		cerr << "\n";
		return RESPONSE_CONTINUE;
	}
#endif
}

// get the directory where the XSTOOL utilities are stored
const char* FindXSTOOLSBinDir(void)
{
	XSError err(cerr);
	string msg;

#ifdef _WINDOWS
	// check the registry
	HKEY key;
	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE,"SOFTWARE\\XESS\\XSTOOLS",0,KEY_READ,&key) == ERROR_SUCCESS)
	{
		// get the installation path
		static char path[200];
		DWORD length, valType;
		length = 200;
		if(RegQueryValueEx(key,"InstallDir",0,&valType,(unsigned char*)path,&length) == ERROR_SUCCESS)
		{
			assert(valType==REG_SZ);
			return path;
		}
	}
#endif
	// check the environment variable if the registry didn't have it
	const char* p = getenv("XSTOOLS_BIN_DIR");
	if(p != NULL)
		return p;
	err.SimpleMsg(XSErrorFatal,(string)"XSTOOLS_BIN_DIR environment variable is not set!!\n");
	return NULL;
}


static long begPos, endPos, eofPos;		// beginning and ending file offsets for XSTOOLS parameter record

// set XSTOOLs parameter value
bool SetXSTOOLSParameter(string& name, string& value)
{
	XSError err(cerr);
	string msg;

#ifdef _WINDOWS
	// check the registry for the parameter value and set it if found
	HKEY appKey;
	DWORD keyDisp;
	
	if(RegCreateKeyEx(HKEY_CURRENT_USER,"SOFTWARE\\XESS",0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&appKey,&keyDisp) == ERROR_SUCCESS)
	{
		RegSetValueEx(appKey,name.data(),0,REG_SZ,(const unsigned char*)(value.data()),value.length());
		RegCloseKey(appKey);
	}
#endif
	// set the parameter value in the XS parameter file
	FILE* fp;
	if(GetXSTOOLSParameter(name) != "")
	{ // remove existing parameter record
		string XSTOOLSParameterFilename = (string)FindXSTOOLSBinDir() + (string)"\\XSPARAM.TXT";
		int fh = open(XSTOOLSParameterFilename.data(),_O_RDWR,0);
		if(fh == -1)
		{
			msg = "XSTOOLS parameter file (" + XSTOOLSParameterFilename + (string)") could not be opened!!\n";
			err.SimpleMsg(XSErrorMajor,msg);
			return false;
		}
		// write the tail end of the file starting at the beginning of the record to be removed
		// begPos, endPos and eofPos set by previous call to GetXSTOOLSParameter()
		char* tail = (char*)malloc(eofPos-endPos);
		lseek(fh,endPos,SEEK_SET);
		read(fh,tail,eofPos-endPos);
		lseek(fh,begPos,SEEK_SET);
		write(fh,tail,eofPos-endPos);
		chsize(fh,eofPos-(endPos-begPos));
		close(fh);
	}

	// append name and value to parameter file
	string XSTOOLSParameterFilename = (string)FindXSTOOLSBinDir() + (string)"\\XSPARAM.TXT";
	fp = fopen(XSTOOLSParameterFilename.data(),"a+");
	if(fp == NULL)
	{
		msg = "XSTOOLS parameter file (" + XSTOOLSParameterFilename + (string)") could not be opened!!\n";
		err.SimpleMsg(XSErrorMajor,msg);
		return false;
	}
	fprintf(fp,"%s %s\n",name.data(),value.data());
	fclose(fp);
	return true;
}

// get XSTOOLs parameter value
string GetXSTOOLSParameter(string& name)
{
	XSError err(cerr);
	string msg;
	char regValString[512];
	DWORD length = 0;

#ifdef _WINDOWS
	// check the registry for the parameter
	HKEY appKey;
	DWORD keyDisp;
	DWORD valType;
	if(RegCreateKeyEx(HKEY_CURRENT_USER,"SOFTWARE\\XESS",0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&appKey,&keyDisp) == ERROR_SUCCESS)
	{
		length = 512;
		if(RegQueryValueEx(appKey,name.data(),0,&valType,(unsigned char*)regValString,&length) == ERROR_SUCCESS)
		{
			assert(valType==REG_SZ);
			regValString[length] = 0;	// terminate string read from registry
		}
		else
			length = 0;
		RegCloseKey(appKey);
	}
#endif
	// check the XS parameter file
	string XSTOOLSParameterFilename = (string)FindXSTOOLSBinDir() + (string)"\\XSPARAM.TXT";
	FILE* fp = fopen(XSTOOLSParameterFilename.data(),"a+");		// create file if it doesn't exist
	if(fp == NULL)
	{
		msg = "XSTOOLS parameter file (" + XSTOOLSParameterFilename + (string)") could not be opened!!\n";
		err.SimpleMsg(XSErrorMajor,msg);
		begPos = endPos = eofPos = 0;
		if(length==0) return (string)"";
		else return regValString;
	}
	fseek(fp,0,SEEK_END);
	eofPos = ftell(fp);
	fseek(fp,0,SEEK_SET);	// go to beginning of file
	char line[512];
	while(!feof(fp))
	{
		begPos = ftell(fp);		// store position of beginning of record
		if(fgets(line,511,fp) == NULL)
		{ // no more text lines in the file
			if(feof(fp))
			{ // end of file
				fclose(fp);
				begPos = endPos = 0;
				if(length==0) return (string)"";
				else return regValString;
			}

			// otherwise, some other error terminated reading of the file
			msg = (string)"File error encountered while reading XSTOOLS parameter file (" + XSTOOLSParameterFilename + (string)")\n";
			err.SimpleMsg(XSErrorMajor,msg);
			fclose(fp);
			begPos = endPos = 0;
			if(length==0) return (string)"";
			else return regValString;
		}
		endPos = ftell(fp);		// store position of beginning of next record

		// got a line of text.  now see if it starts with the name we are searching for...
		char n[512], fileValString[512];
		int numStrings = sscanf(line,"%s %s",n,fileValString);
		if(numStrings==0  || numStrings==EOF)
			continue;	// skip blank lines
		if(sscanf(line,"%s %s",n,fileValString) != 2)
		{
			msg = (string)"Corrupted record encountered while reading XSTOOLS parameter file (" + XSTOOLSParameterFilename + (string)"):\n";
			msg += (string)line;
			err.SimpleMsg(XSErrorMajor,msg);
			fclose(fp);
			begPos = endPos = 0;
			if(length==0) return (string)"";
			else return regValString;
		}
		if((string)n == name)
		{ // found the parameter name.  now return the parameter value...
			fclose(fp);
			if(length==0) return (string)fileValString;
			else return regValString;
		}
		// otherwise keep searching through the parameter text file...
	}
	fclose(fp);
	begPos = endPos = 0;
	if(length==0) return (string)"";
	else return regValString;
}

// get the information about the XS Boards from a file
int GetXSBoardInfo(XSBoardInfo** bInfo)
{
	XSError err(cerr);
	static int numBoards=0;
	static XSBoardInfo* brdInfo=NULL;

	// return board info already read from file on a previous call to this subroutine
	if(numBoards > 0)
	{
		*bInfo = brdInfo;
		return numBoards;
	}

	// otherwise, read board info from file and store it in data structure
	const char* whitespace = " \t\n\f\r";
	string msg;
	string XSBoardInfoFilename = (string)FindXSTOOLSBinDir() + (string)"\\XSBRDINF.TXT";
	FILE* fp = fopen(XSBoardInfoFilename.data(),"r");
	if(fp == NULL)
	{
		msg = "XS Board information file (" + XSBoardInfoFilename + (string)") could not be opened!!\n";
		err.SimpleMsg(XSErrorMajor,msg);
		brdInfo = NULL;
		numBoards = 0;
		*bInfo = brdInfo;
		return numBoards;
	}

	char *line, *field;
	while(!feof(fp))
	{
		if((line = (char*)malloc(512)) == NULL)
		{
			msg = (string)"Ran out of memory while reading entries from the XS Board information file (" + XSBoardInfoFilename + (string)")\n";
			err.SimpleMsg(XSErrorMajor,msg);
			numBoards = 0;
			if(brdInfo!=NULL)
				free(brdInfo);
			brdInfo = NULL;
			fclose(fp);
			*bInfo = brdInfo;
			return numBoards;
		}
		if(fgets(line,512,fp) == NULL)
		{
			if(feof(fp))
			{
				free(line);
				fclose(fp);
				*bInfo = brdInfo;
				return numBoards;
			}
			msg = (string)"File error encountered while reading XS Board information file (" + XSBoardInfoFilename + (string)")\n";
			err.SimpleMsg(XSErrorMajor,msg);
			numBoards = 0;
			if(brdInfo!=NULL)
				free(brdInfo);
			brdInfo = NULL;
			free(line);
			fclose(fp);
			*bInfo = brdInfo;
			return numBoards;
		}

		if((field = strtok(line,whitespace)) == NULL)
		{
			free(line);
			continue;	// skip blank lines
		}
		if(*field == '#')
		{
			free(line);
			continue;	// skip comments
		}

		if((brdInfo = (XSBoardInfo*)realloc(brdInfo, (numBoards+1) * sizeof(XSBoardInfo))) == NULL)
		{
			msg = (string)"Ran out of memory while reading records from XS Board information file (" + XSBoardInfoFilename + (string)")\n";
			err.SimpleMsg(XSErrorMajor,msg);
			numBoards = 0;
			if(brdInfo!=NULL)
				free(brdInfo);
			brdInfo = NULL;
			free(line);
			fclose(fp);
			*bInfo = brdInfo;
			return numBoards;
		}
		brdInfo[numBoards].brdModel = field;

		if((field = strtok(NULL,whitespace)) != NULL)
		{
			if(!strcmp(field,"NULL"))
				field[0] = 0;	// null any unused fields
			brdInfo[numBoards].chipType = field;
		}
		if((field = strtok(NULL,whitespace)) != NULL)
		{
			if(!strcmp(field,"NULL"))
				field[0] = 0;	// null any unused fields
			brdInfo[numBoards].dwnldIntfcBitstreamFile = field;
		}
		if((field = strtok(NULL,whitespace)) != NULL)
		{
			if(!strcmp(field,"NULL"))
				field[0] = 0;	// null any unused fields
			brdInfo[numBoards].ramBitstreamFile = field;
		}
		if((field = strtok(NULL,whitespace)) != NULL)
		{
			if(!strcmp(field,"NULL"))
				field[0] = 0;	// null any unused fields
			brdInfo[numBoards].oscBitstreamFile = field;
		}
		if((field = strtok(NULL,whitespace)) != NULL)
		{
			if(!strcmp(field,"NULL"))
				field[0] = 0;	// null any unused fields
			brdInfo[numBoards].eraseBitstreamFile = field;
		}
		if((field = strtok(NULL,whitespace)) != NULL)
		{
			if(!strcmp(field,"NULL"))
				field[0] = 0;	// null any unused fields
			brdInfo[numBoards].flashBitstreamFile = field;
		}
		if((field = strtok(NULL,whitespace)) != NULL)
		{
			if(!strcmp(field,"NULL"))
				field[0] = 0;	// null any unused fields
			brdInfo[numBoards].flashConfigBitstreamFile = field;
		}
		if((field = strtok(NULL,whitespace)) != NULL)
		{
			if(!strcmp(field,"NULL"))
				field[0] = 0;	// null any unused fields
			brdInfo[numBoards].testIntfcBitstreamFile = field;
		}
		if((field = strtok(NULL,whitespace)) != NULL)
		{
			if(!strcmp(field,"NULL"))
				field[0] = 0;	// null any unused fields
			brdInfo[numBoards].testBitstreamFile = field;
		}
		if((field = strtok(NULL,whitespace)) != NULL)
		{
			if(!strcmp(field,"NULL"))
				field[0] = 0;	// null any unused fields
			brdInfo[numBoards].testObjFile = field;
		}
		else
		{	// if the last field was not read correctly, then an error occurred in that field
			// or one of the earlier fields
			msg = (string)"File error encountered while reading fields of XS Board information file (" + XSBoardInfoFilename + (string)")\n";
			err.SimpleMsg(XSErrorMajor,msg);
			numBoards = 0;
			if(brdInfo!=NULL)
				free(brdInfo);
			brdInfo = NULL;
			free(line);
			fclose(fp);
			*bInfo = brdInfo;
			return numBoards;
		}

		numBoards++;
	}

	fclose(fp);
	*bInfo = brdInfo;
	return numBoards;
}
