/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// XC4KPort objects specialize a CnfgPort to handle programming of XC4000 FPGAs
//

// $Log: xc4kprt.cpp $
// Revision 1.1  2001/12/30 15:49:25  DEVB
// Initial revision
//


#include <dos.h>
#include <stdlib.h>
#include <stdio.h>
#include <fstream.h>
#include <strstrea.h>
#include <time.h>
#include <string.h>

#include <afxwin.h>

#include "Utils.h"
#include "XC4KPrt.h"

/* function: GetInteger */
// compute an integer from a vector of hex numbers in a stream
static long unsigned int GetInteger(istream& is, int len=2);

/* function: GetType */
// get the field type indicator from a stream
static int GetType(istream& is);

/* function: ScanForField */
// pass over fields until a certain field is found
static bool ScanForField(istream& is, unsigned char searchType);



// begin methods and functions



// Instantiate an XC4KPort object on a given parallel port
XC4KPort::XC4KPort(void)
{
	progressGauge = NULL;
}

// Instantiate an XC4KPort object on a given parallel port
XC4KPort::XC4KPort(XSError* e, // pointer to error reporting object
				   unsigned int portNum, // parallel port number
				   unsigned int invMask, // inversion mask for the parallel port
				   unsigned int posCCLK, // bit position in parallel port of CCLK pin
				   unsigned int posPROG, // bit position in parallel port of PROGRAM pin
				   unsigned int posDIN, // bit position in parallel port of DIN pin
				   unsigned int posDONE // bit position in parallel port of DONE pin
				   )
{
	progressGauge = NULL;
	Setup(e,portNum,invMask,posCCLK,posPROG,posDIN,posDONE);
}



// Setup an XC4KPort object on a given parallel port
bool XC4KPort::Setup(XSError* e, // pointer to error reporting object
				   unsigned int portNum, // parallel port number
				   unsigned int invMask, // inversion mask for the parallel port
				   unsigned int posCCLK, // bit position in parallel port of CCLK pin
				   unsigned int posPROG, // bit position in parallel port of PROGRAM pin
				   unsigned int posDIN, // bit position in parallel port of DIN pin
				   unsigned int posDONE // bit position in parallel port of DONE pin
				   )
{
	return CnfgPort::Setup(e,portNum,invMask,posCCLK,posPROG,posDIN,posDONE);
}




static const int DeviceFieldType = 0x62;		// field type for the FPGA device identifier

// returns the chip identifier from the bitstream
string XC4KPort::GetChipType(istream& is)
{
	XSError& err = GetErr();
	
	if(is.eof()!=0)
		return (string)""; // exit if no BIT stream exists
	
	// jump over the first field of the BIT file
	long unsigned int fieldLength = GetInteger(is);
	assert(fieldLength!=0);
	is.ignore(fieldLength);
	
	// process the second field
	fieldLength = GetInteger(is);
	assert(fieldLength==1);
	
	// now look for the field with the chip identifier
	bool status = ScanForField(is,DeviceFieldType);
	assert(status==true);
	fieldLength = GetInteger(is);
	assert(fieldLength>0 && fieldLength<100);
	char chipType[100];
	for(unsigned int i=0; i<fieldLength; i++)
		is >> chipType[i];
	chipType[i] = 0;		// terminate string
	return chipType;		// return the chip type
}



// returns the chip identifier from the bitstream
string XC4KPort::GetChipType(string& bitfileName)
{
	XSError& err = GetErr();
	
	if(bitfileName.compare("")==0)
		return (string)""; // exit if no BIT file was given
	
	ifstream is(bitfileName.data(),ios::binary|ios::nocreate);
	if(is.fail()!=0)
	{ // error if given file could not be opened
		err.SetSeverity(XSErrorMajor);
		err << "could not open " << bitfileName.data() << "\n";
		err.EndMsg();
		return (string)"";
	}
	
	// now get the chip identifier from the bitstream file
	string chipType = GetChipType(is);
	
	is.close();
	
	return chipType;
}



// Initialize FPGA configuration pins
void XC4KPort::InitConfigureFPGA(void)
{
	SetPROG(1);
	SetCCLK(0);
	Out(0x3,6,7);	// set XC4000 mode pins to "slave serial" config. mode
}



// Send out a byte of configuration information
void XC4KPort::ConfigureFPGA(unsigned char b)
{
	// configuration bytes for the XC4000 programming pins only contain
	// values for the DIN pin, so eight clock pulses are needed per byte.
	for(int i=0; i<8; i++)
	{
		SetDIN(b&0x80 ? 1:0);  // byte is sent MSB first
		b <<= 1;
		PulseCCLK();
	}
}



// compute an integer from a vector of hex numbers in a stream
static long unsigned int GetInteger(istream& is, int len)
{
	assert(len<=4); // can't handle lengths greater than 32 bits
	long unsigned int lui = 0;
	for(int i=0; i<len; i++)
	{ // process hex digits starting from MSDigit
		unsigned char digit;
		is.read(&digit,1);
		assert(is.eof()==0);
		lui = (lui<<8) | (long unsigned int)digit;
	}
	return lui;
}



// get the field type indicator from a stream
static int GetType(istream& is)
{
	unsigned char type;
	is.read(&type,1);
	assert(is.eof()==0);
	return type;
}



// pass over fields until a certain field is found
static bool ScanForField(istream& is, unsigned char searchType)
{
	while(true)
	{
		unsigned char type = GetType(is);
		if(type==searchType)
			return true;
		unsigned int length = GetInteger(is);
		is.ignore(length);
		if(is.eof()!=0)
			break;
	}
	return false;
}



static const int BitstreamFieldType = 0x65;	// field type for the bitstream data

// process a stream and send the configuration bitstream to the board
bool XC4KPort::ConfigureFPGA(istream& is)
{
	XSError& err = GetErr();
	
	if(is.eof()!=0)
		return false; // exit if no BIT stream exists

	assert(progressGauge != NULL);	// make sure progress indicator is initialized

	// show the progress indicator
	progressGauge->Report(0);
	
	// jump over the first field of the BIT file
	long unsigned int fieldLength = GetInteger(is);
	assert(fieldLength!=0);
	is.ignore(fieldLength);
	
	// process the second field
	fieldLength = GetInteger(is);
	assert(fieldLength==1);
	
	// now look for the field with the bitstream data
	bool status = ScanForField(is,BitstreamFieldType);
	assert(status==true);
	
	// now we are at the start of the configuration bits
	fieldLength = GetInteger(is,4); // get the length of the bit stream
	assert(fieldLength>0);
#if DEBUG
	cerr << "bit stream field length = " << fieldLength << "\n";
#endif

	// start the configuration process
	PulsePROG();					// pulse the PROGRAM pin to start the configuration of the FPGA
	InsertDelay(300,MILLISECONDS);	// insert 300 ms delay after PROGRAM pin goes high

	for(long unsigned int i=0; i<fieldLength; i++)
	{ // read the bytes and send the config. bits to the board
		unsigned char byte;
		is.read(&byte,1);
		assert(is.eof()==0);	// should not hit end-of-file
		ConfigureFPGA(byte);
		// output some feedback as the configuration proceeds
		if((i%1000)==0)
			progressGauge->Report(is.tellg());
	}
	progressGauge->Report(is.tellg());	// this should set gauge to 100%
	
	// clock a few more times at the end of the configuration process
	// to clean things up...
	for(i=0; i<4; i++)
		PulseCCLK();

	// insert a delay to let the chip initialize
	InsertDelay(100,MILLISECONDS);
	
#if DEBUG
	cerr << "exiting ConfigureFPGA()\n";
#endif
	
	return true;
}



// configure the FPGA with the bit stream stored in a BIT file
bool XC4KPort::ConfigureFPGA(string& fileName)
{
	XSError& err = GetErr();
	
	if(fileName.length()==0)
		return false;  // stop if no SVF or BIT file was given
	
	ifstream is(fileName.data(),ios::binary|ios::nocreate);  // otherwise open BIT or SVF file
	if(is.fail() || is.eof())
	{ // error - couldn't open file
		err.SetSeverity(XSErrorMajor);
		err << "could not open " << fileName.data() << "\n";
		err.EndMsg();
		return false;
	}
	
	// determine the size of the file
	is.seekg(0,ios::end);	// move pointer to end of file
	streampos streamEndPos = is.tellg();	// pointer position = position of end of file

	is.seekg(0,ios::beg);	// return pointer to beginning of file

	progressGauge = new Progress(&err,(string)"Configure FPGA",(string)"Downloading "+StripPrefix(fileName),0,streamEndPos);
	
	bool status = ConfigureFPGA(is);
	
	delete progressGauge;
	
	is.close();  // close-up the hex file
	
	return status;
}
