/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The XC95KPort object specializes the JTAGPort object so that an XC9500 CPLD
// can be configured.
//

// $Log: xc95kprt.cpp $
// Revision 1.1  2001/12/30 15:51:28  DEVB
// Initial revision
//

#include <string.h>
#include <fstream.h>
#include <time.h>
#include "Utils.h"
#include "Hex.h"
#include "Bitstrm.h"
#include "XC95KPrt.h"
using std::string;



// begin methods and functions



// Instantiate an object for an XC9500 CPLD connected to a parallel port
XC95KPort::XC95KPort(void)
{
	progressGauge = NULL;
}

// Instantiate an object for an XC9500 CPLD connected to a parallel port
XC95KPort::XC95KPort(XSError* err,	// pointer to error reporting object
		   unsigned int portNum,	// parallel port number
		   unsigned int invMask,	// inversion mask for the parallel port
		   unsigned int posTCK,	// bit position in parallel port of TCK pin
		   unsigned int posTMS,	// bit position in parallel port of TMS pin
		   unsigned int posTDI,	// bit position in parallel port of TDI pin
		   unsigned int posTDO)	// bit position in parallel port of TDO pin
{
	progressGauge = NULL;
	Setup(err,portNum,invMask,posTCK,posTMS,posTDI,posTDO);
}



// Setup an object for an XC9500 CPLD connected to a parallel port
bool XC95KPort::Setup(XSError* err,	// pointer to error reporting object
		   unsigned int portNum,	// parallel port number
		   unsigned int invMask,	// inversion mask for the parallel port
		   unsigned int posTCK,	// bit position in parallel port of TCK pin
		   unsigned int posTMS,	// bit position in parallel port of TMS pin
		   unsigned int posTDI,	// bit position in parallel port of TDI pin
		   unsigned int posTDO)	// bit position in parallel port of TDO pin
{
	// return true if setup was successful
	return JTAGPort::Setup(err,portNum,invMask,posTCK,posTMS,posTDI,posTDO);
}



// detect the presence and type of the XC9500 CPLD chip on the port
string XC95KPort::GetChipID()
{
	Bitstream bsir(8);	// BSIR for sending IDCODE instruction
	Bitstream bsdr(32);	// BSDR for holding chip ID
	bsir.SetBits(0, 0,1,1,1,1,1,1,1, -1); // IDCODE instruction for reading chip id
	Bitstream null(0);	// zero-length bitstream

	InitTAP();							// initialize the JTAG TAP controller
	LoadBSIRthenBSDR(bsir,null,bsdr);	// get the ID code from the chip
	return bsdr.ToString();
}

// get the 32-bit USERCODE from the XC9500 CPLD chip
string XC95KPort::GetUSERCODE()
{
	Bitstream bsir(8);	// BSIR for sending USERCODE instruction
	Bitstream bsdr(32);	// BSDR for holding USERCODE
	bsir.SetBits(0, 1,0,1,1,1,1,1,1, -1); // USERCODE instruction for reading USERCODE
	Bitstream null(0);	// zero-length bitstream

	InitTAP();							// initialize the JTAG TAP controller
	LoadBSIRthenBSDR(bsir,null,bsdr);	// get the ID code from the chip
//	PromptUser((string)"USERCODE = " + bsdr.ToString() + (string)"\n",PROMPT_OK);
	return bsdr.ToString();
}


#if 0
	// identifiers for the various types of XC9500 CPLD chips in 84-pin PLCC
	Bitstream XC9572ID(28), XC9572ID1(32);
	Bitstream XC9572XLID(28), XC9572XLID1(32);
	Bitstream XC95108ID(28), XC95108ID1(32), XC95108ID2(32), XC95108ID3(32);
	XC9572ID.SetBits ( 0, 1,1,0,0,1,0,0,1, 0,0,0,0,0,0,1,0,
		0,0,0,0,1,0,1,0, 1,0,0,1, -1);
	XC9572ID1.SetBits ( 0, 1,1,0,0,1,0,0,1, 0,0,0,0,0,0,1,0,
		0,0,0,0,1,0,1,0, 1,0,0,1,0,0,0,0, -1);
	XC9572XLID.SetBits ( 0, 1,1,0,0,1,0,0,1, 0,0,0,0,0,0,1,0,
		0,0,0,0,0,1,1,0, 1,0,0,1, -1);
	XC9572XLID1.SetBits ( 0, 1,1,0,0,1,0,0,1, 0,0,0,0,0,0,1,0,
		0,0,0,0,0,1,1,0, 1,0,0,1,0,0,0,0, -1);
	XC95108ID.SetBits (0, 1,1,0,0,1,0,0,1, 0,0,0,0,0,1,1,0,
		0,0,0,0,1,0,1,0, 1,0,0,1, -1);
	XC95108ID1.SetBits(0, 1,1,0,0,1,0,0,1, 0,0,0,0,0,1,1,0,
		0,0,0,0,1,0,1,0, 1,0,0,1,0,0,0,0, -1);
	XC95108ID2.SetBits(0, 1,1,0,0,1,0,0,1, 0,0,0,0,0,1,1,0,
		0,0,0,0,1,0,1,0, 1,0,0,1,1,0,0,0, -1);
	XC95108ID3.SetBits(0, 1,1,0,0,1,0,0,1, 0,0,0,0,0,1,1,0,
		0,0,0,0,1,0,1,0, 1,0,0,1,0,1,0,0, -1);
#endif


// Initialize the XC9500 CPLD for configuration through the JTAG interface
void XC95KPort::InitConfigureCPLD(void)
{
	InitTAP();	// initialize TAP controller of XC95KPort chip to reset state
}



// Send out a byte of configuration information to the XC9500 Flash
void XC95KPort::ConfigureCPLD(unsigned char b)
{
	// JTAG configuration bytes contain values for both TDI and TMS,
	// so only four clock pulses are needed per byte
	for(int i=0; i<4; i++)
	{
		SetTDI(b&1);	// byte is sent out LSB first
		b >>= 1;
		SetTMS(b&1);
		b >>= 1;
		PulseTCK();
	}
}



static string NextWord(string& s)
{
	int pos = s.find_first_not_of(" \t");	// find beginning of word
	if(pos== -1)
	{	// no non-whitespace chars found, so string has no words
		s = "";
		return "";
	}
	s = s.substr(pos,s.length()-pos);	// remove leading whitespace
	pos = s.find_first_of(" \t");		// find end of first keyword
	string word;
	if(pos== -1)
	{	// no more whitespace, so entire string is the last word
		word = s;
		s = "";
	}
	else
	{	// extract the word from the string
		word = s.substr(0,pos);	// get first keyword from command
		s = s.substr(pos,s.length()-pos);	// remove keyword from command
	}
	return word;
}



// Process SVF and send results through JTAG configuration pins
bool XC95KPort::ConfigureCPLD(istream& is)
{
	InitTAP();	// initialize the JTAG state machine
	GoThruTAPSequence(RunTestIdle,-1); // home state for configuration

	unsigned int seconds = clock()/CLOCKS_PER_SEC;

#if DEBUG_SVF
	long start = clock();
#endif
	
	int lineCnt = 0;
	unsigned long elapsedTime = 0;
	progressGauge->Report(0);
	while(true)
	{
		const unsigned int maxLineLength = 400;
		char line[maxLineLength];
		line[0] = 0;
		is.getline(line,400);
		if(is.fail())
			break;
		lineCnt++;
		
		// indicate progress to user
		if((unsigned int)(clock()/CLOCKS_PER_SEC)!=seconds)
		{
			seconds = clock()/CLOCKS_PER_SEC;
			progressGauge->Report(is.tellg());
		}
		
		string cmd = line;	// store SVF command in a string variable
		
		// remove anything after the first ";" or "//" in the command
		unsigned int pos = cmd.find_first_of(';');
		cmd = cmd.substr(0,pos);
		pos = cmd.find("//");
		cmd = cmd.substr(0,pos);
		
#if DEBUG_SVF
		cout << cmd.data() << "\n"; cout.flush();
#endif
		
		if(cmd.length()==0 || cmd.compare("")==0)
			continue;	// must be a comment line
		
		// remove the first keyword from the command
		string keywd = NextWord(cmd);
		if(keywd == "")
			continue;	// must be a blank line

		int status;
		unsigned long len;
		static Bitstream hdr(0), hir(0), tdr(0), tir(0);
		
		if(keywd.compare("TRST")==0)
		{ // TRST command has nothing to do with configuration so ignore it
#if DEBUG_SVF
			cout << "TRST\n"; cout.flush();
#endif
			continue;
		}
		
		else if(keywd.compare("STATE")==0)
		{ // STATE command has nothing to do with configuration so ignore it
#if DEBUG_SVF
			cout << "STATE\n"; cout.flush();
#endif
			continue;
		}
		
		else if(keywd.compare("ENDIR")==0)
		{ // ENDIR has nothing to do with configuration, but check to make
			// sure it is a well-formed command
#if DEBUG_SVF
			cout << "ENDIR\n"; cout.flush();
#endif
			keywd = NextWord(cmd);
			assert(keywd.compare("IDLE")==0);	// make sure it's what should be there
			continue;
		}
		
		else if(keywd.compare("ENDDR")==0)
		{ // ENDDR has nothing to do with configuration, but check to make
			// sure it is a well-formed command
#if DEBUG_SVF
			cout << "ENDDR\n"; cout.flush();
#endif
			keywd = NextWord(cmd);
			assert(keywd.compare("IDLE")==0);	// make sure it's what should be there
			continue;
		}
		
		else if(keywd.compare("HIR")==0)
		{ // initialize the header bits for loading into the IR
#if DEBUG_SVF
			cout << "HIR\n"; cout.flush();
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TDI")==0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word[0]=='(');
			assert(word[word.length()-1]==')');
			Bitstream bs(len);
			unsigned int bsi = 0; // bit stream index
			// process hex number within parentheses
			for(int i=word.length()-2; i>0; i--)
			{
				int hexbits = CharToHex(word[i]);
				for(int j=0; j<4 && bsi<len; j++,bsi++,hexbits>>=1)
					bs.SetBit(bsi,hexbits&1);
			}
			hir = bs; // copy bitstream into IR header bitstream
			continue;
		}
		
		else if(keywd.compare("HDR")==0)
		{ // initialize the header bits for loading into the DR
#if DEBUG_SVF
			cout << "HDR\n"; cout.flush();
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TDI")==0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word[0]=='(');
			assert(word[word.length()-1]==')');
			Bitstream bs(len);
			unsigned int bsi = 0; // bit stream index
			// process hex number within parentheses
			for(int i=word.length()-2; i>0; i--)
			{
				int hexbits = CharToHex(word[i]);
				for(int j=0; j<4 && bsi<len; j++,bsi++,hexbits>>=1)
					bs.SetBit(bsi,hexbits&1);
			}
			hdr = bs; // copy bitstream into IR header bitstream
			continue;
		}
		
		else if(keywd.compare("TIR")==0)
		{ // initialize the trailer bits for loading into the IR
#if DEBUG_SVF
			cout << "TIR\n"; cout.flush();
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TDI")==0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word[0]=='(');
			assert(word[word.length()-1]==')');
			Bitstream bs(len);
			unsigned int bsi = 0; // bit stream index
			// process hex number within parentheses
			for(int i=word.length()-2; i>0; i--)
			{
				int hexbits = CharToHex(word[i]);
				for(int j=0; j<4 && bsi<len; j++,bsi++,hexbits>>=1)
					bs.SetBit(bsi,hexbits&1);
			}
			tir = bs; // copy bitstream into IR header bitstream
			continue;
		}
		
		else if(keywd.compare("TDR")==0)
		{ // initialize the trailer bits for loading into the DR
#if DEBUG_SVF
			cout << "TDR\n"; cout.flush();
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TDI")==0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word[0]=='(');
			assert(word[word.length()-1]==')');
			Bitstream bs(len);
			unsigned int bsi = 0; // bit stream index
			// process hex number within parentheses
			for(int i=word.length()-2; i>0; i--)
			{
				int hexbits = CharToHex(word[i]);
				for(int j=0; j<4 && bsi<len; j++,bsi++,hexbits>>=1)
					bs.SetBit(bsi,hexbits&1);
			}
			tdr = bs; // copy bitstream into IR header bitstream
			continue;
		}
		
		else if(keywd.compare("SIR")==0)
		{ // send a bitstream into the instruction register
#if DEBUG_SVF
			cout << "SIR\n"; cout.flush();
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TDI")==0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word[0]=='(');
			assert(word[word.length()-1]==')');
			// create a bitstream
			Bitstream bs(len);
			unsigned int bsi=0;
			for(int i=word.length()-2; i>0; i--)
			{
				int hexbits = CharToHex(word[i]);
				for(int j=0; j<4 && bsi<len; j++,bsi++,hexbits>>=1)
					bs.SetBit(bsi,hexbits&1);
			}
			// concatenate header, data, and trailer bitstreams
			bs = hir + bs + tir;
#if DEBUG_SVF
			cout << bs << "\n"; cout.flush();
#endif
			// send bitstream into IR and return to Run-Test-Idle state
			GoThruTAPSequence(SelectDRScan,SelectIRScan,CaptureIR,ShiftIR,-1);
			SendRcvBitstream(bs,bs);
			GoThruTAPSequence(PauseIR,Exit2IR,UpdateIR,RunTestIdle,-1);
			continue;
		}
		
		else if(keywd.compare("SDR")==0)
		{ // send a bitstream into the data register
#if DEBUG_SVF
			cout << "SDR\n"; cout.flush();
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TDI")==0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word[0]=='(');
			assert(word[word.length()-1]==')');
			// create a bitstream
			Bitstream bs(len);
			unsigned int bsi=0;
			for(int i=word.length()-2; i>0; i--)
			{
				int hexbits = CharToHex(word[i]);
				for(int j=0; j<4 && bsi<len; j++,bsi++,hexbits>>=1)
					bs.SetBit(bsi,hexbits&1);
			}
			// concatenate header, data, and trailer bitstreams
			bs = hdr + bs + tdr;
#if DEBUG_SVF
			cout << bs << "\n"; cout.flush();
#endif
			// send bitstream into IR and return to Run-Test-Idle state
			GoThruTAPSequence(SelectDRScan,CaptureDR,ShiftDR,-1);
			SendRcvBitstream(bs,bs);
			GoThruTAPSequence(PauseDR,Exit2DR,UpdateDR,RunTestIdle,-1);
			continue;
		}
		
		else if(keywd.compare("RUNTEST")==0)
		{ // insert delay while programming takes place
#if DEBUG_SVF
			cout << "RUNTEST\n";
#endif
			string word = NextWord(cmd);
			assert(word.length()>0);
			status = sscanf(word.data(),"%ld",&len);
			elapsedTime += len;
			assert(status!=0);
			word = NextWord(cmd);
			assert(word.length()>0);
			assert(word.compare("TCK")==0);
#if DEBUG_SVF
			cout << "delay = " << len << " us\n"; cout.flush();
#endif
			InsertDelay(len,MICROSECONDS);
			continue;
		}
		
		// if we get here, then it's some strange junk we don't understand
		XSError& err = GetErr();
		err.SetSeverity(XSErrorMajor);
		err << "unknown SVF command: " << cmd.data() << "\n";
		err.EndMsg();
		InitTAP(); // reset TAP and idle the JTAG logic
		return false;
	}
	
	progressGauge->Report(is.tellg());	// should set gauge to 100%
	
#if DEBUG_SVF
	long finish = clock();
	cerr << "elapsed time = " << (finish-start)/(CLOCKS_PER_SEC) << "\n";
	cerr << "elapsed time = " << elapsedTime << " us\n";
#endif
	
	InitTAP(); // done so reset TAP and idle the JTAG logic
	InsertDelay(100,MILLISECONDS); // let the XC9500 chip initialize
	return true;
}



// configure the CPLD with the bit stream stored in an SVF file
bool XC95KPort::ConfigureCPLD(string& fileName)
{
	XSError& err = GetErr();
	
	if(fileName.length()==0)
		return false;  // stop if no SVF file was given
	
	ifstream is(fileName.data(),ios::binary|ios::nocreate);  // otherwise open SVF file
	if(is.fail() || is.eof())
	{ // error - couldn't open file
		err.SetSeverity(XSErrorMajor);
		err << "could not open " << fileName.data() << "\n";
		err.EndMsg();
		return false;
	}
	
	// determine the size of the file
	is.seekg(0,ios::end);	// move pointer to end of file
	streampos streamEndPos = is.tellg();	// pointer position = position of end of file

	is.seekg(0,ios::beg);	// return pointer to beginning of file

	progressGauge = new Progress(&err,(string)"Configure CPLD",(string)"Downloading "+StripPrefix(fileName),0,streamEndPos);
	
	bool status = ConfigureCPLD(is);
	
	delete progressGauge;
	
	is.close();  // close-up the hex file
	
	return status;
}
