/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The XCBSDR object specializes a bitstream into a boundary scan data register
// that is used to perform RAM operations via the JTAG port.

// $Log: xcbsdr.cpp $
// Revision 1.2  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:05:05  DEVB
// Initial revision
//

#include "xcbsdr.h"

//The bsdr is partitioned into three-bit segments.  Each three bit segment
// interfaces to a pin of the FPLD. The bit functions are:
//		bit 0: controls the output driver of the pin
//				0 = tristate the output driver
//				1 = enable the driver
//		bit 1: stores the value to be output on the pin
//		bit 2: stores the value read from the pin   
//

static const int tristate = 0;		// disable pin driver
static const int enable = 1;		// enable pin driver
static const int bsdrPinEnable = 0;	// pin tristate control bit position offset
static const int bsdrPinOutput = 1;	// pin output bit position offset
static const int bsdrPinInput  = 2;	// pin input bit position offset


// begin methods and functions

XCBSDR::XCBSDR(unsigned int length,			// length of bsdr
			   unsigned int* addressBitPos,	// positions of address bits in bsdr
			   unsigned int* dataBitPos,	// positions of data bits in bsdr
			   unsigned int* controlBitPos	// positions of control bits in bsdr
			   ) : Bitstream(length)
{
	for(int i=0; i<numRAMAddressBits; i++)
		ramAddressBitPos[i] = addressBitPos[i];
	for(i=0; i<numRAMDataBits; i++)
		ramDataBitPos[i] = dataBitPos[i];
	for(i=0; i<numRAMControlBits; i++)
		ramControlBitPos[i] = controlBitPos[i];
}

// place a RAM address into the bsdr bitstream buffer
void XCBSDR::SetRAMAddress(unsigned int address)
{
	for(int i=0; i<numRAMAddressBits; i++,address>>=1)
	{
		// enable the output drivers of the FPLD pin connected to the RAM address pin
		SetBit(ramAddressBitPos[i]+bsdrPinEnable,enable);
		// set the address bit value on the FPLD pin
		SetBit(ramAddressBitPos[i]+bsdrPinOutput,address&1);
	}
}

// place RAM data into the bsdr bitstream buffer
void XCBSDR::SetRAMData(unsigned int data)
{
	for(int i=0; i<numRAMDataBits; i++,data>>=1)
	{
		// enable the output driver of the FPLD pin connected to the RAM data pin
		SetBit(ramDataBitPos[i]+bsdrPinEnable,enable);
		// set the data bit value on the FPLD pin
		SetBit(ramDataBitPos[i]+bsdrPinOutput,data&1);
	}
}

// tristate the FPLD outputs connected to the RAM data bus
void XCBSDR::ReadRAMData(void)
{
	for(int i=0; i<numRAMDataBits; i++)
	{
		// tristate the FPLD pin that drives the RAM data line
		SetBit(ramDataBitPos[i]+bsdrPinEnable,tristate);
	}
}

// get the RAM data value from the bsdr bitstream
unsigned int XCBSDR::GetRAMData()
{
	unsigned int data = 0;
	for(int i=numRAMDataBits-1; i>=0; i--)
		data = (data<<1) | ((*this)[ramDataBitPos[i]+bsdrPinInput] ? 1:0);
	return data;
}

// get the RAM address value from the bsdr bitstream
unsigned long XCBSDR::GetRAMAddress(void)
{
	unsigned int address = 0;
	for(int i=numRAMAddressBits-1; i>=0; i--)
		address = (address<<1) | ((*this)[ramAddressBitPos[i]+bsdrPinInput] ? 1:0);
	return address;
}

// place RAM control pin levels into the bsdr bitstream buffer
void XCBSDR::SetRAMControls(unsigned int ce_, unsigned int oe_, unsigned int we_)
{
	// set the output drivers of the FPLD pins which drive the RAM control pins
	SetBit(ramControlBitPos[0]+bsdrPinEnable,enable);
	SetBit(ramControlBitPos[1]+bsdrPinEnable,enable);
	SetBit(ramControlBitPos[2]+bsdrPinEnable,enable);
	// set the control signal levels on the FPLD pins
	SetBit(ramControlBitPos[0]+bsdrPinOutput,ce_);
	SetBit(ramControlBitPos[1]+bsdrPinOutput,oe_);
	SetBit(ramControlBitPos[2]+bsdrPinOutput,we_);
}
