/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// This object loads the FPGA or EEPROM with a bitstream,
// uploads/downloads the RAM, sets the oscillator frequency,
// and tests the board.
//

// $Log: xs40brd.cpp $
// Revision 1.1  2001/12/30 15:52:35  DEVB
// Initial revision
//



#include <time.h>
#include <string>
using std::string;
#include "XSError.h"
#include "Utils.h"
#include "XS40Brd.h"


// inversion mask for parallel port connection to XS40 Board
// 00000011: bits  7- 0 are attached to data pins D7-D0
// 00000xxx: bits 15-11 are attached to status pins /S7,S6,S5,S4,S3
// xxxx1001: bits 19-16 are attached to control pins /C3,C2,/C1,/C0
const unsigned int invMask	= 0x090003;

// bit position of XC4000 configuration pins within parallel port registers
const unsigned int posCCLK	= 17;
const unsigned int posPROG	= 16;
const unsigned int posDIN	= 19;
const unsigned int posDONE	= 23;	// not used

// bit position of prog. osc. pins within parallel port registers
const unsigned int posOSC	= 0;

// bit positions of RAM access pins within parallel port
const unsigned int posRESET	= 0;
const unsigned int posCLK	= 1;
const unsigned int posDOLSB	= 2;
const unsigned int posDOMSB	= 5;
const unsigned int posDILSB	= 11;
const unsigned int posDIMSB	= 14;
const unsigned int posSTLSB	= 11;
const unsigned int posSTMSB	= 14;

// bit positions of serial EEPROM access pins within parallel port
const unsigned int posSEECLK	= 17;
const unsigned int posSEEOE		= 6;
const unsigned int posSEECE		= 7;
const unsigned int posSEEDIN	= 19;

// bit positions for board test status
const unsigned int posTESTSTATUS	= 14;
const unsigned int posTESTRESET		= 5;



static XSBoardInfo* brdInfo;
static int numBoards;


// begin methods and functions



// create an object for programming the XS40 Board
XS40Board::XS40Board(void)
{
	XSError* err = new XSError(cerr);	// create error-reporting channel
	numBoards = GetXSBoardInfo(&brdInfo);	// get information about all the models of XS boards
	Setup(err,"",1);
}

// setup the object for programming the XS40 Board
bool XS40Board::Setup(XSError* err,
					 const char* model,		// model of XS40 Board
					 unsigned int lptNum
					 )
{
	// store name of board model
	if(brdModel!=NULL)
		free(brdModel);
	brdModel = (char*)malloc(strlen(model)+1);
	if(brdModel==NULL)
		err->SimpleMsg(XSErrorFatal,(string)"ERRLOC: out of memory!!\n");
	strcpy(brdModel,model);

	bool status;
	status = fpga.Setup(err,lptNum,invMask,posCCLK,posPROG,posDIN,posDONE);
	status = status && osc.Setup(err,lptNum,invMask,posOSC);
	status = status && ram.Setup(err,lptNum,invMask,posRESET,posCLK,posDOLSB,posDOMSB,posDILSB,posDIMSB,posSTLSB,posSTMSB,8,20);
	status = status && seeprom.Setup(err,lptNum,invMask,posSEECLK,posSEECE,posSEEOE,posSEEDIN);
	return status;	// return true if no error occurred for any object setup
}

// load a configuration bitstream into the FPGA or Flash
bool XS40Board::Configure(
			string& fileName	// bitstream file
			)
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	int brdIndex;
	for(brdIndex=0; brdIndex<numBoards; brdIndex++)
	{
		if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
			break;
	}
	if(brdIndex>=numBoards)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XS40 Board!\n");
		return false;
	}
	
	string suffix = GetSuffix(fileName);
	if(suffix!="BIT")
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Only .BIT files can be downloaded into the FPGA on the XS40 Board!!\n");
		return false;
	}

	string type = ConvertToUpperCase(fpga.GetChipType(fileName));
	if(type == "")
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"The .BIT file does not identify the target FPGA!!\n");
		return false;
	}
	if(type != brdInfo[brdIndex].chipType)
	{
		string msg;
		msg = ".BIT file is for the ";
		msg += type;
		msg += " but this is not the FPGA on the ";
		msg += brdModel;
		msg += "!!\n";
		errMsg.SimpleMsg(XSErrorMajor,msg);
		return false;
	}
	fpga.InitConfigureFPGA();
	return fpga.ConfigureFPGA(fileName);
}

// load a configuration bitstream into the interface CPLD
bool XS40Board::ConfigureInterface(
			string& fileName	// bitstream file
			)
{
	return false;	// there is no interface CPLD on the XS40 Board, so always return true
}

// download into RAM
bool XS40Board::DownloadRAM(
			string& fileName,		// RAM data file
			bool doStart,			// perform startup operations for RAM downloads
			bool doEnd				// perform terminating operations for RAM downloads
			)
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	string suffix = GetSuffix(fileName);
	if(suffix!="HEX" && suffix!="EXO" && suffix!="XES" && suffix!="MCS")
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Only .HEX, .MCS, .EXO or .XES files can be downloaded into the RAM on the XS40 Board!!\n");
		return false;
	}

	// get the directory where the RAM interface files are kept
	const char* xstoolsBinDir = FindXSTOOLSBinDir();
	
	int brdIndex;
	for(brdIndex=0; brdIndex<numBoards; brdIndex++)
	{
		if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
			break;
	}
	if(brdIndex>=numBoards)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XS40 Board!\n");
		return false;
	}

	if(strlen(brdInfo[brdIndex].ramBitstreamFile)==0)
	{
		string msg;
		msg = brdModel;
		msg += " does not support RAM downloads!!\n";
		errMsg.SimpleMsg(XSErrorMajor,msg);
		return false;
	}

	string bitFileName;
	bitFileName = xstoolsBinDir;
	bitFileName += "\\";
	bitFileName += brdInfo[brdIndex].ramBitstreamFile;

	if(doStart)
	{
		if(Configure(bitFileName) == false)
		{
			errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading RAM interface circuit!!\n");
			return false;
		}
	}

	if(ram.DownloadRAM(fileName) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading into RAM!!\n");
		return false;
	}

	return true;
}

// upload RAM from loAddr to hiAddr into file
bool XS40Board::UploadRAM(
			string& fileName,		// RAM data file
			const char* format,		// format - Intel hex, Moto exo, or XESS
			unsigned int loAddr,	// lower address of RAM data
			unsigned int hiAddr	// upper address of RAM data
			)
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	// get the directory where the RAM interface files are kept
	const char* xstoolsBinDir = FindXSTOOLSBinDir();
	
	int brdIndex;
	for(brdIndex=0; brdIndex<numBoards; brdIndex++)
	{
		if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
			break;
	}
	if(brdIndex>=numBoards)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XS40 Board!\n");
		return false;
	}

	if(strlen(brdInfo[brdIndex].ramBitstreamFile)==0)
	{
		string msg;
		msg = brdModel;
		msg += " does not support RAM uploads!!\n";
		errMsg.SimpleMsg(XSErrorMajor,msg);
		return false;
	}

	string bitFileName;
	bitFileName = xstoolsBinDir;
	bitFileName += "\\";
	bitFileName += brdInfo[brdIndex].ramBitstreamFile;

	if(Configure(bitFileName) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading RAM interface circuit!!\n");
		return false;
	}

	if(ram.UploadRAM(fileName,format,loAddr,hiAddr) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error uploading from RAM!!\n");
		return false;
	}
	return true;
}

// download into serial EEPROM
bool XS40Board::DownloadFlash(
			string& fileName,		// Flash data file
			bool doStart,			// perform startup operations for Flash downloads
			bool doEnd				// perform terminating operations for Flash downloads
			)
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	string suffix = GetSuffix(fileName);
	if(suffix!="BIT")
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Only .BIT files can be downloaded into the serial EEPROM on the XS40 Board!!\n");
		return false;
	}

	// get the directory where the serial EEPROM interface files are kept
	const char* xstoolsBinDir = FindXSTOOLSBinDir();
	
	int brdIndex;
	for(brdIndex=0; brdIndex<numBoards; brdIndex++)
	{
		if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
			break;
	}
	if(brdIndex>=numBoards)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XS40 Board!\n");
		return false;
	}

	if(strlen(brdInfo[brdIndex].flashBitstreamFile)==0)
	{
		string msg;
		msg = brdModel;
		msg += " does not support serial EEPROM programming!!\n";
		errMsg.SimpleMsg(XSErrorMajor,msg);
		return false;
	}

	string instructions;

	if(doStart)
	{
		instructions = "Before programming the XS40 Board serial EEPROM you must:\n";
		instructions += "1) Place shunts on jumpers J4, J6 and J11\n";
		instructions += "2) Remove any shunt on jumper J10\n";
		instructions += "3) Apply power to the board\n";
		if(PromptUser(instructions,PROMPT_OKCANCEL) == RESPONSE_CANCEL)
			return false;
		
		string bitFileName;
		bitFileName = xstoolsBinDir;
		bitFileName += "\\";
		bitFileName += brdInfo[brdIndex].flashBitstreamFile;
		
		if(Configure(bitFileName) == false)
		{
			errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading serial EEPROM interface circuit!!\n");
			return false;
		}
	}

	if(seeprom.ProgramEEPROM(fileName) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading into serial EEPROM!!\n");
		return false;
	}

	if(doEnd)
	{
		instructions = "The XS40 Board serial EEPROM has been programmed!!\n";
		instructions += "Now do these steps to enable configuration of your XS40 Board from the serial EEPROM:\n";
		instructions += "1) Remove the shunts from jumpers J4, J6 and J11\n";
		instructions += "2) Place a shunt on jumper J10\n";
		instructions += "3) Disconnect the parallel port cable\n";
		instructions += "4) Apply power to the board\n";
		PromptUser(instructions,PROMPT_OK);
	}

	return true;
}

// upload Flash from loAddr to hiAddr into file
bool XS40Board::UploadFlash(
			string& fileName,		// bitstream data file
			const char* format,		// format - Intel hex, Moto exo, or XESS
			unsigned int loAddr,	// lower address of RAM data
			unsigned int hiAddr	// upper address of RAM data
			)
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	errMsg.SimpleMsg(XSErrorMajor,(string)"The serial EEPROM on the XS40 Board cannot be read back!!\n");
	return false;
}

// program the oscillator frequency
bool XS40Board::SetFreq(
			 int div,				// divisor for osc. frequency
			 bool extOscPresent		// true if external clock drives oscillator
			 )
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	// get the directory where the serial EEPROM interface files are kept
	const char* xstoolsBinDir = FindXSTOOLSBinDir();
	
	int brdIndex;
	for(brdIndex=0; brdIndex<numBoards; brdIndex++)
	{
		if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
			break;
	}
	if(brdIndex>=numBoards)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XS40 Board!\n");
		return false;
	}

	if(strlen(brdInfo[brdIndex].oscBitstreamFile)==0)
	{
		string msg;
		msg = brdModel;
		msg += " does not support oscillator programming!!\n";
		errMsg.SimpleMsg(XSErrorMajor,msg);
		return false;
	}

	string instructions;
	instructions = "Before setting the XS40 Board frequency you must:\n";
	instructions += "1) Remove the power and downloading cable from your XS40 Board\n";
	instructions += "2) Place a shunt on the \"set\" position of jumper J12\n";
	instructions += "3) Reconnect the power cable\n";
	instructions += "4) Reconnect the downloading cable\n";
	instructions += "5) Click on the OK button\n";
	if(PromptUser(instructions,PROMPT_OKCANCEL) == RESPONSE_CANCEL)
		return false;

	string bitFileName;
	bitFileName = xstoolsBinDir;
	bitFileName += "\\";
	bitFileName += brdInfo[brdIndex].oscBitstreamFile;

	if(Configure(bitFileName) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading clock-set circuit!!\n");
		return false;
	}

	osc.SetOscFrequency(div,extOscPresent);

	instructions = "The frequency of your XS40 Board has been set!!\n";
	instructions += "Now do these steps to activate the oscillator:\n";
	instructions += "1) Remove the power and downloading cable from your XS40 Board\n";
	instructions += "2) Move the shunt to the \"osc\" position of jumper J12\n";
	instructions += "3) Reconnect the power cable\n";
	instructions += "4) Reconnect the downloading cable\n";
	instructions += "5) Click on the OK button\n";
	PromptUser(instructions,PROMPT_OK);

	return true;
}

// test the XS40 Board
bool XS40Board::Test(
			void
			)
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	int brdIndex;
	for(brdIndex=0; brdIndex<numBoards; brdIndex++)
	{
		if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
			break;
	}
	if(brdIndex>=numBoards)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XS40 Board!\n");
		return false;
	}

	string msg;

	if(strlen(brdInfo[brdIndex].testBitstreamFile)==0)
	{
		msg = brdModel;
		msg += " does not support self-test!!\n";
		errMsg.SimpleMsg(XSErrorMajor,msg);
		return false;
	}

	// get the directory where the test interface files are kept
	const char* xstoolsBinDir = FindXSTOOLSBinDir();

	string testprogFileName = xstoolsBinDir;
	testprogFileName += "\\";
	testprogFileName += brdInfo[brdIndex].testObjFile;
	if(DownloadRAM(testprogFileName,true,false) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading XS40 Board self-test program!\n");
		return false;
	}

	fpga.Out(1,posTESTRESET,posTESTRESET);	// make sure micro is reset as soon as FPGA is configured

	string bitFileName = xstoolsBinDir;
	bitFileName += "\\";
	bitFileName += brdInfo[brdIndex].testBitstreamFile;
	if(Configure(bitFileName) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading XS40 Board self-test circuit!\n");
		return false;
	}

	fpga.Out(1,posTESTRESET,posTESTRESET);	// make sure the microcontroller is reset
	InsertDelay(10,MILLISECONDS);
	fpga.Out(0,posTESTRESET,posTESTRESET);	// release the microcontroller and do the test

	Progress* progressGauge = new Progress(&errMsg,(string)"Testing " + (string)brdModel,
			(string)"Testing...",0,100);
	int testProgress = 0;

	int prevStatus = fpga.In(posTESTSTATUS,posTESTSTATUS);
	int currStatus;
	int prevDiff = 0;
	int currDiff;
	const float integrationPeriod = 0.5;	// in seconds
	while(true)
	{
		clock_t startTime = clock();
		clock_t endTime = startTime + integrationPeriod * CLOCKS_PER_SEC;
		currDiff = 0;
		while(clock()<endTime)
		{
			currStatus = fpga.In(posTESTSTATUS,posTESTSTATUS);
			if(currStatus != prevStatus)
				currDiff++;
			prevStatus = currStatus;
		}
		testProgress = testProgress>=100 ? 0 : testProgress+10;
		progressGauge->Report(testProgress);
		
		if(currDiff == 0)
		{	// test failed
			msg = (string)"\nYour " + brdModel + (string)" failed the test!\n\n";
			msg += "Check the following:\n";
			msg += " 1) Is your board connected to the PC parallel port?\n";
			msg += " 2) Is your PC parallel port in ECP mode?\n";
			msg += " 3) Is your board the only device connected to the parallel port?\n";
			msg += " 4) Is your board getting power?\n";
			msg += " 5) Is your board sitting on a non-conductive surface?\n";
			msg += " 6) Is the programmable oscillator frequency greater than 1 MHz?\n";
			msg += " 7) Is the shunt on jumper J12 in the osc position?\n";
			msg += " 8) Is the shunt on jumper J7 in the ext position?\n";
			msg += " 9) Are the shunts on jumpers J4 and J11?\n";
			msg += "10) Are the shunts removed from jumpers J6 and J10?\n";
			errMsg.SimpleMsg(XSErrorMajor,msg);
			delete progressGauge;
			return false;
		}
		else
		{	// test still going
			if((currDiff > 2*prevDiff) && (prevDiff>0))
			{	// test succeeded!!
				msg = (string)"\nYour " + brdModel + (string)" passed the test!\n";
				PromptUser(msg,PROMPT_OK);
				delete progressGauge;
				return true;
			}
		}
		prevDiff = currDiff;
	}
	delete progressGauge;
	return false;
}
