/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// This object loads the CPLD with a bitstream,
// uploads/downloads the RAM, sets the oscillator frequency,
// and tests the board.
//

// $Log: xs95brd.cpp $
// Revision 1.1  2001/12/30 15:55:41  DEVB
// Initial revision
//



#include <time.h>
#include <string>
using std::string;
#include "Utils.h"
#include "XSError.h"
#include "XS95Brd.h"


// inversion mask for parallel port connection to XS95 Board
// 00000011: bits  7- 0 are attached to data pins D7-D0
// 00000xxx: bits 15-11 are attached to status pins /S7,S6,S5,S4,S3
// xxxx1001: bits 19-16 are attached to control pins /C3,C2,/C1,/C0
const unsigned int invMask = 0x090003;

// bit position of XC9572XL JTAG pins within parallel port registers
const unsigned int posTCK  = 17;
const unsigned int posTMS  = 18;
const unsigned int posTDI  = 19;
const unsigned int posTDO  = 15;

// bit position of prog. osc. pins within parallel port registers
const unsigned int posOSC  = 0;

// bit positions for board test status
const unsigned int posTESTSTATUS	= 14;
const unsigned int posTESTRESET		= 5;




static XSBoardInfo* brdInfo;
static int numBoards;



// begin methods and functions



// create an object for programming the XS95 Board
XS95Board::XS95Board(void)
{
	XSError* err = new XSError(cerr);	// create error-reporting channel
	numBoards = GetXSBoardInfo(&brdInfo);	// get information about all the models of XS boards
	Setup(err,"",1);
}

// setup the object for programming the XS95 Board
bool XS95Board::Setup(XSError* err,
					 const char* model,
					 unsigned int lptNum
					 )
{
	// store name of board model
	if(brdModel!=NULL)
		free(brdModel);
	brdModel = (char*)malloc(strlen(model)+1);
	if(brdModel==NULL)
		err->SimpleMsg(XSErrorFatal,(string)"ERRLOC: out of memory!!\n");
	strcpy(brdModel,model);

	bool status;
	status = cpld.Setup(err,lptNum,invMask,posTCK,posTMS,posTDI,posTDO);
	status = status && osc.Setup(err,lptNum,invMask,posOSC);
	status = status && ram.Setup(XC95108Type,err,lptNum,invMask,posTCK,posTMS,posTDI,posTDO);
	return status;	// return true if no error occurred for any object setup
}

// check the chip ID and make sure it is an XC95108
bool XS95Board::CheckChipID(void)
{
	// get the chip ID from the interface cpld and compare it to the chip ID for
	// an XC95108 CPLD (ignoring the first 4 bits which increment for each chip revision)
	char* XC95108ID = "1001010100000110000010010011";
	string chipID = cpld.GetChipID();
	if(strncmp(XC95108ID,chipID.data()+4,strlen(XC95108ID)))
	{ // ID did not match
		string instructions;
		instructions = "The CPLD of your XS95 Board is not responding!!\n\n";
		instructions += "\nChip ID = ";
		instructions += chipID;		// display the ID that was received
		instructions += "\n\n";
		instructions += "1) Is power getting to your XS95 Board?\n";
		instructions += "2) Is the downloading cable attached?\n";
		instructions += "\nContinue anyway?\n";
		if(PromptUser(instructions,PROMPT_OKCANCEL) == RESPONSE_CANCEL)
			return false;
	}
	return true;	// ID matched
}


// load a configuration bitstream into the FPGA or Flash
bool XS95Board::Configure(
	 				string& fileName	// bitstream file
					)
{
	XSError& errMsg = cpld.GetErr(); // setup error channel

	string suffix = GetSuffix(fileName);
	if(suffix!="SVF")
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Only .SVF files can be downloaded into the CPLD on the XS95 Board!!\n");
		return false;
	}

	if(CheckChipID() == false)
		return false;

	cpld.InitConfigureCPLD();
	return cpld.ConfigureCPLD(fileName);
}

// load a configuration bitstream into the interface FPGA
bool XS95Board::ConfigureInterface(
					string& fileName	// bitstream file
					)
{
	return false;
}

// download into RAM
bool XS95Board::DownloadRAM(
			string& fileName,		// RAM data file
			bool doStart,			// perform startup operations for RAM downloads
			bool doEnd				// perform terminating operations for RAM downloads
			)
{
	XSError& errMsg = cpld.GetErr(); // setup error channel

	string suffix = GetSuffix(fileName);
	if(suffix!="HEX" && suffix!="EXO" && suffix!="XES" && suffix!="MCS")
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Only .HEX, .MCS, .EXO or .XES files can be downloaded into the RAM on the XS95 Board!!\n");
		return false;
	}

	if(CheckChipID() == false)
		return false;

	if(ram.DownloadRAM(fileName) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading into RAM!!\n");
		return false;
	}

	return true;
}

// upload RAM from loAddr to hiAddr into file
bool XS95Board::UploadRAM(
					string& fileName,	// RAM data file
					const char* format,		// format - Intel hex, Moto exo, or XESS
					unsigned int loAddr,	// lower address of RAM data
					unsigned int hiAddr	// upper address of RAM data
					)
{
	XSError& errMsg = cpld.GetErr(); // setup error channel

	if(CheckChipID() == false)
		return false;

	if(ram.UploadRAM(fileName,format,loAddr,hiAddr) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error uploading from RAM!!\n");
		return false;
	}
	return true;
}

// download into Flash
bool XS95Board::DownloadFlash(
			string& fileName,		// Flash data file
			bool doStart,			// perform startup operations for Flash downloads
			bool doEnd				// perform terminating operations for Flash downloads
			)
{
	XSError& errMsg = cpld.GetErr(); // setup error channel

	errMsg.SimpleMsg(XSErrorMajor,(string)"The XS95 Board does not have a Flash memory device!!\n");
	return false;
}

// upload Flash from loAddr to hiAddr into file
bool XS95Board::UploadFlash(
					string& fileName,	// Flash data file
					const char* format,		// format - Intel hex, Moto exo, or XESS
					unsigned int loAddr,	// lower address of RAM data
					unsigned int hiAddr	// upper address of RAM data
					)
{
	XSError& errMsg = cpld.GetErr(); // setup error channel

	errMsg.SimpleMsg(XSErrorMajor,(string)"The XS95 Board does not have a Flash memory device!!\n");
	return false;
}

// program the oscillator frequency
bool XS95Board::SetFreq(
			int div,						// divisor for osc. frequency
			bool extOscPresent				// true if external clock drives oscillator
			)
{
	XSError& errMsg = cpld.GetErr(); // setup error channel

	// get the directory where the tools and RAM download files are kept
	const char* xstoolsBinDir = FindXSTOOLSBinDir();
	
	int brdIndex;
	for(brdIndex=0; brdIndex<numBoards; brdIndex++)
		if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
			break;
	if(brdIndex>=numBoards)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XS95 Board!\n");
		return false;
	}

	if(strlen(brdInfo[brdIndex].oscBitstreamFile)==0)
	{
		string msg;
		msg = brdModel;
		msg += " does not support oscillator programming!!\n";
		errMsg.SimpleMsg(XSErrorMajor,msg);
		return false;
	}

	string bitFileName;
	bitFileName = xstoolsBinDir;
	bitFileName += "\\";
	bitFileName += brdInfo[brdIndex].oscBitstreamFile;

	string instructions;
	instructions = "Before setting the XS95 Board frequency you must:\n";
	instructions += "1) Remove the power and downloading cable from your XS95 Board\n";
	instructions += "2) Place a shunt on the \"set\" position of jumper J6\n";
	instructions += "3) Reconnect the power cable\n";
	instructions += "4) Reconnect the downloading cable\n";
	instructions += "5) Click on the OK button\n";
	if(PromptUser(instructions,PROMPT_OKCANCEL) == RESPONSE_CANCEL)
		return false;

	cpld.ConfigureCPLD(bitFileName);
	if(errMsg.IsError())
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading clock-set circuit!!\n");
		return false;
	}

	osc.SetOscFrequency(div,extOscPresent);

	bitFileName = xstoolsBinDir;
	bitFileName += "\\";
	bitFileName += brdInfo[brdIndex].eraseBitstreamFile;
	cpld.ConfigureCPLD(bitFileName);
	if(errMsg.IsError())
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error erasing CPLD!!\n");
		return false;
	}
	instructions = "The frequency of your XS95 Board has been set!!\n";
	instructions += "Now do these steps to activate the oscillator:\n";
	instructions += "1) Remove the power and downloading cable from your XS95 Board\n";
	instructions += "2) Move the shunt to the \"osc\" position of jumper J6\n";
	instructions += "3) Reconnect the power cable\n";
	instructions += "4) Reconnect the downloading cable\n";
	instructions += "5) Click on the OK button\n";
	PromptUser(instructions,PROMPT_OK);

	return true;
}

// test the XS95 Board
bool XS95Board::Test(
					 void
					)
{
	XSError& errMsg = cpld.GetErr(); // setup error channel

	int brdIndex;
	for(brdIndex=0; brdIndex<numBoards; brdIndex++)
	{
		if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
			break;
	}
	if(brdIndex>=numBoards)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XS95 Board!\n");
		return false;
	}

	string msg;

	if(strlen(brdInfo[brdIndex].testBitstreamFile)==0)
	{
		msg = brdModel;
		msg += " does not support self-test!!\n";
		errMsg.SimpleMsg(XSErrorMajor,msg);
		return false;
	}

	// get the directory where the test interface files are kept
	const char* xstoolsBinDir = FindXSTOOLSBinDir();

	cpld.Out(1,posTESTRESET,posTESTRESET);	// make sure micro is reset as soon as CPLD is configured

	string testprogFileName = xstoolsBinDir;
	testprogFileName += "\\";
	testprogFileName += brdInfo[brdIndex].testObjFile;

	if(DownloadRAM(testprogFileName,true,false) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading XS95 Board self-test program!\n");
		return false;
	}

	cpld.Out(1,posTESTRESET,posTESTRESET);	// make sure micro is reset as soon as CPLD is configured

	string bitFileName = xstoolsBinDir;
	bitFileName += "\\";
	bitFileName += brdInfo[brdIndex].testBitstreamFile;

	if(Configure(bitFileName) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading XS95 Board self-test circuit!\n");
		return false;
	}

	cpld.Out(1,posTESTRESET,posTESTRESET);	// make sure the microcontroller is reset
	InsertDelay(10,MILLISECONDS);
	cpld.Out(0,posTESTRESET,posTESTRESET);	// release the microcontroller reset and do the test

	Progress* progressGauge = new Progress(&errMsg,(string)"Testing " + (string)brdModel,
			(string)"Testing...",0,100);
	int testProgress = 0;

	int prevStatus = cpld.In(posTESTSTATUS,posTESTSTATUS);
	int currStatus;
	int prevDiff = 0;
	int currDiff;
	const float integrationPeriod = 0.5;	// in seconds
	while(true)
	{
		clock_t startTime = clock();
		clock_t endTime = startTime + integrationPeriod * CLOCKS_PER_SEC;
		currDiff = 0;
		while(clock()<endTime)
		{
			currStatus = cpld.In(posTESTSTATUS,posTESTSTATUS);
			if(currStatus != prevStatus)
				currDiff++;
			prevStatus = currStatus;
		}
		testProgress = testProgress>=100 ? 0 : testProgress+10;
		progressGauge->Report(testProgress);
		
		if(currDiff == 0)
		{	// test failed
			msg = (string)"\nYour " + brdModel + (string)" failed the test!\n\n";
			msg += "Check the following:\n";
			msg += "1) Is your board connected to the PC parallel port?\n";
			msg += "2) Is your PC parallel port in ECP mode?\n";
			msg += "3) Is your board the only device connected to the parallel port?\n";
			msg += "4) Is your board getting power?\n";
			msg += "5) Is your board sitting on a non-conductive surface?\n";
			msg += "6) Is the programmable oscillator frequency greater than 1 MHz?\n";
			msg += "7) Is the shunt on jumper J6 in the osc position?\n";
			msg += "8) Is the shunt on jumper J7 in the ext position?\n";
			errMsg.SimpleMsg(XSErrorMajor,msg);
			delete progressGauge;
			return false;
		}
		else
		{	// test still going
			if((currDiff > 2*prevDiff) && (prevDiff>0))
			{	// test succeeded!!
				msg = (string)"\nYour " + brdModel + (string)" passed the test!\n";
				PromptUser(msg,PROMPT_OK);
				delete progressGauge;
				return true;
			}
		}
		prevDiff = currDiff;
	}
	delete progressGauge;
	return false;
}
