/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// This object loads the FPGA or Flash with a bitstream,
// uploads/downloads the RAM, sets the oscillator frequency,
// and tests the board.
//

// $Log: xsaboard.cpp $
// Revision 1.1  2001/12/30 15:57:50  DEVB
// Initial revision
//



#include <time.h>
#include <string>
using std::string;
#include "Utils.h"
#include "XSError.h"
#include "XSABoard.h"


// inversion mask for parallel port connection to XSA Board
// 00000011: bits  7- 0 are attached to data pins D7-D0
// 00000xxx: bits 15-11 are attached to status pins /S7,S6,S5,S4,S3
// xxxx1001: bits 19-16 are attached to control pins /C3,C2,/C1,/C0
const unsigned int invMask = 0x090003;

// bit position of Spartan2 configuration pins within parallel port registers
const unsigned int posCCLK = 0;
const unsigned int posPROG = 7;
const unsigned int posDLO  = 2;
const unsigned int posDHI  = 5;
const unsigned int posDONE = 11;

// bit position of XC9572XL JTAG pins within parallel port registers
const unsigned int posTCK  = 17;
const unsigned int posTMS  = 18;
const unsigned int posTDI  = 19;
const unsigned int posTDO  = 15;

// bit position of prog. osc. pins within parallel port registers
const unsigned int posOSC  = 16;

// bit positions of RAM access pins within parallel port
const unsigned int posRRESET	= 0;
const unsigned int posRCLK		= 1;
const unsigned int posRDOLSB	= 2;
const unsigned int posRDOMSB	= 5;
const unsigned int posRDILSB	= 11;
const unsigned int posRDIMSB	= 14;
const unsigned int posRSTLSB	= 11;
const unsigned int posRSTMSB	= 14;

// bit positions of Flash access pins within parallel port
const unsigned int posFRESET	=  0;
const unsigned int posFCLK		=  1;
const unsigned int posFDOLSB	=  2;
const unsigned int posFDOMSB	=  5;
const unsigned int posFDILSB	= 11;
const unsigned int posFDIMSB	= 13;
const unsigned int posFSTLSB	= 11;
const unsigned int posFSTMSB	= 13;

// bit positions for board test status
const unsigned int posTESTSTATUS	= 14;



// USERCODE strings for various circuits programmed into the XSA CPLD
static const string oscIntfcCode			= "00111100001100000011111000100001";
static const string flashIntfcCode			= "00111100001100010011111000100001";
static const string flashConfigIntfcCode	= "00111100001100100011111000100001";
static const string testIntfcCode			= "00111100001101000011111000100001";
static const string dwnldIntfcCode			= "00111100001101000011111000100001";
//											   XXXX||||XXXX||||XXXX||||XXXX||||   


static XSBoardInfo* brdInfo;
static int numBoards;



// begin methods and functions



// create an object for programming the XSA Board
XSABoard::XSABoard(void)
{
	XSError* err = new XSError(cerr);	// create error-reporting channel
	numBoards = GetXSBoardInfo(&brdInfo);	// get information about all the models of XS boards
	Setup(err,"",1);
}

// setup the object for programming the XSA Board
bool XSABoard::Setup(XSError* err,
					 const char* model,		// model of XSA Board
					 unsigned int lptNum
					 )
{
	// store name of board model
	if(brdModel!=NULL)
		free(brdModel);
	brdModel = (char*)malloc(strlen(model)+1);
	if(brdModel==NULL)
		err->SimpleMsg(XSErrorFatal,(string)"ERRLOC: out of memory!!\n");
	strcpy(brdModel,model);

	bool status;
	status = fpga.Setup(err,lptNum,invMask,posCCLK,posPROG,posDLO,posDHI,posDONE);
	status = status && cpld.Setup(err,lptNum,invMask,posTCK,posTMS,posTDI,posTDO);
	status = status && osc.Setup(err,lptNum,invMask,posOSC);
	status = status && ram.Setup(err,lptNum,invMask,posRRESET,posRCLK,posRDOLSB,posRDOMSB,posRDILSB,posRDIMSB,posRSTLSB,posRSTMSB,16,24);
	status = status && flash.Setup(err,lptNum,invMask,posFRESET,posFCLK,posFDOLSB,posFDOMSB,posFDILSB,posFDIMSB,posFSTLSB,posFSTMSB);
	return status;	// return true if no error occurred for any object setup
}

// check the interface chip ID and make sure it is an XC9572XL
bool XSABoard::CheckChipID(void)
{
	// get the chip ID from the interface cpld and compare it to the chip ID for
	// an XC9572XL CPLD (ignoring the first 4 bits which increment for each chip revision)
	char* XC9572XLID = "1001011000000100000010010011";
	string chipID = cpld.GetChipID();
	if(strncmp(XC9572XLID,chipID.data()+4,strlen(XC9572XLID)))
	{
		string instructions;
		instructions = "The interface CPLD of your XSA Board is not responding!!\n\n";
		instructions += "\nChip ID = ";
		instructions += chipID;		// display the ID that was received
		instructions += "\n\n";
		instructions += "1) Is power getting to your XSA Board?\n";
		instructions += "2) Is the downloading cable attached?\n";
		instructions += "\nContinue anyway\n?";
		if(PromptUser(instructions,PROMPT_OKCANCEL) == RESPONSE_CANCEL)
			return false;
	}
	return true;	// ID matched
}

// load a configuration bitstream into the FPGA or Flash
bool XSABoard::Configure(
				string& fileName	// bitstream file
				)
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	if(CheckChipID() == false)
		return false;
	
	string suffix = GetSuffix(fileName);
	if(suffix=="BIT")
	{
		int brdIndex;
		for(brdIndex=0; brdIndex<numBoards; brdIndex++)
		{
			if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
				break;
		}
		if(brdIndex>=numBoards)
		{
			errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XSA Board!\n");
			return false;
		}
		
		string type = ConvertToUpperCase(fpga.GetChipType(fileName));
		if(type == "")
		{
			errMsg.SimpleMsg(XSErrorMajor,(string)"The .BIT file does not identify the target FPGA!!\n");
			return false;
		}
		if(type != brdInfo[brdIndex].chipType)
		{
			string msg;
			msg = ".BIT file is for the ";
			msg += type;
			msg += " but this is not the FPGA on the ";
			msg += brdModel;
			msg += "!!\n";
			errMsg.SimpleMsg(XSErrorMajor,msg);
			return false;
		}

		string usercode = cpld.GetUSERCODE();
		if((usercode==oscIntfcCode) || (usercode==flashIntfcCode) || (usercode==flashConfigIntfcCode))
		{ // the interface CPLD is not configured with the standard downloading circuit
			// get the file where the standard downloading interface bitstream is kept
			const char* xstoolsBinDir = FindXSTOOLSBinDir();
			
			string bitFileName;
			bitFileName = xstoolsBinDir;
			bitFileName += "\\";
			bitFileName += brdInfo[brdIndex].dwnldIntfcBitstreamFile;
			
			if(ConfigureInterface(bitFileName) == false)
			{
				errMsg.SimpleMsg(XSErrorMajor,(string)"Error configuring CPLD with downloading interface circuit!\n");
				return false;
			}
			fpga.EnableFastDownload(true);
		}
		else if((usercode==dwnldIntfcCode) || (usercode==testIntfcCode))
		{ // the CPLD contains the fast downloading interface
			fpga.EnableFastDownload(true);
		}
		else
		{ // can't tell what the CPLD contains, so assume the user loaded it with a legacy, low-speed interface
			fpga.EnableFastDownload(false);
		}
		
		fpga.InitConfigureFPGA();
		return fpga.ConfigureFPGA(fileName);
	}
	else if(suffix=="SVF")
	{
		return ConfigureInterface(fileName);
	}
	
	errMsg.SimpleMsg(XSErrorMajor,(string)"Only .BIT or .SVF files can be downloaded into the FPGA or CPLD on the XSA Board!!");
	return false;
}

// load a configuration bitstream into the interface FPGA
bool XSABoard::ConfigureInterface(
				string& fileName	// bitstream file
				)
{
	string suffix = GetSuffix(fileName);
	if(suffix!="SVF")
		return false;

	if(CheckChipID() == false)
		return false;

	cpld.InitConfigureCPLD();
	return cpld.ConfigureCPLD(fileName);
}

// download into RAM
bool XSABoard::DownloadRAM(
			string& fileName,		// RAM data file
			bool doStart,			// perform startup operations for RAM downloads
			bool doEnd				// perform terminating operations for RAM downloads
			)
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	string suffix = GetSuffix(fileName);
	if(suffix!="HEX" && suffix!="EXO" && suffix!="XES" && suffix!="MCS")
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Only .HEX, .MCS, .EXO or .XES files can be downloaded into the RAM on the XSA Board!!\n");
		return false;
	}
	
	int brdIndex;
	for(brdIndex=0; brdIndex<numBoards; brdIndex++)
	{
		if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
			break;
	}
	if(brdIndex>=numBoards)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XSA Board!\n");
		return false;
	}

	if(strlen(brdInfo[brdIndex].ramBitstreamFile)==0)
	{
		string msg;
		msg = brdModel;
		msg += " does not support RAM downloads!!\n";
		errMsg.SimpleMsg(XSErrorMajor,msg);
		return false;
	}

	// get the directory where the RAM interface files are kept
	const char* xstoolsBinDir = FindXSTOOLSBinDir();

	string bitFileName;
	bitFileName = xstoolsBinDir;
	bitFileName += "\\";
	bitFileName += brdInfo[brdIndex].ramBitstreamFile;

	if(doStart)
	{
		if(Configure(bitFileName) == false)
		{
			errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading RAM interface circuit!!\n");
			return false;
		}
	}

	if(ram.DownloadRAM(fileName) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading into RAM!!\n");
		return false;
	}

	return true;
}

// upload RAM from loAddr to hiAddr into file
bool XSABoard::UploadRAM(
				string& fileName,		// RAM data file
				const char* format,			// format - Intel hex, Moto exo, or XESS
				unsigned int loAddr,	// lower address of RAM data
				unsigned int hiAddr		// upper address of RAM data
				)
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	int brdIndex;
	for(brdIndex=0; brdIndex<numBoards; brdIndex++)
	{
		if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
			break;
	}
	if(brdIndex>=numBoards)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XSA Board!\n");
		return false;
	}

	if(strlen(brdInfo[brdIndex].ramBitstreamFile)==0)
	{
		string msg;
		msg = brdModel;
		msg += " does not support RAM uploads!!\n";
		errMsg.SimpleMsg(XSErrorMajor,msg);
		return false;
	}

	// get the directory where the RAM interface files are kept
	const char* xstoolsBinDir = FindXSTOOLSBinDir();

	string bitFileName;
	bitFileName = xstoolsBinDir;
	bitFileName += "\\";
	bitFileName += brdInfo[brdIndex].ramBitstreamFile;

	if(Configure(bitFileName) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading RAM interface circuit!!\n");
		return false;
	}

	if(ram.UploadRAM(fileName,format,loAddr,hiAddr) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error uploading from RAM!!\n");
		return false;
	}
	return true;
}

// download into Flash
bool XSABoard::DownloadFlash(
			string& fileName,		// Flash data file
			bool doStart,			// perform startup operations for Flash downloads
			bool doEnd				// perform terminating operations for Flash downloads
			)
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	string suffix = GetSuffix(fileName);
	if(suffix!="HEX" && suffix!="EXO" && suffix!="XES" && suffix!="MCS")
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Only .HEx, .MCS, .EXO or .HEX files can be downloaded into the Flash on the XSA Board!!\n");
		return false;
	}
	
	int brdIndex;
	for(brdIndex=0; brdIndex<numBoards; brdIndex++)
	{
		if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
			break;
	}
	if(brdIndex>=numBoards)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XSA Board!\n");
		return false;
	}

	if(strlen(brdInfo[brdIndex].flashBitstreamFile)==0)
	{
		string msg;
		msg = brdModel;
		msg += " does not support Flash programming!!\n";
		errMsg.SimpleMsg(XSErrorMajor,msg);
		return false;
	}

	if(CheckChipID() == false)
		return false;

	bool loadFlashIntfc;
	string usercode = cpld.GetUSERCODE();
	if((usercode==oscIntfcCode) || (usercode==dwnldIntfcCode) || 
		(usercode==flashConfigIntfcCode) || (usercode==testIntfcCode))
	{ // the interface CPLD is not configured with the standard downloading circuit
		loadFlashIntfc = true;
	}
	else if(usercode==flashIntfcCode)
	{ // the CPLD already contains the Flash programming interface bitstream
		loadFlashIntfc = false;
	}
	else
	{ // can't tell what the CPLD contains, so load the Flash programming interface
		loadFlashIntfc = true;
	}

	// get the directory where the serial EEPROM interface files are kept
	const char* xstoolsBinDir = FindXSTOOLSBinDir();

	string bitFileName;
	bitFileName = xstoolsBinDir;
	bitFileName += "\\";
	bitFileName += brdInfo[brdIndex].flashBitstreamFile;
	
	if(doStart || loadFlashIntfc)
	{
		if(ConfigureInterface(bitFileName) == false)
		{
			errMsg.SimpleMsg(XSErrorMajor,(string)"Error configuring CPLD with Flash programming circuit!\n");
			return false;
		}
	}
	
	if(flash.DownloadFlash(fileName,doStart) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error programming the Flash on the XSA Board!\n");
		return false;
	}
	
	bitFileName = xstoolsBinDir;
	bitFileName += "\\";
	bitFileName += brdInfo[brdIndex].flashConfigBitstreamFile;
	
	if(doEnd)
	{
		if(ConfigureInterface(bitFileName) == false)
		{
			errMsg.SimpleMsg(XSErrorMajor,(string)"Error configuring CPLD with Flash configuration circuit!\n");
			return false;
		}
	}

	return true;
}

// upload Flash from loAddr to hiAddr into file
bool XSABoard::UploadFlash(
				string& fileName,		// Flash data file
				const char* format,			// format - Intel hex, Moto exo, or XESS
				unsigned int loAddr,	// lower address of RAM data
				unsigned int hiAddr		// upper address of RAM data
				)
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	int brdIndex;
	for(brdIndex=0; brdIndex<numBoards; brdIndex++)
	{
		if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
			break;
	}
	if(brdIndex>=numBoards)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XSA Board!\n");
		return false;
	}

	if(strlen(brdInfo[brdIndex].flashBitstreamFile)==0)
	{
		string msg;
		msg = brdModel;
		msg += " does not support Flash uploads!!\n";
		errMsg.SimpleMsg(XSErrorMajor,msg);
		return false;
	}

	if(CheckChipID() == false)
		return false;
	
	bool loadFlashIntfc;
	string usercode = cpld.GetUSERCODE();
	if((usercode==oscIntfcCode) || (usercode==dwnldIntfcCode) || 
		(usercode==flashConfigIntfcCode) || (usercode==testIntfcCode))
	{ // the interface CPLD is not configured with the standard downloading circuit
		loadFlashIntfc = true;
	}
	else if(usercode==flashIntfcCode)
	{ // the CPLD already contains the Flash programming interface bitstream
		loadFlashIntfc = false;
	}
	else
	{ // can't tell what the CPLD contains, so load the Flash programming interface
		loadFlashIntfc = true;
	}

	const char* xstoolsBinDir = FindXSTOOLSBinDir();
	string bitFileName;
	
	if(loadFlashIntfc)
	{
		// get the directory where the Flash interface files are kept
		bitFileName = xstoolsBinDir;
		bitFileName += "\\";
		bitFileName += brdInfo[brdIndex].flashBitstreamFile;
		
		if(ConfigureInterface(bitFileName) == false)
		{
			errMsg.SimpleMsg(XSErrorMajor,(string)"Error configuring CPLD with Flash programming circuit!\n");
			return false;
		}
	}

	if(flash.UploadFlash(fileName,format,loAddr,hiAddr) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error uploading from Flash!!\n");
		return false;
	}

	bitFileName = xstoolsBinDir;
	bitFileName += "\\";
	bitFileName += brdInfo[brdIndex].flashConfigBitstreamFile;
	
	if(ConfigureInterface(bitFileName) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error configuring CPLD with Flash configuration circuit!\n");
		return false;
	}

	return true;
}

// program the oscillator frequency
bool XSABoard::SetFreq(
				int div,						// divisor for osc. frequency
				bool extOscPresent				// true if external clock drives oscillator
				)
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	string instructions;
	instructions = "Before setting the XSA Board frequency you must:\n\n";
	instructions += "1) Remove the power and downloading cable from your XSA Board\n";
	instructions += "2) Place a shunt on the \"set\" position of jumper J6\n";
	instructions += "3) Reconnect the power cable\n";
	instructions += "4) Reconnect the downloading cable\n";
	instructions += "5) Click on the OK button\n";
	if(PromptUser(instructions,PROMPT_OKCANCEL) == RESPONSE_CANCEL)
		return false;

	osc.SetOscFrequency(div,extOscPresent);

	instructions = "The frequency of your XSA Board has been set!!\n";
	instructions += "Now do these steps to activate the oscillator:\n\n";
	instructions += "1) Remove the power and downloading cables from your XSA Board\n";
	instructions += "2) Move the shunt to the \"osc\" position of jumper J6\n";
	instructions += "3) Reconnect the power cable\n";
	instructions += "4) Reconnect the downloading cable\n";
	instructions += "5) Click on the OK button\n";
	PromptUser(instructions,PROMPT_OK);
	return true;
}

// test the XSA Board
bool XSABoard::Test(
			void
			)
{
	XSError& errMsg = fpga.GetErr(); // setup error channel

	int brdIndex;
	for(brdIndex=0; brdIndex<numBoards; brdIndex++)
	{
		if(strcmp(brdModel,brdInfo[brdIndex].brdModel) == 0)
			break;
	}
	if(brdIndex>=numBoards)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Unknown type of XSA Board!\n");
		return false;
	}

	string msg;

	if(strlen(brdInfo[brdIndex].testIntfcBitstreamFile)==0)
	{
		msg = brdModel;
		msg += " does not support self-test!!\n";
		errMsg.SimpleMsg(XSErrorMajor,msg);
		return false;
	}

	if(CheckChipID() == false)
		return false;
	
	bool loadTestIntfc;
	string usercode = cpld.GetUSERCODE();
	if((usercode==oscIntfcCode) || (usercode==flashIntfcCode) || 
		(usercode==flashConfigIntfcCode) || (usercode==dwnldIntfcCode))
	{ // the interface CPLD is not configured with the standard testing circuit
		loadTestIntfc = true;
	}
	else if(usercode==testIntfcCode)
	{ // the CPLD already contains the standard testing interface bitstream
		loadTestIntfc = false;
	}
	else
	{ // can't tell what the CPLD contains, so load standard testing interface
		loadTestIntfc = true;
	}
	
	const char* xstoolsBinDir = FindXSTOOLSBinDir();
	string bitFileName;
	
	if(loadTestIntfc)
	{
		// get the directory where the test interface files are kept
		bitFileName = xstoolsBinDir;
		bitFileName += "\\";
		bitFileName += brdInfo[brdIndex].testIntfcBitstreamFile;
		
		if(ConfigureInterface(bitFileName) == false)
		{
			errMsg.SimpleMsg(XSErrorMajor,(string)"Error configuring CPLD with self-test interface circuit!\n");
			return false;
		}
	}

	bitFileName = xstoolsBinDir;
	bitFileName += "\\";
	bitFileName += brdInfo[brdIndex].testBitstreamFile;

	if(Configure(bitFileName) == false)
	{
		errMsg.SimpleMsg(XSErrorMajor,(string)"Error downloading XSA Board self-test circuit!\n");
		return false;
	}

	Progress* progressGauge = new Progress(&errMsg,(string)"Testing " + (string)brdModel,
			(string)"Testing...",0,100);
	int testProgress = 0;

	int prevStatus = cpld.In(posTESTSTATUS,posTESTSTATUS);
	int currStatus;
	int prevDiff = 0;
	int currDiff;
	const float integrationPeriod = 0.5;	// in seconds
	while(true)
	{
		clock_t startTime = clock();
		clock_t endTime = startTime + integrationPeriod * CLOCKS_PER_SEC;
		currDiff = 0;
		while(clock()<endTime)
		{
			currStatus = cpld.In(posTESTSTATUS,posTESTSTATUS);
			if(currStatus != prevStatus)
				currDiff++;
			prevStatus = currStatus;
		}
		testProgress = testProgress>=100 ? 0 : testProgress+10;
		progressGauge->Report(testProgress);
		
		if(currDiff == 0)
		{	// test failed
			msg = (string)"\nYour " + brdModel + (string)" failed the test!\n\n";
			msg += "Check the following:\n";
			msg += "1) Is your board connected to the PC parallel port?\n";
			msg += "2) Is your PC parallel port in ECP mode?\n";
			msg += "3) Is your board the only device connected to the parallel port?\n";
			msg += "4) Is your board getting power?\n";
			msg += "5) Is your board sitting on a non-conductive surface?\n";
			msg += "6) Is the programmable oscillator frequency set to 50 MHz?\n";
			msg += "7) Is the shunt on the \'xs\' position of jumper J9?\n";
			msg += "8) Is the shunt on the \'osc\' position of jumper J6?\n";
			msg += "9) Is the shunt on jumper J2?\n";
			errMsg.SimpleMsg(XSErrorMajor,msg);
			delete progressGauge;
			return false;
		}
		else
		{	// test still going
			if((currDiff > 2*prevDiff) && (prevDiff>0))
			{	// test succeeded!!
				msg = (string)"\nYour " + brdModel + (string)" passed the test!\n";
				PromptUser(msg,PROMPT_OK);
				delete progressGauge;
				return true;
			}
		}
		prevDiff = currDiff;
	}
	delete progressGauge;
	return false;
}
