/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// This object allows the reporting of errors with a standardized format.
//

// $Log: xserror.cpp $
// Revision 1.6  2001/12/30 16:16:24  DEVB
// Added a batch switch that turns off error messages during automated
// runs of the XSTOOLS.
//
// Revision 1.5  2001/01/03 15:15:22  DEVB
// Fixed location of header file for generating pop-up message windows.
//
// Revision 1.4  2000/09/10 16:53:00  DEVB
// Added #ifdef...#endif to remove calls to Windows message box functions
// when xserror object is used in command-line applications.
//
// Revision 1.3  2000/09/10 13:38:57  DEVB
// Error message object now gathers error messages for display in a
// Windows message box.
//
// Revision 1.2  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:02:44  DEVB
// Initial revision
//

#include <assert.h>
#ifdef _WINDOWS
#include <afxwin.h>
#endif
#include "xserror.h"


// begin methods and functions

// init error object with output going to stream s
XSError::XSError(ostream_withassign& s)
{
	for(XSErrorSeverity i=XSErrorMin; i<XSErrorMax; i=(XSErrorSeverity)((int)i+1))
		SetNumErrors(i,0);
	ostream_withassign::operator=(s);
	state = XSErrorInitial;
	severity = XSErrorNone;
	header = "";
	storedMsg = "";
	batch = false;
}

// destruct error object
XSError::~XSError(void)
{
	;
}

// assign contents of one source error object to another
XSError& XSError::operator= (XSError& src)
{
	for(XSErrorSeverity i=XSErrorMin; i<XSErrorMax; i=(XSErrorSeverity)((int)i+1))
		SetNumErrors(i,src.GetNumErrors(i));
	ostream_withassign::operator=((ostream_withassign&)src);
	SetState(src.GetState());
	SetSeverity(src.GetSeverity());
	SetHeader(src.GetHeader());
	storedMsg = this->storedMsg;
	batch = this->batch;
	return *this;
}

// get number of errors of a certain type that have occurred
unsigned int XSError::GetNumErrors(XSErrorSeverity s) const
{
	assert(s>=XSErrorMin && s<XSErrorMax);
	return numErrors[s];
}

// set number of errors of a certain type that have occurred
void XSError::SetNumErrors(XSErrorSeverity s, unsigned int n)
{
	assert(s>=XSErrorMin && s<XSErrorMax);
	numErrors[s] = n;
}

// returns true if errors were recorded by this error object
bool XSError::IsError(void) const
{
	for(XSErrorSeverity i=XSErrorMin; i<XSErrorMax; i=(XSErrorSeverity)((int)i+1))
		if(GetNumErrors(i)) return true; // yes, there were errors
		return false; // no errors were recorded
}

XSErrorSeverity XSError::GetSeverity(void) const
{
	return severity;
}

// set severity of next error message
void XSError::SetSeverity(XSErrorSeverity s)
{
	severity = s;
	switch(GetSeverity())
	{
	case XSErrorFatal:
		(*this) << GetHeader().data() << " FATAL ERROR: ";
		SetState(XSErrorInMessage);
		break;
	case XSErrorMajor:
		(*this) << GetHeader().data() << " MAJOR ERROR: ";
		flush();
		SetState(XSErrorInMessage);
		break;
	case XSErrorMinor:
		(*this) << GetHeader().data() << " MINOR ERROR: ";
		flush();
		SetState(XSErrorInMessage);
		break;
	case XSErrorNone:
		(*this) << GetHeader().data() << ": ";
		flush();
		SetState(XSErrorInMessage);
		break;
	default:
		SetSeverity(XSErrorMinor);
		(*this) << "\nerror severity was incorrectly set!\n";
		flush();
		EndMsg();
		break;
	}
}

XSErrorState XSError::GetState(void) const
{
	return state;
}

void XSError::SetState(XSErrorState s )
{
	state = s;
}

// set header string for each error message
void XSError::SetHeader(string& h)
{
	header = h;
}

string& XSError::GetHeader(void)
{
	return header;
}

// enable/disable messages to user
void XSError::EnableBatch(bool b)
{
	batch = b;	// disable user prompts (i.e. enable batch processing) when true
}

// end an error message
void XSError::EndMsg(void)
{
	switch(GetSeverity())
	{
	case XSErrorFatal:
		SetNumErrors(XSErrorFatal,GetNumErrors(XSErrorFatal)+1);
		operator<<("Abnormal termination of program\n");
#ifdef _WINDOWS
		if(!batch) AfxMessageBox(storedMsg.data(),MB_ICONSTOP);
#endif
		exit(1);
		break;
	case XSErrorMajor:
		SetNumErrors(XSErrorMajor,GetNumErrors(XSErrorMajor)+1);
#ifdef _WINDOWS
		if(!batch) AfxMessageBox(storedMsg.data(),MB_ICONEXCLAMATION);
#endif
		storedMsg = "";
		SetState(XSErrorInitial);
		break;
	case XSErrorMinor:
		SetNumErrors(XSErrorMinor,GetNumErrors(XSErrorMinor)+1);
#ifdef _WINDOWS
		if(!batch) AfxMessageBox(storedMsg.data(),MB_ICONEXCLAMATION);
#endif
		storedMsg = "";
		SetState(XSErrorInitial);
		break;
	case XSErrorNone:
		SetNumErrors(XSErrorNone,GetNumErrors(XSErrorNone)+1);
#ifdef _WINDOWS
		if(!batch) AfxMessageBox(storedMsg.data(),MB_ICONINFORMATION);
#endif
		storedMsg = "";
		SetState(XSErrorInitial);
		break;
	default:
		SetSeverity(XSErrorMinor);
		(*this) << "\nerror severity was not set!\n";
		flush();
#ifdef _WINDOWS
		if(!batch) AfxMessageBox(storedMsg.data(),MB_ICONINFORMATION);
#endif
		storedMsg = "";
		SetState(XSErrorInitial);
		break;
	}
}

// a simple interface for sending an error message
void XSError::SimpleMsg(XSErrorSeverity s, string& msg)
{
	SetSeverity(s);
	(*this) << msg.data();
	EndMsg();
}

// overload << operator to assist in displaying error messages in a window
XSError& XSError::operator<<(long n)
{
	char num[30];
	sprintf(num,"%ld",n);
	string s = num;
	storedMsg = storedMsg + s;
	(ostream_withassign&)(*this) << num;
	return *this;
}

// overload << operator to assist in displaying error messages in a window
XSError& XSError::operator<<(const char* msg)
{
	string s = msg;
	storedMsg = storedMsg + s;
	(ostream_withassign&)(*this) << msg;
	return *this;
}

// overload << operator to assist in displaying error messages in a window
XSError& XSError::operator<<(string& msg)
{
	storedMsg = storedMsg + msg;
	(ostream_withassign&)(*this) << msg.data();
	return *this;
}

// overload << operator to assist in displaying error messages in a window
XSError& XSError::operator<<(ostream& s)
{
	(ostream_withassign&)(*this) << s;
	return *this;
}

