/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// This object allows the reporting of errors with a standardized format.
//

// $Log: xserror.h $
// Revision 1.4  2001/12/30 16:16:24  DEVB
// Added a batch switch that turns off error messages during automated
// runs of the XSTOOLS.
//
// Revision 1.3  2000/09/10 13:38:57  DEVB
// Error message object now gathers error messages for display in a
// Windows message box.
//
// Revision 1.2  1999/03/30 18:27:41  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:51:22  DEVB
// Initial revision
//

#ifndef XSError_H
#define XSError_H

#include <iostream.h>
#include <string>

using std::string;

#define ERRLOC	__FILE__(__LINE__)

// severity of errors
typedef enum
{
	XSErrorMin,		// minimum error index
		XSErrorNone,    // no error (not initialized)
		XSErrorMinor,   // minor error (no abort)
		XSErrorMajor,
		XSErrorFatal,	// fatal error (causes abort)
		XSErrorMax		// maximum error index
} XSErrorSeverity;

// states of the error object
typedef enum
{
	XSErrorInitial,   // just starting error message
		XSErrorInMessage, // currently printing error message
} XSErrorState;


// error object which manages printing error messages and terminating
// the main program if necessary.
// Example:
//	XSError err(cerr);
//	err.setHeader("XESS ");
//	err.setSeverity(XSErrorMinor);
//	err << "this is a minor error" << endl;
//	err.endMsg();
//  err.simpleMsg(XSErrorMinor,"this is another minor error");
class XSError : public ostream_withassign
{
	public:

	/* begin XSError methods */

	/* method: XSError */
	// init error object with output going to stream s
	 XSError(ostream_withassign& s);

	/* method: ~XSError */
	// destruct error object
	 ~XSError(void);

	/* method: operator= */
	// assign contents of one source error object to another
	XSError& operator=(XSError& src);

	/* method: GetNumErrors */
	// get number of errors of a certain type that have occurred
	unsigned int GetNumErrors(XSErrorSeverity s) const;

	/* method: SetNumErrors */
	// set number of errors of a certain type that have occurred
	void SetNumErrors(XSErrorSeverity s, unsigned int n);

	/* method: IsError */
	// returns true if errors were recorded by this error object
	bool IsError(void) const;

	/* method: SetSeverity */
	// set severity of next error message
	void SetSeverity(XSErrorSeverity s);

	/* method: SetHeader */
	// set header string for each error message
	void SetHeader(string& h);

	/* method: EnableBatch */
	// enable/disable messages to user
	void EnableBatch(bool b);

	/* method: EndMsg */
	// end an error message
	void EndMsg(void);

	/* method: SimpleMsg */
	// a simple interface for sending an error message
	void SimpleMsg(XSErrorSeverity s, string& msg);

	/* method: operator<< */
	// overload << operator to assist in displaying error messages in a window
	XSError& operator<<(long n);

	/* method: operator<< */
	// overload << operator to assist in displaying error messages in a window
	XSError& operator<<(const char*  msg);

	/* method: operator<< */
	// overload << operator to assist in displaying error messages in a window
	XSError& operator<<(string& msg);

	/* method: operator<< */
	// overload << operator to assist in displaying error messages in a window
	XSError& operator<<(ostream& s);

	/* end XSError methods */

	private:

	/* method: GetSeverity */
	XSErrorSeverity GetSeverity(void) const;

	/* method: SetState */
	void SetState(XSErrorState s );

	/* method: GetState */
	XSErrorState GetState(void) const;

	/* method: GetHeader */
	string& GetHeader(void);

	unsigned int numErrors[XSErrorMax];		// counters for various grades of errors
	XSErrorState state;						// records state of the error reporting process
	XSErrorSeverity severity;				// severity of current error report
	string header;							// header for each error message
	string storedMsg;						// stored error message for display in window
	bool batch;								// disables messages to user when true
};

/* begin XSError functions */

/* end XSError functions */

#endif
